package fr.inria.gforge.dspace.overlay.benchmarks.analyzer.parser;

import fr.inria.gforge.dspace.overlay.benchmarks.measurement.AddStatementMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.AddStatementsMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.PeerMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.PeersMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.QueryMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.TaskTimeMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.TasksTimeMeasurementResults;

/**
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: BenchmarkBean.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class BenchmarkBean implements Comparable<BenchmarkBean> {

	public enum Type {
		LOCAL, REMOTE
	};

	private Type type;

	private int nbPeers;

	private TasksTimeMeasurementResults tasksTimeMeasurementResults;

	private AddStatementsMeasurementResults addRandomStatementsMeasurementResults;

	private QueryMeasurementResults queryAllMeasurementResults;
	
	private PeersMeasurementResults peersMeasurementResults;

	private double addRandomStatementsTimeAverage;

	private double addRandomStatementsSingleTripAverage;

	private double addRandomStatementsReturnTripAverage;

	public BenchmarkBean(
			Type type,
			int nbPeers,
			TasksTimeMeasurementResults tasksTimeMeasurementResults,
			AddStatementsMeasurementResults addRandomStatementsMeasurementResults,
			QueryMeasurementResults queryAllMeasurementResults,
			PeersMeasurementResults peersMeasurementResults) {
		this.type = type;
		this.nbPeers = nbPeers;
		this.tasksTimeMeasurementResults = tasksTimeMeasurementResults;
		this.addRandomStatementsMeasurementResults = addRandomStatementsMeasurementResults;
		this.queryAllMeasurementResults = queryAllMeasurementResults;
		this.peersMeasurementResults = peersMeasurementResults;

		double[] averages = this.computeAddRandomStatementsAverages();
		this.addRandomStatementsTimeAverage = averages[0];
		this.addRandomStatementsSingleTripAverage = averages[1];
		this.addRandomStatementsReturnTripAverage = averages[2];
	}

	/**
	 * Returns an array composed of three averages. The first one is the average
	 * time to execute add random statement task. The second is the number of
	 * steps for round trip average and the last is the number of steps for
	 * return trip average.
	 * 
	 * @return an array composed of three averages. The first one is the average
	 *         time to execute add random statement task. The second is the
	 *         number of steps for round trip average and the last is the number
	 *         of steps for return trip average.
	 */
	private double[] computeAddRandomStatementsAverages() {
		double[] averages = new double[3];

		for (AddStatementMeasurementResults result : this.addRandomStatementsMeasurementResults
				.getMeasurementsResults()) {
			averages[0] += result.getExecutionTime();
			averages[1] += result.getNbHopsSingleTrip();
			averages[2] += result.getNbHopsReturnTrip();
		}

		averages[0] /= this.addRandomStatementsMeasurementResults
				.getMeasurementsResults().size();
		averages[1] /= this.addRandomStatementsMeasurementResults
				.getMeasurementsResults().size();
		averages[2] /= this.addRandomStatementsMeasurementResults
				.getMeasurementsResults().size();

		return averages;
	}

	public double getAddStatementsTimeAverage() {
		return this.addRandomStatementsTimeAverage;
	}

	public double getAddStatementsSingleTripAverage() {
		return this.addRandomStatementsSingleTripAverage;
	}

	public double getAddStatementsReturnTripAverage() {
		return this.addRandomStatementsReturnTripAverage;
	}

	public TasksTimeMeasurementResults getTasksTimeMeasurementResults() {
		return this.tasksTimeMeasurementResults;
	}

	public AddStatementsMeasurementResults getAddRandomStatementsMeasurementResults() {
		return this.addRandomStatementsMeasurementResults;
	}

	public QueryMeasurementResults getQueryAllMeasurementResults() {
		return queryAllMeasurementResults;
	}

	public PeersMeasurementResults getPeersMeasurementResults() {
		return this.peersMeasurementResults;
	}

	public int getNbPeers() {
		return this.nbPeers;
	}

	public Type getType() {
		return this.type;
	}

	@Override
	public int compareTo(BenchmarkBean bb) {
		if (this.nbPeers < bb.getNbPeers()) {
			return -1;
		} else if (this.nbPeers > bb.getNbPeers()) {
			return 1;
		} else {
			return 0;
		}
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("* Benchmark with ");
		buf.append(this.nbPeers);

		buf.append("\n  > tasks executed:\n");
		for (TaskTimeMeasurementResults result : 
				this.tasksTimeMeasurementResults.getMeasurementsResults()) {
			buf.append("    - ");
			buf.append(result.getTaskSimpleName());
			buf.append(", ");
			buf.append(result.getExecutionTime());
			buf.append(" ms\n");
		}
		buf.append("  > add random statements:\n");
		for (AddStatementMeasurementResults result : 
				this.addRandomStatementsMeasurementResults.getMeasurementsResults()) {
			buf.append("    - startTime=");
			buf.append(result.getStartTime());
			buf.append(", execution time=");
			buf.append(result.getExecutionTime());
			buf.append(" ms");
			buf.append(", nbHopsForSingleTrip=");
			buf.append(result.getNbHopsSingleTrip());
			buf.append(", nbHopsForReturnTrip=");
			buf.append(result.getNbHopsReturnTrip());
			buf.append("\n");
		}
		buf.append("  > peers information:\n");
		for (PeerMeasurementResults result : 
				this.peersMeasurementResults.getMeasurementsResults()) {
			buf.append("    - id=");
			buf.append(result.getIdentifier());
			buf.append(", nbStatements=");
			buf.append(result.getNbStatementsManaged());
			buf.append("\n");
		}

		return buf.toString();
	}

}
