package fr.inria.gforge.dspace.overlay.benchmarks;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.inria.gforge.dspace.overlay.benchmarks.measurement.AbstractMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.TaskTimeMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.measurement.TasksTimeMeasurementResults;
import fr.inria.gforge.dspace.overlay.benchmarks.tasks.AbstractTask;

/**
 * A benchmark is composed of a list of {@link AbstractTask}s to measure.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: Benchmark.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public abstract class Benchmark {

	private final static Logger logger = LoggerFactory.getLogger(Benchmark.class);
	
    private ExecutorService threadsPool;
    
    private List<AbstractTask> tasksToExecute;
    
    private List<AbstractMeasurementResults> measurementResults;
    
    private boolean running = false;
    
	public boolean add(AbstractMeasurementResults e) {
		return this.measurementResults.add(e);
	}

	/**
	 * Returns the name of the file in which measurements are saved.
	 * 
	 * @return the name of the file in which measurements are saved.
	 */
    public abstract String getOutputFilename();
    
	public List<AbstractMeasurementResults> getMeasurementResults() {
		return this.measurementResults;
	}

	/**
	 * Returns the task to execute for setting up the benchmark.
	 * 
	 * @return the task to execute for setting up the benchmark.
	 */
    protected abstract AbstractTask getTaskToExecuteForSetUp();
    
    /**
     * Returns a list of tasks to execute.
     * 
     * @return a list of tasks to execute.
     */
    protected abstract List<AbstractTask> getTasksToExecute();
    
	/**
	 * Returns the task to execute for tearing down the benchmark.
	 * 
	 * @return the task to execute for tearing down the benchmark.
	 */
    protected abstract AbstractTask getTaskToExecuteForTearDown();
    
    public synchronized void run(boolean exportMeasurements) {
        if (!this.running) {
            this.running = true;
        }
        
        int nbThreads = System.getProperty("dspace.overlay.benchmarks.threads") != null ?
        					Integer.parseInt(System.getProperty("dspace.overlay.benchmarks.threads")) :
        						8 * Runtime.getRuntime().availableProcessors();
        
        this.threadsPool = Executors.newFixedThreadPool(nbThreads);
        
        this.tasksToExecute = this.getTasksToExecute();
        this.tasksToExecute.add(0, this.getTaskToExecuteForSetUp());
        this.tasksToExecute.add(this.tasksToExecute.size(), this.getTaskToExecuteForTearDown());
        
        this.measurementResults = 
        	new ArrayList<AbstractMeasurementResults>();
        this.measurementResults.add(
        		new TasksTimeMeasurementResults(
        				new ArrayList<TaskTimeMeasurementResults>(
        						this.tasksToExecute.size())));
        
        long taskMeasurementTime;
        
        for (AbstractTask task : this.tasksToExecute) {
        	taskMeasurementTime = this.measureTaskTime(task);
        	
        	((TasksTimeMeasurementResults)this.measurementResults.get(0)).
        		getMeasurementsResults().add(
        			new TaskTimeMeasurementResults(
        					task.getClass().getSimpleName(), 
        					taskMeasurementTime));
        	logger.info("{} executed in {} ms", 
        				task.getClass().getSimpleName(),
        				taskMeasurementTime);
        }

        this.threadsPool.shutdown();
        
        if (exportMeasurements) {
        	this.exportMeasurementResults();
        }
        
        this.running = false;
    }
    
    /**
     * Keeps the time to execute the specified task.
     * 
     * @param task the task to measure.
     * 
     * @return the measurement time computed.
     */
    private long measureTaskTime(AbstractTask task) {
    	long start = System.currentTimeMillis();
    	task.run();
    	return System.currentTimeMillis() - start;
    }
    
    public void exportMeasurementResults() {
        File parent = 
        	new File(System.getProperty("dspace.bundle.home", ".")
                + File.separator + "benchmarks");
        if (!parent.exists()) {
            parent.mkdirs();
        }

        File file = new File(parent.toString(), this.getOutputFilename());
        if (!file.exists()) {
            try {
                file.createNewFile();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        BufferedWriter writer = null;
        try {
            writer = new BufferedWriter(new FileWriter(file));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
			e.printStackTrace();
		}
        
        
        try {
        	this.writeHeader(writer);
        	this.writeMeasurementResults(writer);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
				writer.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
        }
    }
        
    protected abstract void writeHeader(BufferedWriter writer) throws IOException;
    
    protected void writeMeasurementResults(BufferedWriter writer) throws IOException {
    	for (AbstractMeasurementResults result: this.measurementResults) {
    		result.writeResults(writer);
    	}
    }
    
    public ExecutorService getThreadsPool() {
        return this.threadsPool;
    }
    
}
