package fr.inria.gforge.dspace.overlay.benchmarks.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;

import org.ontoware.rdf2go.model.impl.StatementImpl;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.URI;
import org.openrdf.model.Statement;
import org.openrdf.rdf2go.ConversionUtil;
import org.openrdf.rio.RDFHandler;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFParseException;
import org.openrdf.rio.RDFParser;
import org.openrdf.rio.helpers.RDFParserBase;

/**
 * Parses OWL files and returns a list of {@link org.ontoware.rdf2go.model.Statement}s.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: NTriplesParser.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class NTriplesParser {

	public static String DEFAULT_ONTOLOGY = "";

	/**
	 * Parses the specified file.
	 * 
	 * @param inputStream
	 *            the stream to use to load triples.
	 * @return a list of {@link Statement}s by parsing the specified file.
	 * 
	 * @throws RDFParseException
	 * @throws RDFHandlerException
	 * @throws IOException
	 */
	public static List<org.ontoware.rdf2go.model.Statement> parse(InputStream inputStream)
			throws RDFParseException, RDFHandlerException, IOException {
		return parse(inputStream, DEFAULT_ONTOLOGY);
	}

	
	
	/**
	 * Parses the specified file with the specified ontology.
	 * 
	 * @param stream
	 *            the stream to use to load triples.
	 * @param ontology
	 *            the ontology to use.
	 * @return a list of {@link Statement}s by parsing the specified file.
	 * @throws RDFParseException
	 * @throws RDFHandlerException
	 * @throws IOException
	 */
	public static List<org.ontoware.rdf2go.model.Statement> parse(InputStream stream,
			String ontology) throws RDFParseException, RDFHandlerException,
			IOException {
		final List<org.ontoware.rdf2go.model.Statement> statements = 
			new ArrayList<org.ontoware.rdf2go.model.Statement>(10000);

		RDFParserBase parser = new org.openrdf.rio.ntriples.NTriplesParser();
		parser.setRDFHandler(new RDFHandler() {
			public void startRDF() throws RDFHandlerException {
			}

			public void handleStatement(Statement arg0)
					throws RDFHandlerException {
				statements.add(
						new StatementImpl(
								(URI) ConversionUtil.toRdf2go(arg0.getContext()), 
								(Resource) ConversionUtil.toRdf2go(arg0.getSubject()), 
								ConversionUtil.toRdf2go(arg0.getPredicate()), 
								ConversionUtil.toRdf2go(arg0.getObject())));
			}

			public void handleNamespace(String arg0, String arg1)
					throws RDFHandlerException {
			}

			public void handleComment(String arg0) throws RDFHandlerException {
			}

			public void endRDF() throws RDFHandlerException {
			}
		});

		parser.setDatatypeHandling(RDFParser.DatatypeHandling.IGNORE);
//		parser.setVerifyData(true);
//		parser.setStopAtFirstError(true);

		try {
			parser.parse(stream, ontology);
		} finally {
			stream.close();
		}

		return statements;
	}

	/**
	 * Parses all owl files from the specified directory.
	 * 
	 * @param fileDir
	 *            the directory to load owl files.
	 * @param ontology
	 *            the ontology to use.
	 * @return a list of {@link Statement}s by parsing OWL files.
	 * @throws RDFParseException
	 * @throws RDFHandlerException
	 * @throws IOException
	 */
	public static List<org.ontoware.rdf2go.model.Statement> parseDirectory(
			File fileDir, String ontology) throws RDFParseException,
			RDFHandlerException, IOException {
		FilenameFilter filterOWL = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(".owl");
			}
		};

		final List<org.ontoware.rdf2go.model.Statement> statements = 
				new ArrayList<org.ontoware.rdf2go.model.Statement>(10000);

		for (final File file : fileDir.listFiles(filterOWL)) {
			try {
				statements.addAll(parse(new FileInputStream(file)));
			} catch (RDFParseException e) {
				e.printStackTrace();
			} catch (RDFHandlerException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		return statements;
	}

	public static void write(
			List<org.ontoware.rdf2go.model.Statement> statements, File output) {
		ObjectOutputStream oos = null;
		try {
			oos = new ObjectOutputStream(new FileOutputStream(output));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		try {
			for (org.ontoware.rdf2go.model.Statement stmt : statements) {
				oos.writeObject(stmt);
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				oos.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

	}

	public static void parseAndWrite(File dirToWriteTo) {
		FilenameFilter filterOWL = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(".owl");
			}
		};

		for (File file : dirToWriteTo.listFiles(filterOWL)) {
			try {
				write(parse(new FileInputStream(file)), new File(
						dirToWriteTo, file.getName().replace(".owl", ".rdf")));
			} catch (RDFParseException e) {
				e.printStackTrace();
			} catch (RDFHandlerException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	public static void read(File dir) {
		FilenameFilter filterRDF = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(".rdf");
			}
		};

		List<org.ontoware.rdf2go.model.Statement> statements = 
			new ArrayList<org.ontoware.rdf2go.model.Statement>(10000);
		Object objectRead = null;

		for (File file : dir.listFiles(filterRDF)) {
			ObjectInputStream oos = null;
			try {
				oos = new ObjectInputStream(new FileInputStream(file));
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}

			try {
				while ((objectRead = oos.readObject()) != null) {
					statements
							.add((org.ontoware.rdf2go.model.Statement) objectRead);
				}
			} catch (IOException e) {
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} finally {
				try {
					oos.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

}