package fr.inria.gforge.dspace.overlay;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.util.ProActiveRandom;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.TriplePattern;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.api.messages.query.AskQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.ConstructQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.DescribeQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.SelectQuery;
import fr.inria.gforge.dspace.overlay.api.messages.response.AskResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.ConstructResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.DescribeResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.SelectResponse;
import fr.inria.gforge.dspace.overlay.messages.query.can.AddStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementsQueryMessage;
import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;
import fr.inria.gforge.dspace.overlay.tracker.SemanticChordTracker;
import fr.inria.gforge.dspace.overlay.util.RDF2GoBuilder;
import fr.inria.gforge.dspace.overlay.util.SemanticHelper;

/**
 * A <code>SpaceReference</code> is used in order to avoid multiple access to
 * {@link SemanticChordTracker} if multiple operations multiple operations must
 * be performed on a specified space.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SpaceReference.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class SpaceReference implements Serializable {

    private static final long serialVersionUID = 1L;

    private URI spaceURI;

    private List<SemanticChordTracker> remoteTrackers;

    private List<SemanticPeer> remotePeersCAN;

    /**
     * Constructs a new space reference for the specified space with the given
     * trackers and peers belonging to the previous space.
     * 
     * @param spaceURI
     *            the space {@link URI} associated to the space reference.
     * @param remoteTrackers
     *            the trackers used in order to retrieve peers belonging to the
     *            specified space if peers already retrieved are dead.
     * @param remotePeersCAN
     *            the {@link SemanticPeer}s associated to this space.
     */
    public SpaceReference(URI spaceURI, SemanticChordTracker[] remoteTrackers,
            SemanticPeer[] remotePeersCAN) {
        this.spaceURI = spaceURI;

        this.remoteTrackers = new ArrayList<SemanticChordTracker>();
        for (SemanticChordTracker tracker : remoteTrackers) {
            this.remoteTrackers.add(tracker);
        }

        this.remotePeersCAN = new ArrayList<SemanticPeer>();
        for (SemanticPeer peer : remotePeersCAN) {
            this.remotePeersCAN.add(peer);
        }
    }

    /**
     * Constructs a new space reference for the specified space with the given
     * trackers and peers belonging to the previous space.
     * 
     * @param spaceURI
     *            the space {@link URI} associated to the space reference.
     * @param remoteTrackers
     *            the trackers used in order to retrieve peers belonging to the
     *            specified space if peers already retrieved are dead.
     * @param remotePeersCAN
     *            the {@link SemanticPeer}s associated to this space.
     */
    public SpaceReference(URI spaceURI, List<SemanticChordTracker> remoteTrackers,
            List<SemanticPeer> remotePeersCAN) {
        this.spaceURI = spaceURI;
        this.remoteTrackers = remoteTrackers;
        this.remotePeersCAN = remotePeersCAN;
    }

    /**
     * Returns a {@link SemanticPeer} randomly chosen from the maintained list
     * of peers associated to this space reference.
     * 
     * @return a {@link SemanticPeer} randomly chosen from the maintained list
     *         of peers associated to this space reference.
     */
    public SemanticPeer getRandomSemanticPeer() {
        return this.remotePeersCAN.get(
                ProActiveRandom.nextInt(this.remotePeersCAN.size()));
    }

    /**
     * Returns all the {@link SemanticPeer} stubs of type CAN maintained by the
     * space reference.
     * 
     * @return all the {@link SemanticPeer} stubs of type CAN maintained by the
     *         space reference.
     */
    public List<SemanticPeer> getAllSemanticPeers() {
        return this.remotePeersCAN;
    }

    /*
     * Operations available on a specified space
     */

    public void addAll(Iterator<? extends Statement> it)
            throws SemanticSpaceException {
        while (it.hasNext()) {
            this.addStatement(it.next());
        }
    }

    public void addStatement(Statement statement) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI, statement)));
    }

    public void addStatement(Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI, subject, predicate, literal)));
    }

    public void addStatement(Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI,
                        subject, predicate, literal, languageTag)));
    }

    public void addStatement(Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI,
                        subject, predicate, literal, datatypeURI)));
    }

    public void addStatement(String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI, subjectURIString, predicate, literal)));
    }

    public void addStatement(String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI,
                        subjectURIString, predicate, literal, languageTag)));

    }

    public void addStatement(String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new AddStatementQueryMessage(this.spaceURI,
                        subjectURIString, predicate, literal, datatypeURI)));
    }

    public boolean contains(Statement stmt) throws SemanticSpaceException {
        return this.contains(
                    stmt.getSubject(),
                    stmt.getPredicate(),
                    stmt.getObject());
    }

    public boolean contains(ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        return this.contains(subject, predicate, object.toString());
    }

    public boolean contains(ResourceOrVariable subject, UriOrVariable predicate,
            String plainLiteral) throws SemanticSpaceException {
        StringBuffer query = new StringBuffer();
        query.append("ASK { GRAPH<");
        query.append(this.spaceURI);
        query.append("> { ");
        if (subject == null) {
            query.append("?s ");
        } else {
            query.append(subject);
            query.append(" ");
        }
        if (predicate == null) {
            query.append("?p ");
        } else {
            query.append(predicate);
            query.append(" ");
        }
        if (plainLiteral == null) {
            query.append("?o ");
        } else {
            query.append(plainLiteral);
        }
        query.append(" } . }");

        try {
            return ((AskResponse) PAFuture.getFutureValue(
                        this.getRandomSemanticPeer().send(
                                new AskQuery(this.spaceURI, query.toString())))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public ClosableIterator<Statement> findStatements(TriplePattern triplepattern)
            throws SemanticSpaceException {
        return this.sparqlConstruct(
                SemanticHelper.createConstructSparqlFrom(
                        this.spaceURI, triplepattern)).iterator();
    }

    public void removeAll() throws SemanticSpaceException {
        this.removeStatements(null, null, null);
    }

    public void removeAll(Iterator<? extends Statement> it)
            throws SemanticSpaceException {
        while (it.hasNext()) {
            this.removeStatement(it.next());
        }
    }

    public void removeStatement(Statement statement) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI, statement)));
    }

    public void removeStatement(Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI, subject, predicate, literal)));
    }

    public void removeStatement(Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI,
                        subject, predicate, literal, languageTag)));
    }

    public void removeStatement(Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI,
                        subject, predicate, literal, datatypeURI)));
    }

    public void removeStatement(String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer()
                .send(
                        new RemoveStatementQueryMessage(this.spaceURI, subjectURIString, predicate,
                                literal)));
    }

    public void removeStatement(String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI,
                        subjectURIString, predicate, literal, languageTag)));
    }

    public void removeStatement(String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementQueryMessage(this.spaceURI,
                        subjectURIString, predicate, literal, datatypeURI)));
    }

    public void removeStatements(ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer().send(
                new RemoveStatementsQueryMessage(this.spaceURI,
                        RDF2GoBuilder.createStatement(
                                (Resource) subject, (URI) predicate, (Node) object))));
    }

    public void removeStatements(TriplePattern triplePattern)
            throws SemanticSpaceException {
        this.removeStatements(triplePattern.getSubject(),
                              triplePattern.getPredicate(),
                              triplePattern.getObject());
    }

    public boolean sparqlAsk(String query) throws SemanticSpaceException {
        try {
            return ((AskResponse) PAFuture.getFutureValue(
                    this.getRandomSemanticPeer().send(
                            new AskQuery(this.spaceURI, query)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public ClosableIterable<Statement> sparqlConstruct(String query)
            throws SemanticSpaceException {
        try {
            return ((ConstructResponse) PAFuture.getFutureValue(
                    this.getRandomSemanticPeer().send(
                            new ConstructQuery(this.spaceURI, query)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }
    
    public ClosableIterable<Statement> sparqlDescribe(String query)
    throws SemanticSpaceException {
        try {
            return ((DescribeResponse) PAFuture.getFutureValue(
                    this.getRandomSemanticPeer().send(
                            new DescribeQuery(this.spaceURI, query)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public QueryResultTable sparqlSelect(String queryString)
            throws SemanticSpaceException {
        try {
            return ((SelectResponse) PAFuture.getFutureValue(
                    this.getRandomSemanticPeer().send(
                            new SelectQuery(this.spaceURI, queryString)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

}
