package fr.inria.gforge.dspace.overlay.connector;

import java.io.IOException;
import java.util.Arrays;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.node.Node;
import org.objectweb.proactive.core.node.NodeException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.NetworkNotJoinedException;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.tracker.Tracker;
import org.ontoware.rdf2go.model.node.URI;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.api.SemanticFactory;
import fr.inria.gforge.dspace.overlay.api.SemanticSpaceManagement;
import fr.inria.gforge.dspace.overlay.deployment.NodeProvider;
import fr.inria.gforge.dspace.overlay.deployment.NodeProviderKey;
import fr.inria.gforge.dspace.overlay.kernel.SemanticSpaceOverlayKernel;
import fr.inria.gforge.dspace.overlay.messages.query.chord.NewPeerOnCANQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.chord.NewPeersOnCANQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.chord.RemovePeerOnCANQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.chord.SpaceExistsQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.response.chord.SpaceExistsResponseMessage;
import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;
import fr.inria.gforge.dspace.overlay.overlay.chord.SemanticSpaceChordOverlay;
import fr.inria.gforge.dspace.overlay.tracker.SemanticChordTracker;

/**
 * ManagementConnector is used to manage a deployed SOA4ALL Space architecture.
 * It implements all operations defined in {@link SemanticSpaceManagement}.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: ManagementConnector.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class ManagementConnector implements SemanticSpaceManagement {

    private SemanticChordTracker tracker;

    private NodeProvider nodeProvider;

    private ExecutorService threadPool;
    
    private boolean autoRemoveDatastoreAtShutdown = false;
    
    private ManagementConnector() {
        this.threadPool = Executors.newCachedThreadPool();
        
        Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
            public void run() {
                threadPool.shutdown();
            }
        }));
    }

    /**
     * Constructs new ManagementConnector with the specified
     * <code>trackerURL</code> and <code>nodeProviderURL</code>.
     * 
     * @param trackerURL
     *            the URL to use in order to lookup a new
     *            {@link SemanticChordTracker}.
     * 
     * @param nodeProviderURL
     *            the URL to use in order to lookup a new {@link NodeProvider}.
     */
    public ManagementConnector(String trackerURL, String nodeProviderURL) {
        this();
        try {
            this.tracker = PAActiveObject.lookupActive(SemanticChordTracker.class, trackerURL);
            this.nodeProvider = PAActiveObject.lookupActive(NodeProvider.class, nodeProviderURL);
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Constructs a new ManagementConnector with the specified
     * <code>tracker</code> and <code>nodeProvider</code>.
     * 
     * @param tracker
     *            the {@link SemanticChordTracker} to use.
     * 
     * @param nodeProvider
     *            the {@link NodeProvider} to use.
     */
    public ManagementConnector(SemanticChordTracker tracker, NodeProvider nodeProvider) {
        this();
        this.tracker = tracker;
        this.nodeProvider = nodeProvider;
    }

    /**
     * Constructs a new ManagementConnector with the specified
     * <code>tracker</code>.
     * 
     * @param tracker
     *            the tracker to use.
     */
    public ManagementConnector(SemanticChordTracker tracker) {
        this();
        this.tracker = tracker;
    }

    /**
     * Creates a new {@link Peer} of type Chord and adds it on the Chord network
     * associated to the current tracker. The {@link Peer} created is deployed
     * on the JVM of the current object if the {@link NodeProvider} of the
     * current object is <code>null</code> . Otherwise, the new {@link Peer} is
     * deployed by using a {@link Node} provided by the {@link NodeProvider}.
     */
    public void newPeerOnChord() {
        SemanticPeer peer = null;
        if (this.nodeProvider != null) {
            try {
                peer = SemanticFactory.newActiveSemanticPeer(
                            Arrays.asList((Tracker) tracker),
                            new SemanticSpaceChordOverlay(),
                            this.nodeProvider.getNextNode(NodeProviderKey.PEERS_CHORD));
            } catch (ActiveObjectCreationException e) {
                e.printStackTrace();
            } catch (NodeException e) {
                e.printStackTrace();
            }
        } else {
            try {
                peer = SemanticFactory.newActiveSemanticPeer(
                            Arrays.asList((Tracker) tracker),
                            new SemanticSpaceChordOverlay());
            } catch (ActiveObjectCreationException e) {
                e.printStackTrace();
            } catch (NodeException e) {
                e.printStackTrace();
            }
        }

        this.tracker.addOnNetwork(peer);
    }

    /**
     * Creates and adds the specified number of peers of type Chord on the Chord
     * network.
     * 
     * The {@link Peer}s created are deployed on the JVM of the current object
     * if the {@link NodeProvider} of the current object is <code>null</code> .
     * Otherwise, the new {@link Peer}s are deployed by using {@link Node}s
     * provided by the {@link NodeProvider}. For each new {@link Peer}, a new
     * {@link SemanticSpaceOverlayKernel} is created.
     * 
     * @param nb
     *            the number of {@link Peer} of type Chord to add.
     */
    public void newPeersOnChordInParallel(int nb) {
        final SemanticPeer[] peers;
        final CountDownLatch doneSignal = new CountDownLatch(nb);
        
        if (nodeProvider == null) {
            peers = SemanticFactory.newActiveSemanticChordPeersInParallel(
                                            nb, Arrays.asList((Tracker)this.tracker));
        } else {
            peers = SemanticFactory.newActiveSemanticChordPeersInParallel(
                    nb, Arrays.asList(
                            (Tracker)this.tracker), 
                            this.nodeProvider.getNextNodes(
                                    NodeProviderKey.PEERS_CHORD, nb));
        }
        
        for (int i = 0; i < nb; i++) {
            final int index=  i;
            this.threadPool.execute(new Runnable() {
                public void run() {
                    tracker.addOnNetwork(peers[index]);
                    doneSignal.countDown();
                }
            });
        }
        
        try {
            doneSignal.await();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    /**
     * {@inheritDoc}
     * 
     * If the space already exists, a new peer of type CAN is added.
     * 
     * @see #newPeerOnCAN(URI)
     */
    public void createSpace(URI spaceURI) throws SemanticSpaceException {
        this.newPeerOnCAN(spaceURI);
    }

    /**
     * {@inheritDoc}
     * 
     * If the space already exists, a new peer of type CAN is added.
     * 
     * @see #newPeerOnCAN(URI, SemanticSpaceOverlayKernel)
     */
    public void createSpace(URI spaceURI, SemanticSpaceOverlayKernel remoteKernel)
            throws SemanticSpaceException {
        this.newPeerOnCAN(spaceURI, remoteKernel);
    }
    
    /**
     * Creates a new {@link Peer} of type CAN and force it to join an existing
     * peer from the network associated to the specified <code>spaceURI</code>.
     * If the space specified doesn't exist it is automatically created. The
     * {@link Peer} created is deployed on the JVM of the current object if the
     * {@link NodeProvider} of the current object is <code>null</code> .
     * Otherwise, the new {@link Peer} is deployed by using a {@link Node}
     * provided by the {@link NodeProvider}.
     * 
     * @param spaceURI
     *            the space on which the peer is added.
     * 
     * @throws SemanticSpaceException
     *             if the Chord network doesn't exist.
     */
    public void newPeerOnCAN(URI spaceURI) throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0) {
            try {
                PAFuture.waitFor(this.tracker.getRandomPeer().send(
                        new NewPeerOnCANQueryMessage(
                                spaceURI,
                                SemanticFactory.newActiveSemanticSpaceOverlayKernel(
                                        new Tracker[] { this.tracker }, 
                                        this.nodeProvider, 
                                        this.autoRemoveDatastoreAtShutdown))));
            } catch (ActiveObjectCreationException e) {
                e.printStackTrace();
            } catch (NodeException e) {
                e.printStackTrace();
            }
        } else {
            throw new SemanticSpaceException("Chord network not initialized.");
        }
    }

    /**
     * Creates <code>nb</code> {@link Peer}s of type CAN in parallel and add
     * them on the specified space. If the space does not exist, it is created.
     * 
     * The {@link Peer}s created are deployed on the JVM of the current object
     * if the {@link NodeProvider} of the current object is <code>null</code> .
     * Otherwise, the new {@link Peer}s are deployed by using {@link Node}s
     * provided by the {@link NodeProvider}. For each new {@link Peer}, a new
     * {@link SemanticSpaceOverlayKernel} is created.
     * 
     * @param spaceURI
     *            identify the space on which the new peers join.
     * @param nb
     *            the number of peers to add.
     * @throws SemanticSpaceException
     *             if the Chord network doesn't exist or if a
     *             {@link NodeProvider} is set and it cannot provide
     *             <code>nb</code> {@link Node}s for {@link Peer}s of type CAN.
     */
    public void newPeersOnCANInParallel(URI spaceURI, int nb) throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0) {
            SemanticPeer[] peers;
            
            if (this.nodeProvider != null) {
                Node[] nodes = new Node[nb];
                for (int i=0; i<nb; i++) {
                    nodes[i] = this.nodeProvider.getNextNode(NodeProviderKey.PEERS_CAN);
                }
                
                Map<Node, SemanticSpaceOverlayKernel> kernels = 
                    SemanticFactory.newActiveSemanticSpaceOverlayKernelsInParallel(
                            new Tracker[] { this.tracker }, 
                            this.autoRemoveDatastoreAtShutdown, 
                            nodes);
                
                peers = SemanticFactory.newActiveSemanticCANPeersInParallel(
                            spaceURI, Arrays.asList((Tracker)this.tracker), kernels);
            } else {
                SemanticSpaceOverlayKernel[] kernels = 
                    SemanticFactory.newActiveSemanticSpaceOverlayKernelsInParallel(
                            nb, 
                            new Tracker[] { this.tracker }, 
                            this.autoRemoveDatastoreAtShutdown);
                
                URI[] spaceURIs = new URI[nb];
                for (int i=0; i<nb; i++) {
                    spaceURIs[i] = spaceURI;
                }
                                          
                peers = SemanticFactory.newActiveSemanticCANPeersInParallel(
                        spaceURIs, Arrays.asList((Tracker)this.tracker), kernels);
            }
            
            PAFuture.waitFor(this.tracker.getRandomPeer().send(
                            new NewPeersOnCANQueryMessage(spaceURI, peers)));
        } else {
            throw new SemanticSpaceException("Chord network isn't initialized.");
        }
    }
    
    /**
     * Creates a new {@link Peer} of type CAN and force it to join an existing
     * peer from the network associated to the specified <code>spaceURI</code>.
     * If the space specified doesn't exist it is automatically created. The
     * {@link Peer} created is deployed on the JVM of the specified
     * <code>kernel</code>.
     * 
     * @param spaceURI
     *            the space on which the peer is added.
     * 
     * @param remoteKernel
     *            the kernel associated to the new peer created. The peer
     *            created is deployed by using the same {@link Node} as the
     *            kernel specified.
     * 
     * @throws SemanticSpaceException
     *             if the Chord network doesn't exist.
     */
    public void newPeerOnCAN(URI spaceURI, SemanticSpaceOverlayKernel remoteKernel)
            throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0) {
            PAFuture.waitFor(this.tracker.getRandomPeer().send(
                    new NewPeerOnCANQueryMessage(
                            spaceURI,
                            (SemanticSpaceOverlayKernel) PAActiveObject.getStubOnThis())));
        } else {
            throw new SemanticSpaceException();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void removePeerOnCAN(URI spaceURI) throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0 ||
            !this.spaceExists(spaceURI)) {
            throw new SemanticSpaceException("The specified space does not exist.");
        }

        PAFuture.waitFor(this.tracker.getRandomPeer().send(
                new RemovePeerOnCANQueryMessage(spaceURI)));
    }

    /**
     * {@inheritDoc}
     */
    public void removePeerOnChord() throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0) {
            try {
                this.tracker.getRandomPeer().leave();
            } catch (NetworkNotJoinedException e) {
                e.printStackTrace();
            }
        } else {
            throw new SemanticSpaceException("Chord network not initialized.");
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean spaceExists(URI spaceURI) throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() > 0) {
            return ((SpaceExistsResponseMessage)PAFuture.getFutureValue(
                    this.tracker.getRandomPeer().send(
                        new SpaceExistsQueryMessage(spaceURI)))).getValue();
        } else {
            throw new SemanticSpaceException("Chord network not initialized.");
        }
    }

    /**
     * Returns a boolean indicating if the datastore created when a new
     * {@link SemanticSpaceOverlayKernel} is instantiated must be removed when
     * they are shutdown or not.
     * 
     * @return a boolean indicating if the datastore created when a new
     *         {@link SemanticSpaceOverlayKernel} is instantiated must be removed
     *         when they are shutdown or not.
     */
    public boolean isAutoRemoveDatastoreAtShutdown() {
        return this.autoRemoveDatastoreAtShutdown;
    }

    /**
     * Sets a boolean indicating if the datastore created when a new
     * {@link SemanticSpaceOverlayKernel} is instantiated must be removed when
     * they are shutdown or not.
     * 
     * @param autoRemoveDatastoreAtShutdown
     *            the value to set.
     */
    public void setAutoRemoveDatastoreAtShutdown(boolean autoRemoveDatastoreAtShutdown) {
        this.autoRemoveDatastoreAtShutdown = autoRemoveDatastoreAtShutdown;
    }

}
