package fr.inria.gforge.dspace.overlay.deployment;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.core.ProActiveException;
import org.objectweb.proactive.core.node.Node;
import org.objectweb.proactive.extensions.gcmdeployment.PAGCMDeployment;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.tracker.Tracker;
import org.objectweb.proactive.gcmdeployment.GCMApplication;
import org.objectweb.proactive.gcmdeployment.GCMVirtualNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * NodeProvider is an active object which is in charge for providing
 * {@link Node}s for peers, trackers and kernels.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: NodeProvider.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class NodeProvider implements Serializable {

    private static final long serialVersionUID = 1L;

    private static final Logger logger = LoggerFactory.getLogger(NodeProvider.class);
    
    private final static int NODES_ACQUISITION_TIMEOUT = 2000;

    private File pathToGCMADescriptor;

    private Map<NodeProviderKey, NodeProviderEntry> entries = 
		new HashMap<NodeProviderKey, NodeProviderEntry>();
    
    private String bindingName;

    private GCMApplication gcmad;

    public NodeProvider() {

    }

    public NodeProvider(File pathToGCMADescriptor) {
        this.pathToGCMADescriptor = pathToGCMADescriptor;
    }

    /**
     * Start deployment by acquiring {@link Node}s.
     * 
     * @param maxTrackerNodesToAcquire
     *            the maximum number of nodes for {@link Tracker}s the
     *            {@link NodeProvider} will acquire. However, the GCMA file must
     *            contain a sufficient number of machines to use in order to
     *            acquire the specified number of nodes. This parameter allows
     *            to acquire less node than specified in the GCMA files. To
     *            acquire all possible nodes as specified in the GCMA file you
     *            can use the value <code>-1</code>.
     * @param maxChordNodesToAcquire
     *            the maximum number of nodes for Chord {@link Peer}s the
     *            {@link NodeProvider} will acquire. However, the GCMA file must
     *            contain a sufficient number of machines to use in order to
     *            acquire the specified number of nodes. This parameter allows
     *            to acquire less node than specified in the GCMA files. To
     *            acquire all possible nodes as specified in the GCMA file you
     *            can use the value <code>-1</code>.
     * @param maxCANNodesToAcquire
     *            the maximum number of nodes for CAN {@link Peer}s the
     *            {@link NodeProvider} will acquire. However, the GCMA file must
     *            contain a sufficient number of machines to use in order to
     *            acquire the specified number of nodes. This parameter allows
     *            to acquire less node than specified in the GCMA files. To
     *            acquire all possible nodes as specified in the GCMA file you
     *            can use the value <code>-1</code>.
     */
    public void deploy(int maxTrackerNodesToAcquire, int maxChordNodesToAcquire, int maxCANNodesToAcquire) {
        try {
            this.gcmad = PAGCMDeployment.loadApplicationDescriptor(
                                            this.pathToGCMADescriptor);
        } catch (ProActiveException e) {
            e.printStackTrace();
        }
        this.gcmad.startDeployment();
        this.gcmad.waitReady();

        this.entries.put(NodeProviderKey.NODE_PROVIDER, 
                new NodeProviderEntry(
                        Arrays.asList(
                                this.gcmad.getVirtualNode("nodeProviderVN").getANode())));

        this.entries.put(NodeProviderKey.TRACKERS, 
                new NodeProviderEntry(
                        maxTrackerNodesToAcquire == -1
                        ? this.acquireAllNodes("trackersVN")
                        : this.acquireNodes("trackersVN", maxTrackerNodesToAcquire)));
        
        this.entries.put(NodeProviderKey.PEERS_CHORD, 
                new NodeProviderEntry(
                        maxChordNodesToAcquire == -1
                        ? this.acquireAllNodes("chordPeersVN")
                                : this.acquireNodes("chordPeersVN", maxChordNodesToAcquire)));

        this.entries.put(NodeProviderKey.PEERS_CAN, 
                new NodeProviderEntry(
                        maxCANNodesToAcquire == -1
                        ? this.acquireAllNodes("canPeersVN")
                        : this.acquireNodes("canPeersVN", maxCANNodesToAcquire)));
        

        if (logger.isDebugEnabled()) {
            StringBuffer buf = new StringBuffer();
            buf.append("Deployment started...");
            buf.append("\n* ");
            buf.append(this.entries.get(NodeProviderKey.PEERS_CHORD).getNodes().size());
            buf.append(" nodes available for peers of type Chord\n");
            buf.append(this.dump(this.entries.get(NodeProviderKey.PEERS_CHORD).getNodes()));
            buf.append("* ");
            buf.append(this.entries.get(NodeProviderKey.PEERS_CAN).getNodes().size());
            buf.append(" nodes available for peers of type CAN\n");
            buf.append(this.dump(this.entries.get(NodeProviderKey.PEERS_CAN).getNodes()));
            buf.append("* ");
            buf.append(this.entries.get(NodeProviderKey.TRACKERS).getNodes().size());
            buf.append(" node available for trackers\n");
            buf.append(this.dump(this.entries.get(NodeProviderKey.TRACKERS).getNodes()));
            logger.debug(buf.toString());
        }
    }

    private String dump(List<Node> nodes) {
        StringBuffer buf = new StringBuffer();
        for (Node node : nodes) {
            buf.append("- ");
            buf.append(node.getNodeInformation().getURL());
            buf.append("\n");
        }
        return buf.toString();
    }

    private List<Node> acquireNodes(String virtualNodeId, int nb) {
        GCMVirtualNode virtualNode = this.gcmad.getVirtualNode(virtualNodeId);

        List<Node> nodes = new ArrayList<Node>();
        Node node;

        for (int i = 0; i < nb; i++) {
            node = virtualNode.getANode(NODES_ACQUISITION_TIMEOUT);
            if (node == null) {
                throw new IllegalStateException(
                        "Cannot acquire " + nb + " nodes because only " + i + " are available.");
            } else {
                nodes.add(node);
            }
        }

        return nodes;
    }

    private List<Node> acquireAllNodes(String virtualNodeId) {
        boolean waitingForNodesAcquisition = true;
        GCMVirtualNode virtualNode = this.gcmad.getVirtualNode(virtualNodeId);

        List<Node> nodes = virtualNode.getNewNodes();

        while (waitingForNodesAcquisition) {
            Thread.yield();

            List<Node> newNodes = virtualNode.getNewNodes();
            if (newNodes.size() > 0) {
                nodes.addAll(newNodes);
            } else {
                waitingForNodesAcquisition = false;
            }

            if (logger.isInfoEnabled()) {
                logger.info("Nodes acquisition for VN '{}', size is {}.",
                        virtualNodeId, nodes.size());
            }
        }

        return nodes;
    }

    public List<Node> getAllNodes(NodeProviderKey key) {
        return this.entries.get(key).getNodes();
    }

    public Node getNextNode(NodeProviderKey key) {
        return this.entries.get(key).getNextNode();
    }
    
    public Node[] getNextNodes(NodeProviderKey key, int nb) {
        List<Node> nodes = new ArrayList<Node>(nb);
        
        for (int i=0; i<nb; i++) {
            nodes.add(this.getNextNode(key));
        }
        
        return nodes.toArray(new Node[]{});
    }

    public String register() {
        if (this.bindingName == null) {
            try {
                this.bindingName = 
                    PAActiveObject.registerByName(
                	    PAActiveObject.getStubOnThis(),
                            "node-provider");
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        return this.bindingName;
    }

}
