package fr.inria.gforge.dspace.overlay.initializers;

import java.io.File;
import java.util.UUID;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.core.node.Node;
import org.objectweb.proactive.core.node.NodeException;

import fr.inria.gforge.dspace.overlay.api.SemanticFactory;
import fr.inria.gforge.dspace.overlay.connector.ManagementConnector;
import fr.inria.gforge.dspace.overlay.connector.OperationsConnector;
import fr.inria.gforge.dspace.overlay.deployment.NodeProvider;
import fr.inria.gforge.dspace.overlay.deployment.NodeProviderKey;
import fr.inria.gforge.dspace.overlay.tracker.SemanticChordTracker;

/**
 * Initializes a global network on a local machine or by distributing the active
 * objects on several machines. A global network is represented by a two levels
 * architecture: at the upper level a Chord network where each node contains
 * references to several spaces and at the lower level, one or several CAN
 * network for each space.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: GlobalNetworkInitializer.java 4614 2010-05-20 08:19:05Z
 *          laurentp $
 */
public class GlobalNetworkInitializer extends SemanticNetworkInitializer<SemanticChordTracker> {

    private NodeProvider nodeProvider;

    public GlobalNetworkInitializer() {
        
    }

    /**
     * Creates a new network By default no space nor chord peers are created and
     * the peers will be deployed on the current machine.
     */
    public void setUpNetworkOnLocalMachine() {
        if (super.initialized) {
            throw new IllegalStateException("Network already initialized");
        }

        try {
            super.trackers = new SemanticChordTracker[] {
                    SemanticFactory.newActiveSemanticChordTracker(
                            UUID.randomUUID().toString()) };
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (NodeException e) {
            e.printStackTrace();
        }

        super.initialized = true;
    }

    /***
     * Creates a new network. By default no space nor Chord peers are created
     * and the peers will be deployed using the information found in the file
     * located in <code>pathToGCMA</code>.
     * 
     * @param pathToGCMA
     *            the Grid Component Model Application file to use in order to
     *            know what are the machine to use for objects deployment.
     * 
     * @param maxChordNodesToAcquire
     *            the maximum number of Chord nodes the {@link NodeProvider}
     *            will acquire (Peers are deployed using {@link Node}). However,
     *            the GCMA file must contain a sufficient number of machines to
     *            use in order to acquire the specified number of nodes. This
     *            parameter allows to acquire less node than specified in the
     *            GCMA files. To acquire all possible nodes as specified in the
     *            GCMA file you can use the value <code>-1</code>.
     * 
     * @param maxCANNodesToAcquire
     *            the maximum number of CAN nodes the {@link NodeProvider} will
     *            acquire (Peers are deployed using {@link Node}). However, the
     *            GCMA file must contain a sufficient number of machines to use
     *            in order to acquire the specified number of nodes. This
     *            parameter allows to acquire less node than specified in the
     *            GCMA files. To acquire all possible nodes as specified in the
     *            GCMA file you can use the value <code>-1</code>.
     */
    public void setUpNetworkOnMultipleMachine(File pathToGCMA,
                                              int maxChordNodesToAcquire,
                                              int maxCANNodesToAcquire) {
        if (super.initialized) {
            throw new IllegalStateException("Network already initialized");
        }

        NodeProvider nodeProvider = new NodeProvider(pathToGCMA);
        nodeProvider.deploy(1, maxChordNodesToAcquire, maxCANNodesToAcquire);
        try {
            this.nodeProvider = PAActiveObject.turnActive(
                    nodeProvider, nodeProvider.getNextNode(NodeProviderKey.NODE_PROVIDER));
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (NodeException e) {
            e.printStackTrace();
        }
        this.nodeProvider.register();

        try {
            super.trackers = new SemanticChordTracker[] {
                    SemanticFactory.newActiveSemanticChordTracker(
                            UUID.randomUUID().toString(), 
                            UUID.randomUUID().toString(), 
                            nodeProvider.getNextNode(NodeProviderKey.TRACKERS)) };
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (NodeException e) {
            e.printStackTrace();
        }

        super.initialized = true;
    }
    
    /**
     * Returns the {@link NodeProvider} which has been initialized when the
     * network is setup on several machines.
     * 
     * @return the {@link NodeProvider} which has been initialized when the
     *         network is setup on several machines or <code>null</code>.
     */
    public NodeProvider getNodeProvider() {
        return this.nodeProvider;
    }
    
    
    /**
     * Returns a {@link ManagementConnector} in order to manage the network
     * initialized.
     * 
     * @return a {@link ManagementConnector} in order to manage the network
     *         initialized.
     */
    public ManagementConnector getManagementConnector() {
        if (this.nodeProvider != null) {
            return new ManagementConnector(super.getRandomTracker(), this.nodeProvider);
        } else {
            return new ManagementConnector(super.getRandomTracker());
        }
    }

    /**
     * Returns an {@link OperationsConnector} in order to perform operations on
     * the current network.
     * 
     * @return an {@link OperationsConnector} in order to perform operations on
     *         the current network.
     */
    public OperationsConnector getOperationsConnector() {
        return new OperationsConnector(super.getRandomTracker());
    }
    
    /**
     * {@inheritDoc}
     */
    public void tearDownNetwork() {
        if (this.nodeProvider != null) {
            PAActiveObject.terminateActiveObject(this.nodeProvider, true);
            this.nodeProvider = null;
        }
    }

}
