package fr.inria.gforge.dspace.overlay.initializers;

import java.io.File;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.node.Node;
import org.objectweb.proactive.core.node.NodeException;
import org.objectweb.proactive.extensions.structuredp2p.api.TrackerFactory;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.StructuredP2PException;
import org.objectweb.proactive.extensions.structuredp2p.overlay.OverlayType;
import org.objectweb.proactive.extensions.structuredp2p.tracker.Tracker;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.node.URI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.api.SemanticFactory;
import fr.inria.gforge.dspace.overlay.api.messages.query.AskQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.ConstructQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.SelectQuery;
import fr.inria.gforge.dspace.overlay.api.messages.response.AskResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.ConstructResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.SelectResponse;
import fr.inria.gforge.dspace.overlay.deployment.NodeProvider;
import fr.inria.gforge.dspace.overlay.deployment.NodeProviderKey;
import fr.inria.gforge.dspace.overlay.kernel.SemanticSpaceOverlayKernel;
import fr.inria.gforge.dspace.overlay.messages.query.can.AddStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementsQueryMessage;
import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;
import fr.inria.gforge.dspace.overlay.util.RDF2GoBuilder;
import fr.inria.gforge.dspace.overlay.util.SemanticHelper;

/**
 * Initializes a space network on a local machine or by distributing the active
 * objects on several machines. A space network is represented by a CAN network.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SpaceNetworkInitializer.java 4597 2010-05-17 08:38:32Z laurentp
 *          $
 */
public class SpaceNetworkInitializer extends SemanticNetworkInitializer<Tracker> {

    private final static Logger logger = LoggerFactory.getLogger(SpaceNetworkInitializer.class);

    private ExecutorService threadsPool = Executors.newCachedThreadPool();
    
    private URI spaceURI;

    private boolean isLocalSetup = true;
    
    public SpaceNetworkInitializer(URI spaceURI) {
        this.spaceURI = spaceURI;
    }

    /**
     * Setup a new space network composed of the specified number of peers of type CAN.
     * 
     * @param nbPeersCAN
     *            the number of peers to add on the network.
     */
    public void setUpNetworkOnLocalMachine(int nbPeersCAN) {
        if (super.initialized) {
            throw new IllegalStateException("Network already initialized");
        }

        try {
            Tracker tracker = TrackerFactory.newActiveTracker(OverlayType.CAN);
            super.trackers = new Tracker[] { tracker };
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (NodeException e) {
            e.printStackTrace();
        }
        
        SemanticSpaceOverlayKernel[] kernels = 
            SemanticFactory.newActiveSemanticSpaceOverlayKernelsInParallel(
                    nbPeersCAN, super.trackers, true);

         
        URI[] spaceURIs = new URI[nbPeersCAN];
        for (int i = 0; i < spaceURIs.length; i++) {
           spaceURIs[i] = this.spaceURI;
       }
 
        SemanticPeer[] peers = SemanticFactory.newActiveSemanticCANPeersInParallel(
                spaceURIs, Arrays.asList(super.trackers), kernels);
        
        this.addOnNetworkInParallel(peers);

        super.initialized = true;
    }

    /**
     * Setup a new space network using the specified GCMA descriptor.
     * 
     * @param pathToGCMA
     *            the path to the descriptor to use.
     * 
     * @param maxCANNodesToAcquire
     *            the maximum number of nodes for peers of type CAN to acquire.
     */
    public void setUpNetworkOnMultipleMachine(File pathToGCMA, int maxCANNodesToAcquire) {
        if (super.initialized) {
            throw new IllegalStateException("Network already initialized.");
        }
        this.isLocalSetup = false;

        NodeProvider nodeProvider = new NodeProvider(pathToGCMA);
        nodeProvider.deploy(1, 0, maxCANNodesToAcquire);

        try {
            super.trackers = new Tracker[] { TrackerFactory.newActiveTracker(
                    OverlayType.CAN, nodeProvider.getNextNode(NodeProviderKey.TRACKERS)) };
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (NodeException e) {
            e.printStackTrace();
        }
        
        // Kernels use the same nodes as peers because
        // they must be on the same JVM.
        Node[] nodesCAN = nodeProvider.getAllNodes(
                NodeProviderKey.PEERS_CAN).toArray(
                        new Node[] {});

        Map<Node, SemanticSpaceOverlayKernel> deployedKernels =
                   SemanticFactory.newActiveSemanticSpaceOverlayKernelsInParallel(
                           this.trackers, true, nodesCAN);
        SemanticSpaceOverlayKernel[] kernels = 
        	new SemanticSpaceOverlayKernel[deployedKernels.values().size()];

        int i = 0;
        for (SemanticSpaceOverlayKernel kernel : deployedKernels.values()) {
            kernels[i] = kernel;
            i++;
        }

        SemanticPeer[] peers = SemanticFactory.newActiveSemanticCANPeersInParallel(
                                   this.spaceURI, Arrays.asList(super.trackers),
                                   deployedKernels);
       this.addOnNetworkInParallel(peers);

        super.initialized = true;
    }

    private void addOnNetworkInParallel(final SemanticPeer[] peers) {
        final CountDownLatch doneSignal = new CountDownLatch(peers.length);
        
        for (int i=0; i<peers.length; i++) {
            final int index = i;
            this.threadsPool.execute(new Runnable() {
                public void run() {
                    try {
                        getRandomTracker().addOnNetwork(peers[index]);
                    } catch (StructuredP2PException e) {
                        e.printStackTrace();
                    } finally {
                        doneSignal.countDown();
                    }
                }
            });
        }
        
        try {
            doneSignal.await();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }
    
    public void addStatement(String subject, String predicate, String object) {
        try {
            this.addStatement(RDF2GoBuilder.toStatement(subject, predicate, object));
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }
    }

    public void addStatement(Statement stmt) {
        PAFuture.waitFor(super.getRandomTracker().getRandomPeer().send(
                            new AddStatementQueryMessage(
                                    this.spaceURI, stmt)));

        if (logger.isInfoEnabled()) {
            StringBuffer buf = new StringBuffer();
            buf.append("Add operation has been performed ");
            buf.append("for triple <");
            buf.append(stmt.getSubject());
            buf.append(",");
            buf.append(stmt.getPredicate());
            buf.append(",");
            buf.append(stmt.getObject());
            buf.append(">");
            buf.append(".");
            logger.info(buf.toString());
        }
    }

    public void addRandomStatements(int nb) {
        for (int i = 0; i < nb; i++) {
            this.threadsPool.execute(new Runnable() {
                public void run() {
                    addStatement(SemanticHelper.generateRandomStatement(spaceURI));
                }
            });
        }
    }

    public boolean isLocalSetup() {
        return isLocalSetup;
    }

    public void removeStatement(String subject, String predicate, String object) {
        try {
            super.getRandomTracker().getRandomPeer().send(
                    new RemoveStatementQueryMessage(
                            this.spaceURI, 
                            RDF2GoBuilder.toStatement(
                                    subject, predicate, object)));
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }
    }

    public void removeStatements(String subject, String predicate, String object) {
        try {
            super.getRandomTracker().getRandomPeer().send(
                    new RemoveStatementsQueryMessage(
                            this.spaceURI, 
                            RDF2GoBuilder.createStatement(
                                    subject == null ? null : RDF2GoBuilder.toSubject(subject), 
                                    predicate == null ? null : RDF2GoBuilder.toPredicate(predicate), 
                                    object == null ? null : RDF2GoBuilder.toObject(object))));
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }
    }

    public boolean sparqlAsk(String askQuery) {
        Boolean response = false;
        try {
            response = ((AskResponse) PAFuture.getFutureValue(
                            super.getRandomTracker().getRandomPeer().send(
                                    new AskQuery(this.spaceURI, askQuery)))).getResult();
        } catch (Exception e) {
            e.printStackTrace();
        }

        return response;
    }

    public Set<Statement> sparqlConstruct(String constructQuery) {
        ConstructResponse response = null;
        Set<Statement> result = new HashSet<Statement>();

        try {
            response = (ConstructResponse) PAFuture.getFutureValue(
                            super.getRandomTracker().getRandomPeer().send(
                                    new ConstructQuery(this.spaceURI, constructQuery)));
            
            ClosableIterator<Statement> it = response.getResult().iterator();
            while (it.hasNext()) {
                result.add(it.next());
            }

        } catch (DispatchException e) {
            e.printStackTrace();
        }
        
        return result;
    }

    public QueryResultTable sparqlSelect(String selectQuery) {
        QueryResultTable result = null;
        
        try {
            result = ((SelectResponse) PAFuture.getFutureValue(
                        super.getRandomTracker().getRandomPeer().send(
                                new SelectQuery(this.spaceURI, selectQuery)))).getResult();
        } catch (DispatchException e) {
            e.printStackTrace();
        }

        return result;
    }

    public void tearDownNetwork() {
        this.threadsPool.shutdown();
    }

    public URI getSpaceURI() {
        return this.spaceURI;
    }
    
}
