package fr.inria.gforge.dspace.overlay.messages.query.can;

import java.util.List;

import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.AbstractCANOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.Zone;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.inria.gforge.dspace.overlay.api.messages.query.SPARQLQuery;
import fr.inria.gforge.dspace.overlay.messages.response.can.AnycastResponseMessage;
import fr.inria.gforge.dspace.overlay.reasoner.RangeQueryCondition;
import fr.inria.gforge.dspace.overlay.reasoner.RangeQueryOperator;
import fr.inria.gforge.dspace.overlay.router.can.AnycastQueryRouter;
import fr.inria.gforge.dspace.overlay.validator.AnycastConstraintsValidator;

/**
 * @author Laurent Pellegrino
 * 
 * @version $Id: RangeQueryMessage.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class RangeQueryMessage extends SemanticQueryMessage {

    private final static transient Logger logger = LoggerFactory.getLogger(RangeQueryMessage.class);

    private static final long serialVersionUID = 1L;

    private List<RangeQueryCondition> conditions;

    public RangeQueryMessage(SPARQLQuery query, final List<RangeQueryCondition> conditions) {
        super(query, new CoordinateValue[] { null, null, null });
        this.conditions = conditions;

        /*
         * Extract information in order to set keyToReach for routing on peers
         * which don't validate constraints
         */
        CoordinateValue[] keyToReach = new CoordinateValue[] { null, null, null };

        for (RangeQueryCondition condition : conditions) {
            if (condition.getOperator() == RangeQueryOperator.Type.GREATER
                    || condition.getOperator() == RangeQueryOperator.Type.GREATER_EQUALS) {
                keyToReach[condition.getIndex()] = condition.getCoordinate();
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Set keyToReach to (" + keyToReach[0] + ", " + keyToReach[1] + ", "
                    + keyToReach[2] + ")");
        }
        super.setKeyToReach(keyToReach);
        logger.debug("New RangeQueryMessage created");
    }

    /**
     * @return the conditions
     */
    public List<RangeQueryCondition> getConditions() {
        return this.conditions;
    }

    public AnycastQueryRouter<RangeQueryMessage> getRouter() {
        ConstraintsValidator<CoordinateValue[]> validator =  new AnycastConstraintsValidator<CoordinateValue[]>() {
            public boolean validatesKeyConstraints(StructuredOverlay overlay, CoordinateValue[] key) {
                return this.validatesKeyConstraints(((AbstractCANOverlay) overlay).getZone(), key);
            }

            public boolean validatesKeyConstraints(Zone zone, CoordinateValue[] key) {
                for (RangeQueryCondition condition : conditions) {
                    if (condition.getOperator() == RangeQueryOperator.Type.LESS
                            && zone.contains(condition.getIndex(), condition.getCoordinate()) > 0) {
                        return false;
                    }
                    if (condition.getOperator() == RangeQueryOperator.Type.LESS_EQUALS
                            && zone.contains(condition.getIndex(), condition.getCoordinate()) > 0) {
                        return false;
                    }
                    if (condition.getOperator() == RangeQueryOperator.Type.GREATER
                            && zone.contains(condition.getIndex(), condition.getCoordinate()) < 0) {
                        return false;
                    }
                    if (condition.getOperator() == RangeQueryOperator.Type.GREATER_EQUALS
                            && zone.contains(condition.getIndex(), condition.getCoordinate()) <= 0) {
                        return false;
                    }
                }
                return true;
            }
        };
        
        return new AnycastQueryRouter<RangeQueryMessage>(validator) {
            public void onPeerWhichValidatesKeyConstraints(AbstractCANOverlay overlay,
                    AnycastQueryMessage msg) {

            }
        };
    }

    public AnycastResponseMessage<?> createResponseMessage() {
        return super.getSparqlQuery().createResponseMessage(this);
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();

        for (int i = 0; i < this.conditions.size(); i++) {
            buf.append(this.conditions.get(i));
            if (i < this.conditions.size() - 1) {
                buf.append(" && ");
            }
        }

        return buf.toString();
    }

}
