package fr.inria.gforge.dspace.overlay.overlay;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.objectweb.proactive.extensions.structuredp2p.api.messages.Query;
import org.objectweb.proactive.extensions.structuredp2p.api.messages.Response;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.PostProcessException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.PreProcessException;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;
import org.objectweb.proactive.extensions.structuredp2p.messages.response.AbstractResponseMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.QueryManager;
import org.openrdf.query.MalformedQueryException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceQueryException;
import fr.inria.gforge.dspace.overlay.api.messages.query.SPARQLQuery;
import fr.inria.gforge.dspace.overlay.messages.query.can.AnycastQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.SemanticQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.AnycastResponseMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.SemanticResponseMessage;
import fr.inria.gforge.dspace.overlay.reasoner.ParsedSPARQLQuery;
import fr.inria.gforge.dspace.overlay.reasoner.SPARQLQueryIdentifier;
import fr.inria.gforge.dspace.overlay.reasoner.SPARQLQueryReasoner;
import fr.inria.gforge.dspace.overlay.router.can.AnycastResponseRouter;
import fr.inria.gforge.dspace.overlay.util.SPARQLQueryFilter;

/**
 * {@link SemanticQueryManager} is an implementation of {@link QueryManager}. It
 * is used for converting initial query from public API to private API,
 * dispatching query and to create response from public API by using responses
 * retrieved.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SemanticQueryManager.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class SemanticQueryManager extends QueryManager {

    private static final long serialVersionUID = 1L;

    private static final transient Logger logger = 
	LoggerFactory.getLogger(SemanticQueryManager.class);

    private SPARQLQueryReasoner parser = new SPARQLQueryReasoner();
    
    /*
     * Table containing responses UUID that are performing queryDataStore
     * operation.
     */
    private Map<UUID, Boolean> responsesWaitingForQueryDataStore = new HashMap<UUID, Boolean>();
    
    public SemanticQueryManager() {

    }

    public Map<UUID, Boolean> getResponsesWaitingForQueryDataStore() {
        return this.responsesWaitingForQueryDataStore;
    }

    /**
     * {@inheritDoc}
     */
    public Response dispatch(Query query) throws DispatchException {
        if (!(query instanceof SPARQLQuery)) {
            throw new DispatchException("Only SPARQL queries can be dispatched.");
        }

        try {
            ParsedSPARQLQuery parsedQuery = this.parser.decompose((SPARQLQuery) query);

            if (logger.isDebugEnabled()) {
                StringBuffer buf = new StringBuffer();
                buf.append("Query '");
                buf.append(query.toString());
                buf.append("' is dispatched as:\n");
                for (SPARQLQuery subQuery : parsedQuery.getSubQueries()) {
                    buf.append("  --> ");
                    buf.append(subQuery.toString());
                    buf.append("\n");
                }
                logger.debug(buf.toString().substring(0, buf.length() - 1));
            }

            List<SynchronousMessage<?>> subQueryResponses = 
                		new ArrayList<SynchronousMessage<?>>(
                			parsedQuery.getSubQueries().size());

            for (SPARQLQuery subQuery : parsedQuery.getSubQueries()) {
                subQueryResponses.add(this.process(this.preProcess(subQuery)));
            }

            SemanticResponseMessage<?> semanticQueryResponseMessage = null;
            if (parsedQuery.requireFiltration()) {
                try {
                    semanticQueryResponseMessage = 
                        SPARQLQueryFilter.getInstance().filter(
                        	(SPARQLQuery) query,
				    subQueryResponses.toArray(
					    new SemanticResponseMessage[] {}));
                } catch (SemanticSpaceException e) {
                    e.printStackTrace();
                }
                semanticQueryResponseMessage.setLatency(this.computeLatency(subQueryResponses));
            } else {
                semanticQueryResponseMessage = 
                    ((SemanticResponseMessage<?>) subQueryResponses.get(0));
            }

            return this.postProcess(query, semanticQueryResponseMessage);
        } catch (MalformedQueryException e) {
            throw new DispatchException("An error occured while parsing query", e);
	}
    }

    private int computeLatency(List<SynchronousMessage<?>> subQueries) {
        int max = 0;
        for (SynchronousMessage<?> subQuery : subQueries) {
            max = Math.max(max, ((AnycastResponseMessage<?>)subQuery).getLatency());
        }
        return max;
    }

    /**
     * Merge a synchronous message with an existing one in the table.
     * 
     * @param msg
     *            the synchronous message to merge with.
     */
    public SynchronousMessageEntry mergeAndSetSynchronousMessage(SynchronousMessage<?> msg) {
            SynchronousMessageEntry entry = 
                super.getSynchronousMessagesReceived().get(msg.getUUID());
            AnycastResponseMessage<?> msgFromMessagesReceived = 
                (AnycastResponseMessage<?>) entry.getSynchronousMessage();
            AnycastResponseMessage<?> msgFromParameter = 
                (AnycastResponseMessage<?>) msg;
            
            if (msgFromMessagesReceived != null) {
                msgFromParameter.addAll(msgFromMessagesReceived.getDataRetrieved());
                
                int nbStepsToAdd = 1;
                
                if (msgFromParameter.getOutboundHopCount() > 
                    entry.getNbStepsWhenSynchronizationPointCreated()) {
                    nbStepsToAdd += 
                        msgFromParameter.getOutboundHopCount() - 
                        entry.getNbStepsWhenSynchronizationPointCreated();
                }
                
                msgFromParameter.setNbStepsForSingleTrip(
                        msgFromMessagesReceived.getOutboundHopCount() + 
                        nbStepsToAdd);
                msgFromParameter.setNbSteps(
                        msgFromMessagesReceived.getReturnHopCount() +
                        msgFromParameter.getReturnHopCount());
            }
            entry.setSynchronousMessage(msgFromParameter);
            return entry;
    }

    /**
     * Sets the specified {@link SynchronousMessage} in the table of messages
     * received.
     * 
     * @param msg
     *            the message to add.
     * 
     * @return the {@link SynchronousMessageEntry} which has been just added in
     *         the table.
     */
    public SynchronousMessageEntry setSynchronousMessage(SynchronousMessage<?> msg) {
        SynchronousMessageEntry entry = super.getSynchronousMessagesReceived().get(msg.getUUID());
        entry.setSynchronousMessage(msg);
        return entry;
    }

    /**
     * {@inheritDoc}
     */
    public SynchronousMessage<?> preProcess(Query query) throws PreProcessException {
        SynchronousMessage<?> msg = null;

        if (query instanceof SPARQLQuery) {
            try {
                msg = this.createSynchronousMessage((SPARQLQuery) query);
            } catch (SemanticSpaceQueryException e) {
                throw new PreProcessException("Cannot identify query", e);
            }
        } else {
            throw new PreProcessException("Unknown query type");
        }

        return msg;
    }

    private SemanticQueryMessage createSynchronousMessage(SPARQLQuery query) 
                                            throws SemanticSpaceQueryException {
        return SPARQLQueryIdentifier.createQueryMessage(query);
    }

    /**
     * {@inheritDoc}
     */
    public Response postProcess(Query originalQuery, AbstractResponseMessage<?> response) 
                        throws PostProcessException {
        if (logger.isDebugEnabled()) {
            logger.debug("Response: nbHopsForSingleTrip={}, nbHopsForReturnTrip={}, latency={}", 
                    new Object[] { 
                        response.getOutboundHopCount(), 
                        response.getReturnHopCount(), 
                        response.getLatency() });
        }
        
        return response.createResponse();
    }

    public void route(SynchronousMessage<?> msg) {
//        msg.incrementNbStepsBy(1);
        
        if (msg instanceof AnycastQueryMessage) {
            if (logger.isDebugEnabled()) {
                logger.debug("Route with instance of AnycastQueryMessage");
            }
            this.processAnycastQueryMessage((AnycastQueryMessage) msg);
        } else if (msg instanceof AnycastResponseMessage<?>) {
            if (logger.isDebugEnabled()) {
                logger.debug("Route with instance of AnycastResponseMessage");
            }
            this.processAnycastQueryResponseMessage((AnycastResponseMessage<?>) msg);
        } else {
            if (logger.isDebugEnabled()) {
                logger.debug("Route with instance of {}", msg.getClass());
            }
            super.route(msg);
        }
    }

    /**
     * Called by all peers receiving the query except for the first peer on
     * which the query is started (hop 0).
     * 
     * @param query
     *            the query to handle.
     * 
     * @see Peer#route(SynchronousMessage)
     */
    public void processAnycastQueryMessage(AnycastQueryMessage query) {
        // Have we seen this request before ?
        if (super.getSynchronousMessagesIdentifiersMet().contains(query.getUUID())) {
            if (logger.isDebugEnabled()) {
                logger.debug("On peer which has already received the query {}, sent empty response", 
                            query.getUUID());
            }
            query.getAnycastRoutingList().removeLast().getPeerStub().route(
                                                query.createResponseMessage());
        } else {
            super.addSynchronousMessageIdentifierMet(query.getUUID());
            query.route(super.overlay);
        }
    }

    /**
     * Called by all {@link Peer}s which have met the query even the initiator
     * of the query.
     * 
     * @param response
     *            the response to handle.
     * 
     * @see Peer#route(SynchronousMessage)
     * @see AnycastResponseRouter#route(org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay,
     *      SynchronousMessage)
     */
    public void processAnycastQueryResponseMessage(AnycastResponseMessage<?> response) {
        SynchronousMessageEntry entry = super.overlay.mergeAndSetSynchronousMessage(response);
        entry.incrementNbResponsesReceivedBy(1);

        // Wait for the sub chain because only each
        // peer knows how many messages have been sent
        // from them

        // FIXME in order to avoid synchronization for response on
        // each peer, it is possible to send the number of responses
        // to wait to the initiator of the response while we are sending
        // the query to neighbors.
        if (entry.getStatus() == SynchronousMessageEntry.Status.ALL_RESPONSES_RECEIVED) {
            synchronized (this.responsesWaitingForQueryDataStore) {
                while (this.responsesWaitingForQueryDataStore.get(response.getUUID()) != null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug("All responses received on {} but query datastore not yet done", overlay);
                    }
                    try {
                        this.responsesWaitingForQueryDataStore.wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
            
            // Call the response router in order to know where to go:
            // either we still have peers to visit in our stack, or we are
            // on the first peer of the call chain
            entry.getSynchronousMessage().incrementNbStepsBy(1);
            entry.getSynchronousMessage().route(super.overlay);
            
            // The response has been handled and sent back so we
            // can remove it from the table. However, only if we
            // are not on the initiator of the query or it cannot
            // retrieve the result
            if (response.getAnycastRoutingList().size() > 0) {
                super.overlay.getQueryManager().
                    getSynchronousMessagesReceived().
                        remove(response.getUUID());
            }
            
            if (logger.isDebugEnabled()) {
                logger.debug("Peer {} has received all responses", super.overlay);
            }
        }
    }

//    /**
//     * Called when the current object is unserialized in order to initialize
//     * transient fields.
//     * 
//     * @return the current object instance.
//     * @throws ObjectStreamException
//     */
//    protected Object readResolve() throws ObjectStreamException {
//        this.inMemoryDataStore = SesameModelFactory.createModel();
//        this.inMemoryDataStore.open();
//        return this;
//    }

}
