package fr.inria.gforge.dspace.overlay.overlay.chord;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.proactive.core.util.ProActiveRandom;
import org.ontoware.rdf2go.model.node.URI;

import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;

/**
 * SemanticResource is a resource managed by {@link SemanticSpaceChordOverlay}.
 * It contains information necessary for CAN/Chord binding in Semantic Spaces.
 * 
 * Each resource contains a list of stored peers and only a
 * {@link SemanticResource} knows this list. Each peer in this list are of type
 * CAN. For these reasons, a SemanticResource plays the role of a tracker by
 * maintaining a list of peers.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SemanticResource.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class SemanticResource implements Serializable {

    private static final long serialVersionUID = 1L;

    private URI spaceURI;

    /**
     * Stubs of type CAN maintained.
     */
    private List<SemanticPeer> remotePeers = new ArrayList<SemanticPeer>();

    /**
     * Constructs a new resource with the specified <code>spaceURI</code> and
     * <code>remotePeer</code> of type CAN.
     * 
     * @param spaceURI
     *            the space associated to this resource.
     * @param remotePeer
     *            the stub of type CAN to maintained.
     */
    public SemanticResource(URI spaceURI, SemanticPeer remotePeer) {
        this.spaceURI = spaceURI;
        this.remotePeers.add(remotePeer);
    }

    public boolean add(SemanticPeer remotePeer) {
        if (this.remotePeers.size() == 0) {
            return this.remotePeers.add(remotePeer);
        } else {
            if (ProActiveRandom.nextDouble() 
                    < SemanticSpaceChordOverlay.probabilityToStoreCANPeer) {
                return this.remotePeers.add(remotePeer);
            }
            return true;
        }
    }

    public URI getSpaceURI() {
        return this.spaceURI;
    }

    /**
     * Returns all the {@link SemanticPeer}s stubs of type CAN maintained.
     * 
     * @return all the {@link SemanticPeer}s stubs of type CAN maintained.
     */
    public List<SemanticPeer> getAll() {
        return this.remotePeers;
    }

    /**
     * Returns a {@link SemanticPeer} randomly selected from the
     * {@link SemanticPeer}s stubs of type CAN maintained.
     * 
     * @return a {@link SemanticPeer} randomly selected from the
     *         {@link SemanticPeer}s stubs of type CAN maintained.
     */
    public SemanticPeer getRandomPeer() {
        return this.remotePeers.get(
                ProActiveRandom.nextInt(
                        this.remotePeers.size()));
    }
    
    /**
     * Removes a {@link SemanticPeer} randomly selected from the
     * {@link SemanticPeer}s stubs of type CAN maintained.
     * 
     * @return The {@link SemanticPeer} stub removed.
     */
    public SemanticPeer removeRandomPeer() {
        return this.remotePeers.remove(
                ProActiveRandom.nextInt(
                        this.remotePeers.size()));
    }
    
    /**
     * {@inheritDoc}
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append(this.remotePeers.size());
        buf.append(" peers maintained for space '");
        buf.append(this.spaceURI);
        buf.append("'");
        return buf.toString();
    }

}
