package fr.inria.gforge.dspace.overlay.overlay.chord;

import java.util.Iterator;
import java.util.Set;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.core.node.NodeException;
import org.objectweb.proactive.core.util.ProActiveRandom;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.StructuredP2PException;
import org.objectweb.proactive.extensions.structuredp2p.overlay.chord.AbstractChordOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.chord.Entry;
import org.objectweb.proactive.extensions.structuredp2p.overlay.chord.Key;
import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.TriplePattern;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.api.SemanticFactory;
import fr.inria.gforge.dspace.overlay.kernel.SemanticSpaceOverlayKernel;
import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;
import fr.inria.gforge.dspace.overlay.overlay.SemanticStructuredOverlay;
import fr.inria.gforge.dspace.overlay.overlay.can.SemanticSpaceCANOverlay;

/**
 * SemanticSpaceChordOverlay extends {@link AbstractChordOverlay} in order to
 * manage {@link SemanticResource} as resources. This overlay is used to perform
 * the binding between Chord and CAN networks in Semantic Spaces.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SemanticSpaceChordOverlay.java 3127 2010-01-06 15:59:42Z
 *          laurentp $
 */
public class SemanticSpaceChordOverlay extends AbstractChordOverlay<SemanticResource> 
                                        implements SemanticStructuredOverlay {

    private static final Logger logger = LoggerFactory.getLogger(SemanticSpaceCANOverlay.class);
    
    private static final long serialVersionUID = 1L;
    
    public static double probabilityToStoreCANPeer = 0.5;

    public SemanticSpaceChordOverlay() {
        super();
    }

    /**
     * Returns a {@link SemanticPeer} stub of type CAN (randomly selects)
     * associated with the specified {@link Key} or <code>null</code> if no
     * entry is found for the specified key.
     * 
     * @param hashedSpaceURI
     *            the key to use.
     * @return a {@link SemanticPeer} stub of type CAN (randomly selects)
     *         associated with the specified {@link Key} or <code>null</code> if
     *         no entry is found for the specified key.
     */
    public SemanticPeer getRandomPeer(Key hashedSpaceURI) {
        Entry<SemanticResource> entryFound = super.findEntry(hashedSpaceURI);

        if (entryFound == null || entryFound.getValue().getAll().size() == 0) {
            return null;
        }

        return entryFound.getValue().getRandomPeer();
    }

    /**
     * Returns a {@link SemanticResource} for the specified {@link Key} or
     * <code>null</code> if no entry is found for the specified key.
     * 
     * @param hashedSpaceURI
     *            the key to use.
     * @return a {@link SemanticResource} for the specified {@link Key} or
     *         <code>null</code> if no entry is found for the specified key.
     */
    public SemanticResource getSemanticResource(Key hashedSpaceURI) {
        Entry<SemanticResource> entryFound = super.findEntry(hashedSpaceURI);

        if (entryFound == null || entryFound.getValue().getAll().size() == 0) {
            return null;
        }

        return entryFound.getValue();
    }

    /**
     * Puts a specified {@link SemanticPeer} of type CAN on the local peer by
     * supposing that the local peer manages the specified
     * <code>hashedSpaceURI</code>. This function creates a new space if it does
     * not exist.
     * 
     * @param spaceURI
     *            the spaceURI to use.
     * 
     * @param remotePeer
     *            the {@link SemanticPeer} stub of type CAN to maintained.
     * @return <code>true</code> if operation has succeeded, <code>false</code>
     *         otherwise.
     */
    public boolean put(URI spaceURI, SemanticPeer remotePeer) {
        Key hashedSpaceURI = new Key(spaceURI.toString());
        Entry<SemanticResource> entryFound = this.getEntry(hashedSpaceURI);

        if (entryFound == null) {
            // creates the space with one initial peer of type CAN
            return super.addEntry(
                    new Entry<SemanticResource>(
                            hashedSpaceURI, new SemanticResource(spaceURI, remotePeer)));
        } else {
            // space already exists, we add a new peer of type CAN
            try {
                remotePeer.join(entryFound.getValue().getRandomPeer());
            } catch (StructuredP2PException e) {
                e.printStackTrace();
                return false;
            }

            if (ProActiveRandom.nextDouble() <= SemanticSpaceChordOverlay.probabilityToStoreCANPeer) {
                if (logger.isDebugEnabled()) {
                    logger.debug("New CAN peer stored on node {}.", this.getKey());
                }
                return entryFound.getValue().add(remotePeer);
            }

            return true;
        }
    }

    /**
     * Add a new peer of type CAN on the network associated to the specified
     * spaceURI. This method must be called on the peer which manages the hash
     * of spaceURI.
     * 
     * @param spaceURI
     *            the spaceURI to use in order to identify the space on which to
     *            add a new peer of type CAN.
     * 
     * @param kernel
     *            the {@link SemanticSpaceOverlayKernel} associated to the new
     *            CAN peer created.
     */
    public boolean addNewCanPeer(URI spaceURI, SemanticSpaceOverlayKernel kernel) {
        Key hashedSpaceURI = new Key(spaceURI.toString());
        Entry<SemanticResource> entry = this.getEntry(hashedSpaceURI);

        if (entry == null) {
            return false;
        } else {
            SemanticPeer newCanPeer = null;
            try {
                newCanPeer = SemanticFactory.newActiveSemanticPeer(
                        ((SemanticPeer)super.getLocalPeer()).getTrackers(),
                        new SemanticSpaceCANOverlay(spaceURI, kernel));
            } catch (ActiveObjectCreationException e) {
                e.printStackTrace();
            } catch (NodeException e) {
                e.printStackTrace();
            }
            try {
                newCanPeer.join(entry.getValue().getRandomPeer());
            } catch (StructuredP2PException e) {
                e.printStackTrace();
            }
            return entry.getValue().add(newCanPeer);
        }
    }

    /*
     * Implementation of SemanticSpaceOperations interface
     */

    public void addAll(URI space, Iterator<? extends Statement> other)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, Statement statement) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public boolean contains(URI space, Statement s) throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return false;
    }

    public boolean contains(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return false;
    }

    public boolean contains(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            String plainLiteral) throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return false;
    }

    public ClosableIterator<Statement> findStatements(URI space, TriplePattern triplepattern)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return null;
    }

    public Set<URI> listSpaces() {
        // TODO Auto-generated method stub
        return null;
    }

    public void removeAll(URI space) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeAll(URI space, Iterator<? extends Statement> statements)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, Statement statement) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatements(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public void removeStatements(URI space, TriplePattern triplePattern)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub

    }

    public boolean sparqlAsk(URI space, String query) throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return false;
    }

    public ClosableIterable<Statement> sparqlConstruct(URI space, String query)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return null;
    }

    public ClosableIterable<Statement> sparqlDescribe(URI space, String query)
	    throws SemanticSpaceException {
	// TODO Auto-generated method stub
	return null;
    }
    
    public QueryResultTable sparqlSelect(URI space, String queryString)
            throws SemanticSpaceException {
        // TODO Auto-generated method stub
        return null;
    }

}
