package fr.inria.gforge.dspace.overlay.router.can;

import java.util.Map;
import java.util.UUID;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.AbstractCANOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.NeighborEntry;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.NeighborTable;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.Zone;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.requests.BlockingRequestReceiverException;
import org.objectweb.proactive.extensions.structuredp2p.router.Router;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.inria.gforge.dspace.overlay.messages.query.can.AnycastQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.AnycastRoutingEntry;
import fr.inria.gforge.dspace.overlay.messages.response.can.AnycastResponseMessage;
import fr.inria.gforge.dspace.overlay.overlay.SemanticQueryManager;
import fr.inria.gforge.dspace.overlay.overlay.can.SemanticSpaceCANOverlay;

/**
 * @author Laurent Pellegrino
 * 
 * @version $Id: AnycastQueryRouter.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public abstract class AnycastQueryRouter<T extends AnycastQueryMessage> 
                extends Router<AnycastQueryMessage, CoordinateValue[]> {

    private static final transient Logger logger = 
        LoggerFactory.getLogger(AnycastQueryRouter.class);

    public AnycastQueryRouter(ConstraintsValidator<CoordinateValue[]> validator) {
        super(validator);
    }
    
    /**
     * This method is called when a message reach a peer which validates
     * constraints of routing just before the next routing step.
     * 
     * @param overlay
     *            the {@link AbstractCANOverlay} of the peer which validates
     *            constraints.
     * 
     * @param msg the message which is handled.
     */
    public abstract void onPeerWhichValidatesKeyConstraints(AbstractCANOverlay overlay, AnycastQueryMessage msg);

    protected void performHandle(final StructuredOverlay overlay, final AnycastQueryMessage msg) {
        NeighborTable neighborsToSendTo = getNeighborsToSendTo(overlay, msg);

        /*
         * At this step there are three cases :
         * 
         * 1) current peer has no neighbors means that the query can only be
         * handled by himself;
         * 
         * 2) neighborsToSendTo equals 0 means that we don't have to route query
         * anymore: we are on a leaf and response must be returned;
         * 
         * 3) neighborsToSendTo > 0 means we have to perform many send operation
         * and the current peer must await for the number of responses sent.
         */
        if (((SemanticSpaceCANOverlay) overlay).getNeighborTable().size() == 0) {
            super.onDestinationReached(overlay, msg);
            overlay.getSynchronousMessagesReceived().put(
                    msg.getUUID(), new SynchronousMessageEntry(1, msg.getHopCount()));
            AnycastResponseMessage<?> response = msg.createResponseMessage();
            response.queryDataStoreAndStoreData(overlay);
            response.incrementNbStepsBy(1);
            overlay.route(response);
        } else if (neighborsToSendTo.size() == 0) {
            super.onDestinationReached(overlay, msg);
            AnycastResponseMessage<?> response = 
                (AnycastResponseMessage<?>) msg.createResponseMessage();
            response.queryDataStoreAndStoreData(overlay);
            response.incrementNbStepsBy(1);

            if (msg.getAnycastRoutingList().size() > 0) {
                AnycastRoutingEntry entry = response.getAnycastRoutingList().removeLast();
                if (!entry.validatesKeyConstraints()) {
                    response.setRouteResponseOnPeersValidatingConstraints(false);
                }
                entry.getPeerStub().route(response);
            } else {
                overlay.getSynchronousMessagesReceived().put(
                        msg.getUUID(), new SynchronousMessageEntry(1, msg.getHopCount()));
                overlay.route(response);
            }
        } else {
            // Adds entry to table in order to know the
            // number of responses awaited
            overlay.getSynchronousMessagesReceived().put(
                    msg.getUUID(), 
                    new SynchronousMessageEntry(
                            neighborsToSendTo.size(), msg.getHopCount()));
            
            ((SemanticQueryManager) overlay.getQueryManager()).
                getResponsesWaitingForQueryDataStore().put(msg.getUUID(), true);
            
            new Thread(new Runnable() {
                public void run() {
                    /*
                     * Performs query datastore while query is sent to neighbors in
                     * order to overlap the network communications.
                     */
                    SynchronousMessageEntry entry = overlay.getSynchronousMessagesReceived().get(msg.getUUID());
                    
                    if (entry.getSynchronousMessage() == null) {
                        AnycastResponseMessage<?> response = msg.createResponseMessage();
                        response.queryDataStoreAndStoreData(overlay);
                        overlay.mergeAndSetSynchronousMessage(response);
                    } else {
                        ((AnycastResponseMessage<?>)entry.
                                getSynchronousMessage()).
                                    queryDataStoreAndStoreData(overlay);
                    }
                                       
                    ((SemanticQueryManager) overlay.getQueryManager()).
                        getResponsesWaitingForQueryDataStore().
                            remove(msg.getUUID());
                    
                    synchronized (((SemanticQueryManager) overlay.getQueryManager()).
                            getResponsesWaitingForQueryDataStore()) {
                        ((SemanticQueryManager) overlay.getQueryManager()).
                            getResponsesWaitingForQueryDataStore().notifyAll();
                    }
                }
            }).start();
            
            // Constructs the routing list used by responses for return trip
            msg.getAnycastRoutingList().add(new AnycastRoutingEntry(
                                overlay.getIdentifier(), 
                                overlay.getRemotePeer(), 
                                true));
            
            msg.incrementNbStepsBy(1);
            for (NeighborEntry neighbor : neighborsToSendTo) {
                neighbor.getStub().route(msg);
            }
        }
        
        if (logger.isDebugEnabled()) {
            logger.debug("Query sent to {} neighbor(s).", neighborsToSendTo.size());
        }
    }

    private NeighborTable getNeighborsToSendTo(final StructuredOverlay overlay,
                                               final AnycastQueryMessage msg) {
        NeighborTable neighborsToSendTo = new NeighborTable();
        neighborsToSendTo.addAll(
                ((SemanticSpaceCANOverlay) overlay).getNeighborTable());

        for (int dimension = 0; dimension < AbstractCANOverlay.NB_DIMENSIONS; dimension++) {
            for (int direction = 0; direction < 2; direction++) {
                for (NeighborEntry entry : neighborsToSendTo.get(dimension, direction).values()) {
                    AnycastRoutingEntry entryCommingFromSender = msg.getAnycastRoutingList()
                            .getRoutingResponseEntryBy(entry.getIdentifier());
                    /*
                     * If message contains the neighbor identifier in its
                     * routing list, we know that this neighbor has
                     * already receive the query and we can remove it from the
                     * neighbors we need to send the query: it avoids sometimes a
                     * remote call.
                     */
                    if (entryCommingFromSender != null) {
                        neighborsToSendTo.remove(entry.getIdentifier(), dimension, direction);
                    } else {
                        if (logger.isDebugEnabled()) {
                            StringBuilder buf = new StringBuilder();
                            buf.append("* On ");
                            buf.append(overlay.getLocalPeer());
                            buf.append(" checking ");
                            buf.append(entry.getStub());
                            buf.append(", validatesKeyConstraints=");
                            buf.append(this.validatesKeyConstraints(msg, entry.getZone()));
                            logger.debug(buf.toString());
                        }

                        if (!this.validatesKeyConstraints(msg, entry.getZone())) {
                            neighborsToSendTo.remove(entry.getIdentifier(), dimension, direction);
                        }
                    }
                }
            }
        }
        
        return neighborsToSendTo;
    }
    
    private boolean validatesKeyConstraints(AnycastQueryMessage msg, Zone zone) {
        for (int i = 0; i < msg.getKeyToReach().length; i++) {
            // if coordinate is null we skip the test
            if (msg.getKeyToReach()[i] != null) {
                // the specified overlay does not contains the key
                if (zone.contains(i, msg.getKeyToReach()[i]) != 0) {
                    return false;
                }
            }
        }
        return true;
    }

    protected void performRoute(StructuredOverlay overlay, AnycastQueryMessage msg) {
        AbstractCANOverlay overlayCAN = ((AbstractCANOverlay) overlay);
        CoordinateValue[] coordinatesToReach = msg.getKeyToReach();

        if (super.validatesKeyConstraints(overlayCAN, msg.getKeyToReach())) {
            if (logger.isDebugEnabled()) {
                logger.debug("Route on peer ({}) which validates constraints.",
                        overlay);
            }
            this.onPeerWhichValidatesKeyConstraints(overlayCAN, msg);
            this.performHandle(overlay, msg);
        } else { // we must find a peer which valid the constraints
            int direction;
            int pos;

            if (logger.isDebugEnabled()) {
                logger.debug("Route in order to find a peer which validates constraints");
            }
            for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
                if (coordinatesToReach[dim] == null) {
                    continue;
                }

                direction = NeighborTable.SUPERIOR_DIRECTION;
                pos = overlayCAN.contains(dim, coordinatesToReach[dim]);
                // pos NOT equals 0 means the coordinates are in the zone of the
                // neighbour
                if (pos == -1) {
                    direction = NeighborTable.INFERIOR_DIRECTION;
                }

                // Step 1 : routing in order to found a peer which validates
                // constraints
                if (pos != 0) {
                    Map<UUID, NeighborEntry> neighbors = overlayCAN.getNeighborTable().get(dim,
                            direction);

                    if (neighbors.size() > 0) {
                        Peer nearestPeer;

                        if (coordinatesToReach[AbstractCANOverlay.getNextDimension(dim)] == null) {
                            nearestPeer = 
                                overlayCAN.getNeighborTable().getRandomNeighbor(dim, direction);
                        } else {
                            nearestPeer = overlayCAN.getNeighborTable()
                                    .getRandomNeighborValidatingConstraints(
                                            overlayCAN.getZone(), 
                                            coordinatesToReach, 
                                            dim,
                                            direction);
                        }

                        overlay.getSynchronousMessagesReceived().put(msg.getUUID(),
                                new SynchronousMessageEntry(1, msg.getHopCount()));
                        msg.getAnycastRoutingList().add(
                                new AnycastRoutingEntry(overlay.getIdentifier(), 
                                        overlay.getRemotePeer(), false));

                        msg.incrementNbStepsBy(1);
                        try {
                            nearestPeer.route(msg);
                        } catch (BlockingRequestReceiverException e) {
                            // overlay.getQueryManager().bufferizeSynchronousMessage(this);
                            e.printStackTrace();
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        break;
                    } else {
                        logger.error("No neighbour to route to and message hasn't reached final destination");
                    }
                }
            }
        }
    }

}
