package fr.inria.gforge.dspace.overlay.router.can;

import org.objectweb.proactive.extensions.structuredp2p.messages.response.AbstractResponseMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.router.Router;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.inria.gforge.dspace.overlay.messages.query.can.AnycastRoutingEntry;
import fr.inria.gforge.dspace.overlay.messages.response.can.AnycastResponseMessage;
import fr.inria.gforge.dspace.overlay.overlay.SemanticQueryManager;
import fr.inria.gforge.dspace.overlay.overlay.can.SemanticSpaceCANOverlay;

/**
 * Route an {@link AbstractResponseMessage} in {@link SemanticSpaceCANOverlay}
 * by using anycast routing.
 * 
 * @param <T>
 *            the type of response to route.
 * @param <D>
 *            data type contained by response.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: AnycastResponseRouter.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class AnycastResponseRouter<T extends AnycastResponseMessage<D>, D> extends
        Router<AnycastResponseMessage<D>, CoordinateValue[]> {

    public AnycastResponseRouter(ConstraintsValidator<CoordinateValue[]> validator) {
        super(validator);
    }

    private static final transient Logger logger = 
        LoggerFactory.getLogger(AnycastResponseRouter.class);

    protected void performHandle(StructuredOverlay overlay, AnycastResponseMessage<D> msg) {
        // We are on the first peer of the call chain
        // and it has already handled the query
        synchronized (overlay.getSynchronousMessagesReceived()) {
            overlay.getSynchronousMessagesReceived().notifyAll();
        }
    }

    /**
     * This method is called depending on
     * {@link SemanticQueryManager#processAnycastQueryResponseMessage(AnycastResponseMessage)}
     * .
     * 
     * @param overlay
     *            the overlay to use for performing the route operation.
     * 
     * @param msg
     *            the response message to route.
     * 
     * @see SemanticSpaceCANOverlay
     */
    protected void performRoute(StructuredOverlay overlay, AnycastResponseMessage<D> msg) {
        if (msg.getAnycastRoutingList().size() > 0) {
            AnycastRoutingEntry entry = msg.getAnycastRoutingList().removeLast();
            if (entry.validatesKeyConstraints()) {
                if (logger.isDebugEnabled()) {
                    logger.debug(
                            "On peer {}, route response on peer {} validating constraints.", 
                            overlay, 
                            entry.getPeerStub());
                }
            } else {
                msg.setRouteResponseOnPeersValidatingConstraints(false);
                if (logger.isDebugEnabled()) {
                    logger.debug(
                            "On peer {}, route response on peer {} not validating constraints.", 
                            overlay, 
                            entry.getPeerStub());
                }
            }
            
            msg.incrementNbStepsBy(1);
            entry.getPeerStub().route(msg);
        } else {
            this.performHandle(overlay, msg);
        }
    }

}
