package fr.inria.gforge.dspace.overlay.tracker;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.objectweb.proactive.extensions.structuredp2p.messages.response.chord.TurnAroundResponseMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.chord.Key;
import org.objectweb.proactive.extensions.structuredp2p.tracker.Tracker;
import org.objectweb.proactive.extensions.structuredp2p.tracker.TrackerChord;
import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.TriplePattern;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.SpaceReference;
import fr.inria.gforge.dspace.overlay.api.messages.query.AskQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.ConstructQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.SelectQuery;
import fr.inria.gforge.dspace.overlay.api.messages.response.AskResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.ConstructResponse;
import fr.inria.gforge.dspace.overlay.api.messages.response.SelectResponse;
import fr.inria.gforge.dspace.overlay.kernel.operations.datastore.DatastoreResponseOperation;
import fr.inria.gforge.dspace.overlay.messages.query.can.AddStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.can.RemoveStatementsQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.query.chord.ListSpacesQueryMessage;
import fr.inria.gforge.dspace.overlay.operations.chord.GetRandomPeerOperation;
import fr.inria.gforge.dspace.overlay.operations.chord.GetRandomPeerResponseOperation;
import fr.inria.gforge.dspace.overlay.operations.chord.GetSemanticResourceOperation;
import fr.inria.gforge.dspace.overlay.operations.chord.GetSemanticResourceResponseOperation;
import fr.inria.gforge.dspace.overlay.overlay.SemanticPeer;
import fr.inria.gforge.dspace.overlay.util.RDF2GoBuilder;
import fr.inria.gforge.dspace.overlay.util.SemanticHelper;

/**
 * SemanticChordTracker is a {@link TrackerChord} which can be used as an entry
 * point for user operations on Semantic Spaces.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SemanticChordTracker.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class SemanticChordTracker extends TrackerChord implements SemanticTracker {

    private static final long serialVersionUID = 1L;

    public SemanticChordTracker() {
        super();
    }

    public SemanticChordTracker(String associatedNetworkName) {
        super(associatedNetworkName);
    }

    public SemanticChordTracker(String associatedNetworkName, String trackerName) {
        super(associatedNetworkName, trackerName);
    }

    /**
     * Returns a {@link SemanticPeer} of type CAN, randomly chosen for the
     * specified space {@link URI}.
     * 
     * @param space
     *            the space {@link URI} to look for.
     * @return a {@link SemanticPeer} randomly chosen for the specified space
     *         {@link URI}.
     * @throws SemanticSpaceException
     *             if the specified space doesn't exist.
     */
    private SemanticPeer getRandomSemanticPeer(URI space) throws SemanticSpaceException {
        SemanticPeer canPeer = ((GetRandomPeerResponseOperation) PAFuture.getFutureValue(super
                .getRandomPeer().send(
                        new GetRandomPeerOperation(new Key(space.toString()))))).getRemotePeer();

        if (canPeer == null) {
            throw new SemanticSpaceException("Space '" + space + "' not found");
        }

        return canPeer;
    }

    /**
     * Returns a {@link SpaceReference} for the specified <code>spaceURI</code>.
     * 
     * @param spaceURI
     *            the space URI associated to the {@link SpaceReference} to get.
     * @return a {@link SpaceReference} for the specified space
     *         <code>spaceURI</code>.
     */
    public SpaceReference getSpaceReference(URI spaceURI) {
        List<SemanticChordTracker> trackers = new ArrayList<SemanticChordTracker>();
        Iterator<Tracker> it = super.getTrackersGroup().iterator();
        while (it.hasNext()) {
            trackers.add((SemanticChordTracker) it.next());
        }
        
        return new SpaceReference(spaceURI, trackers, 
                ((GetSemanticResourceResponseOperation) PAFuture.getFutureValue(
                        super.getRandomPeer().send(
                               new GetSemanticResourceOperation(
                                       new Key(spaceURI.toString()))))).getSemanticResource().getAll());
    }

    /*
     * SemanticSpaceOperations implementation
     */
    public void addAll(URI space, Iterator<? extends Statement> iterator)
            throws SemanticSpaceException {
        while (iterator.hasNext()) {
            this.addStatement(space, iterator.next());
        }
    }

    public void addStatement(URI space, Statement statement) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, statement)));

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subject, predicate, literal)));

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subject, predicate, literal, languageTag)));

    }

    public void addStatement(URI space, Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subject, predicate, literal, datatypeURI)));
    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subjectURIString, predicate, literal)));

    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subjectURIString, predicate, literal,
                        languageTag)));
    }

    public void addStatement(URI space, String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new AddStatementQueryMessage(space, subjectURIString, predicate, literal,
                        datatypeURI)));
    }

    public boolean contains(URI space, Statement stmt) throws SemanticSpaceException {
        return this.contains(space, stmt.getSubject(), stmt.getPredicate(), stmt.getObject());
    }

    public boolean contains(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        return this.contains(space, subject, predicate, object.toString());
    }

    @SuppressWarnings("unchecked")
    public boolean contains(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            String plainLiteral) throws SemanticSpaceException {

        StringBuffer query = new StringBuffer();
        query.append("ASK { GRAPH<");
        query.append(space);
        query.append("> { ");
        if (subject == null) {
            query.append("?s ");
        } else {
            query.append(subject);
            query.append(" ");
        }
        if (predicate == null) {
            query.append("?p ");
        } else {
            query.append(predicate);
            query.append(" ");
        }
        if (plainLiteral == null) {
            query.append("?o ");
        } else {
            query.append(plainLiteral);
        }
        query.append(" } . }");

        try {
            return ((DatastoreResponseOperation<Boolean>) PAFuture.getFutureValue(
                    ((GetRandomPeerResponseOperation) PAFuture.getFutureValue(
                            super.getRandomPeer().send(
                                    new GetRandomPeerOperation(
                                            new Key(space.toString()))))).getRemotePeer().send(
                                                    new AskQuery(space, query.toString()))))
                    .getValue();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public URI createFederation(Set<URI> spaces) throws SemanticSpaceException {
        throw new SemanticSpaceException("Not yet implemented");
    }

    public void createSpaceRelation(URI sSpaces, URI relation, URI oSpace)
            throws SemanticSpaceException {
        throw new SemanticSpaceException("Not yet implemented");

    }

    public void deleteFederation(URI fed) throws SemanticSpaceException {
        throw new SemanticSpaceException("Not yet implemented");
    }

    public ClosableIterator<Statement> findStatements(URI space, TriplePattern triplepattern)
            throws SemanticSpaceException {
        return this.sparqlConstruct(space,
                SemanticHelper.createConstructSparqlFrom(space, triplepattern)).iterator();
    }

    @SuppressWarnings("unchecked")
    public Set<URI> listSpaces() {
        List<URI> spacesURI = ((TurnAroundResponseMessage<URI>) PAFuture.getFutureValue(super
                .getRandomPeer().send(new ListSpacesQueryMessage()))).getValues();
        return new HashSet<URI>(spacesURI);
    }

    public void registerEndPoint(URI endpoint) throws SemanticSpaceException {
        throw new SemanticSpaceException("Not yet implemented");
    }

    public void removeAll(URI space) throws SemanticSpaceException {
        this.removeStatements(space, null, null, null);
    }

    public void removeAll(URI space, Iterator<? extends Statement> it)
            throws SemanticSpaceException {
        while (it.hasNext()) {
            this.removeStatement(space, it.next());
        }
    }

    public void removeStatement(URI space, Statement statement) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, statement)));
    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subject, predicate, literal)));
    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subject, predicate, literal, languageTag)));
    }

    public void removeStatement(URI space, Resource subject, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subject, predicate, literal, datatypeURI)));
    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal)
            throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subjectURIString, predicate, literal)));
    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal,
            String languageTag) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subjectURIString, predicate, literal,
                        languageTag)));
    }

    public void removeStatement(URI space, String subjectURIString, URI predicate, String literal,
            URI datatypeURI) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementQueryMessage(space, subjectURIString, predicate, literal,
                        datatypeURI)));
    }

    public void removeStatements(URI space, ResourceOrVariable subject, UriOrVariable predicate,
            NodeOrVariable object) throws SemanticSpaceException {
        PAFuture.waitFor(this.getRandomSemanticPeer(space).send(
                new RemoveStatementsQueryMessage(space,
                        RDF2GoBuilder.createStatement(
                                (Resource) subject, (URI) predicate, (Node) object))));
    }

    public void removeStatements(URI space, TriplePattern triplePattern)
            throws SemanticSpaceException {
        this.removeStatements(space, triplePattern.getSubject(), triplePattern.getPredicate(),
                triplePattern.getObject());
    }

    public boolean sparqlAsk(URI space, String query) throws SemanticSpaceException {
        try {
            return ((AskResponse) PAFuture.getFutureValue(this.getRandomSemanticPeer(space)
                    .send(new AskQuery(space, query)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public ClosableIterable<Statement> sparqlConstruct(URI space, String query)
            throws SemanticSpaceException {
        try {
            return ((ConstructResponse) PAFuture.getFutureValue(this.getRandomSemanticPeer(
                    space).send(new ConstructQuery(space, query)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

    public QueryResultTable sparqlSelect(URI space, String queryString)
            throws SemanticSpaceException {
        try {
            return ((SelectResponse) PAFuture.getFutureValue(this.getRandomSemanticPeer(
                    space).send(new SelectQuery(space, queryString)))).getResult();
        } catch (DispatchException e) {
            throw new SemanticSpaceException(e);
        }
    }

}
