package fr.inria.gforge.dspace.overlay.util;

import org.ontoware.rdf2go.RDF2Go;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.TriplePattern;
import org.ontoware.rdf2go.model.node.BlankNode;
import org.ontoware.rdf2go.model.node.DatatypeLiteral;
import org.ontoware.rdf2go.model.node.LanguageTagLiteral;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.PlainLiteral;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;
import org.ontoware.rdf2go.model.node.impl.URIImpl;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;

import com.ontotext.trree.rdf2go.OwlimModelFactory;

/**
 * Provides some static methods in order to create {@link RDF2Go}
 * objects.
 * 
 * By default the model is created using the {@link OwlimModelFactory}. It is
 * possible to specify an another factory to use for the model creation by
 * calling {@link RDF2GoBuilder#setModelFactory(String)}.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: RDF2GoBuilder.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class RDF2GoBuilder {

    private static Model model;

    static {
        org.ontoware.rdf2go.RDF2Go.register("com.ontotext.trree.rdf2go.OwlimModelFactory");
        model = org.ontoware.rdf2go.RDF2Go.getModelFactory().createModel();
    }

    public static BlankNode createBlankNode() {
        return model.createBlankNode();
    }

    public static BlankNode createBlankNode(String value) {
        return model.createBlankNode(value);
    }

    public static DatatypeLiteral createDataTypeLiteral(String arg0, URI arg1) {
        return model.createDatatypeLiteral(arg0, arg1);
    }

    public static LanguageTagLiteral createLanguageTagLiteral(String arg0, String arg1) {
        return model.createLanguageTagLiteral(arg0, arg1);
    }

    public static PlainLiteral createPlainliteral(String arg0) {
        return model.createPlainLiteral(arg0);
    }

    public static Statement createStatement(Resource subject, URI predicate, Node object) {
        return model.createStatement(subject, predicate, object);
    }

    public static TriplePattern createTriplePattern(ResourceOrVariable subject,
            UriOrVariable predicate, NodeOrVariable object) {
        return model.createTriplePattern(subject, predicate, object);
    }

    public static URI createURI(String uri) {
        return model.createURI(uri);
    }

    /**
     * Parses a specified subject as String in order to create a new subject
     * using RDF2Go types.
     * 
     * @param subject
     *            the subject to parse.
     * @return the RDF2Go subject created.
     * @throws SemanticSpaceException
     *             if subject is null.
     */
    public static Resource toSubject(String subject) throws SemanticSpaceException {
        if (subject == null) {
            throw new SemanticSpaceException("subject cannot be null");
        }

        Resource mySubject = null;

        if (subject.equals("_:")) {
            mySubject = model.createBlankNode();
        } else if (subject.startsWith("_:")) {
            mySubject = model.createBlankNode(subject.replaceFirst("_:", ""));
        } else {
            mySubject = model.createURI(subject);
        }

        return mySubject;
    }

    /**
     * Parses a specified predicate as String in order to create a new predicate
     * using RDF2Go types.
     * 
     * @param predicate
     *            the predicate to parse.
     * @return the RDF2Go predicate created.
     * @throws SemanticSpaceException
     *             if predicate is null.
     */
    public static URI toPredicate(String predicate) throws SemanticSpaceException {
        if (predicate == null) {
            throw new SemanticSpaceException("predicate cannot be null");
        }

        return model.createURI(predicate);
    }

    /**
     * Parses a specified object as String in order to create a new object
     * using RDF2Go types.
     * 
     * @param object
     *            the object to parse.
     * @return the RDF2Go object created.
     * @throws SemanticSpaceException
     *             if object is null.
     */
    public static Node toObject(String object) throws SemanticSpaceException {
        if (object == null) {
            throw new SemanticSpaceException("object cannot be null");
        }

        Node myObject = null;

        if (object.equals("_:")) {
            myObject = model.createBlankNode();
        } else if (object.startsWith("_:")) {
            myObject = model.createBlankNode(object.replaceFirst("_:", ""));
        } else {
            try {
                // check if a valid URI
                // openrdf URIImpl does not not work propertly in this case
                new org.openrdf.model.impl.URIImpl(object);
                myObject = new URIImpl(object);
            } catch (Exception e) {
                // when no valid uri create plain literal
                myObject = model.createPlainLiteral(object);
            }
        }

        return myObject;
    }

    /**
     * Parses a single statement from a given Statement as a triple of Strings.
     * 
     * @param subject
     *            must be a blank node or URI.
     * @param predicate
     *            must be a valid URI.
     * @param object
     *            must be must be a valid URI, blank node or literal.
     * @return a new {@link Statement}.
     * @throws SemanticSpaceException
     *             if triple is not valid.
     */
    public static Statement toStatement(String subject, String predicate, String object)
            throws SemanticSpaceException {
        if (subject == null || predicate == null || object == null) {
            throw new SemanticSpaceException("subject, predicate and object cannot be null");
        }

        return model.createStatement(toSubject(subject), toPredicate(predicate), toObject(object));
    }

    public static Model getModel() {
        return model;
    }

    public static void setModelFactory(String modelFactoryClass) {
        org.ontoware.rdf2go.RDF2Go.register(modelFactoryClass);
        model = org.ontoware.rdf2go.RDF2Go.getModelFactory().createModel();
    }

}
