package fr.inria.gforge.dspace.overlay.util;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.ModelSet;
import org.ontoware.rdf2go.model.Statement;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.api.messages.query.AskQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.ConstructQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.DescribeQuery;
import fr.inria.gforge.dspace.overlay.api.messages.query.SPARQLQuery;
import fr.inria.gforge.dspace.overlay.datastore.SesameModelFactory;
import fr.inria.gforge.dspace.overlay.messages.response.can.AskResponseMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.ConstructResponseMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.DescribeResponseMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.SelectResponseMessage;
import fr.inria.gforge.dspace.overlay.messages.response.can.SemanticResponseMessage;
import fr.inria.gforge.dspace.overlay.rdf2go.wrappers.ClosableIterableWrapper;
import fr.inria.gforge.dspace.overlay.rdf2go.wrappers.QueryResultTableWrapper;

/**
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SPARQLQueryFilter.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class SPARQLQueryFilter {

    private static SPARQLQueryFilter instance;
    
    private transient ModelSet inMemoryDataStore;
    
    public SPARQLQueryFilter() {
	this.inMemoryDataStore = SesameModelFactory.createModelSet();
        this.inMemoryDataStore.open();
    }
    
    /**
     * 
     * 
     * @param query
     *            the initial {@link SPARQLQuery}.
     * @param subQueryResponses
     *            a list containing the responses associated to the initial
     *            SPARQL query which has been decomposed into sub-queries.
     * @return an unique response after having filtered sub responses.
     * @throws SemanticSpaceException
     *             if a problem occurs when data are filtered.
     */
    @SuppressWarnings("unchecked")
    public synchronized SemanticResponseMessage<?> filter(SPARQLQuery query,
            SynchronousMessage<?>... subQueryResponses) throws SemanticSpaceException {
        Model model = this.inMemoryDataStore.getModel(query.getSpaceURI());
        model.setAutocommit(false);
        ClosableIterator<Statement> it;
        ClosableIterableWrapper data;
        for (SynchronousMessage<?> semanticQueryResponseMessage : subQueryResponses) {
            data = ((SemanticResponseMessage<ClosableIterableWrapper>) 
                        semanticQueryResponseMessage).mergeAndGetDataRetrieved();
            if (data == null) {
                continue;
            }
            it = data.toRDF2Go().iterator();

            while (it.hasNext()) {
                model.addStatement(it.next());
            }
        }
        model.commit();

        ConstructResponseMessage subQuery =
            ((ConstructResponseMessage) subQueryResponses[0]);
        
        SemanticResponseMessage<?> response;
        
        if (query instanceof AskQuery) {
            response = new AskResponseMessage(
        	    	subQuery, subQuery.getKeyToReach(), 
        	    	this.inMemoryDataStore.sparqlAsk(query.toString()));
        } else if (query instanceof ConstructQuery) {
            response = new ConstructResponseMessage(
        	    	subQuery, 
        	    	subQuery.getKeyToReach(), 
        	    	new ClosableIterableWrapper(
        	    		this.inMemoryDataStore.sparqlConstruct(
        	    			query.toString())));
        } else if (query instanceof DescribeQuery) {
            response = new DescribeResponseMessage(
    	    	subQuery, 
    	    	subQuery.getKeyToReach(), 
    	    	new ClosableIterableWrapper(
    	    		this.inMemoryDataStore.sparqlDescribe(
    	    			query.toString())));
        } else {
            response = new SelectResponseMessage(
    	    	subQuery, 
    	    	subQuery.getKeyToReach(), 
    	    	new QueryResultTableWrapper(this.inMemoryDataStore.sparqlSelect(query.toString())));
        }

        model.removeAll();
        model.commit();
        
        return response;
    }
    
    
    public static SPARQLQueryFilter getInstance() {
	if (instance == null) {
	    instance = new SPARQLQueryFilter();
	}
	
	return instance;
    }
    
}
