package fr.inria.gforge.dspace.overlay.core.overlay;

import junit.framework.Assert;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.objectweb.proactive.api.PAFuture;
import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.node.URI;

import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import fr.inria.gforge.dspace.overlay.SpaceReference;
import fr.inria.gforge.dspace.overlay.connector.ManagementConnector;
import fr.inria.gforge.dspace.overlay.initializers.GlobalNetworkInitializer;
import fr.inria.gforge.dspace.overlay.messages.query.chord.NbReferencesStoredQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.response.chord.NbReferencesStoredResponseMessage;
import fr.inria.gforge.dspace.overlay.tracker.SemanticChordTracker;
import fr.inria.gforge.dspace.overlay.util.RDF2GoBuilder;
import fr.inria.gforge.dspace.overlay.util.SemanticHelper;

/**
 * Test {@link SpaceReference} operations.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: TestSpaceReference.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class TestSpaceReference {

    private static GlobalNetworkInitializer networkInitializer =
            new GlobalNetworkInitializer();

    private static URI defaultSpace = RDF2GoBuilder.createURI("http://www.inria.fr");
    
    private static SpaceReference spaceReference;

    @BeforeClass
    public static void setUp() {
        networkInitializer.setUpNetworkOnLocalMachine();
        
        ManagementConnector managementConnector = networkInitializer.getManagementConnector();
        managementConnector.setAutoRemoveDatastoreAtShutdown(true);
        managementConnector.newPeersOnChordInParallel(3);
        try {
            managementConnector.createSpace(defaultSpace);
            managementConnector.newPeersOnCANInParallel(defaultSpace, 3);
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        spaceReference = ((SemanticChordTracker) networkInitializer.getRandomTracker()).
                                                            getSpaceReference(defaultSpace);

        Assert.assertTrue(((NbReferencesStoredResponseMessage) PAFuture.getFutureValue(
                networkInitializer.getRandomTracker().getRandomPeer().send(
                        new NbReferencesStoredQueryMessage(defaultSpace)))).getNbReferences() >= 1);
    }

    @Test
    public void testSpaceReference() {
        this.testFindStatements();
        this.testAddStatement();
        this.testSparqlAsk();
        this.testSparqlConstruct();
        this.testRemoveAll();
        this.testGetAllSemanticPeers();
    }
    
    public void testFindStatements() {
        ClosableIterator<Statement> result = null;
        try {
            result = spaceReference.findStatements(
                    RDF2GoBuilder.createTriplePattern(null, null, null));
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        Assert.assertEquals(0, SemanticHelper.size(result));
    }

    public void testAddStatement() {
        try {
            spaceReference.addStatement(
                    RDF2GoBuilder.createURI("http://www.inria.fr"),
                        RDF2GoBuilder.createURI("http://www.inria.fr/#located"),
                            "France");
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        ClosableIterator<Statement> result = null;
        try {
            result = spaceReference.findStatements(
                    RDF2GoBuilder.createTriplePattern(null, null, null));
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        Assert.assertEquals(1, SemanticHelper.size(result));
    }

    public void testSparqlAsk() {
        boolean result = false;
        try {
            result = spaceReference.sparqlAsk("ASK { ?s ?p ?o }");
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        Assert.assertTrue(result);
    }

    public void testSparqlConstruct() {
        ClosableIterable<Statement> result = null;

        try {
            result = spaceReference.sparqlConstruct(
                        "CONSTRUCT { ?s ?p ?o } WHERE { GRAPH<" +
                                defaultSpace +
                                "> { ?s ?p ?o } }");
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        Assert.assertEquals(1, SemanticHelper.size(result));
    }

    public void testRemoveAll() {
        try {
            spaceReference.removeAll();
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }
        
        boolean result = false;
        try {
            result = spaceReference.sparqlAsk("ASK { ?s ?p ?o }");
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }

        Assert.assertFalse(result);
    }

    public void testGetAllSemanticPeers() {
        Assert.assertTrue(spaceReference.getAllSemanticPeers().size() > 0);
    }

    @AfterClass
    public static void tearDown() {
        networkInitializer.tearDownNetwork();
    }

}
