package fr.inria.gforge.dspace.rdf2go;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.net.URL;
import java.util.Map;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.api.PAFuture;
import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.exception.LockException;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.Syntax;
import org.ontoware.rdf2go.model.impl.AbstractModel;
import org.ontoware.rdf2go.model.node.BlankNode;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;

import at.sti2.semanticspaces.api.exceptions.OperationNotSupportedException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.api.exceptions.SpaceModelRuntimeException;
import at.sti2.semanticspaces.api.exceptions.SpaceNotExistsException;
import fr.inria.gforge.dspace.overlay.SpaceReference;
import fr.inria.gforge.dspace.overlay.messages.query.chord.SpaceExistsQueryMessage;
import fr.inria.gforge.dspace.overlay.messages.response.chord.SpaceExistsResponseMessage;
import fr.inria.gforge.dspace.overlay.tracker.SemanticChordTracker;
import fr.inria.gforge.dspace.overlay.util.RDF2GoBuilder;

/**
 * Wrapper class for the overlay in order to distribute data over a distributed
 * network.
 * 
 * When constructed, an OverlayModel is associated to a specified space. All
 * operations can only be performed on this space. If the specified space
 * doesn't exist when the model is initialized a {@link SpaceNotExistsException}
 * is thrown.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: OverlayModel.java 5038 2010-08-19 09:09:50Z plaurent $
 */
public class OverlayModel extends AbstractModel {

    private static final long serialVersionUID = 1L;

    /**
     * The Space URI associated to this Model.
     */
    private URI space;

    /**
     * The SpaceReference associated to this Space in order to avoid to have
     * communication with a tracker each time an operation must be handled.
     */
    private SpaceReference spaceReference;

    /**
     * The tracker to use in order to have an entry point in the structured p2p
     * network.
     */
    private SemanticChordTracker tracker;

    /**
     * Creates a new OverlayModel for the specified space by using the given
     * <code>remoteTracker</code>.
     * 
     * @param space
     *            the space associated to this model.
     * @param remoteTracker
     *            the {@link SemanticChordTracker} stub to use.
     * @throws SemanticSpaceException
     *             if the Chord network hasn't been initialized or if the space
     *             doesn't exist.
     */
    public OverlayModel(URI space, SemanticChordTracker remoteTracker) throws SemanticSpaceException {
        this.space = space;
        this.tracker = remoteTracker;
        
        this.setSpaceReference();
    }
    
    /**
     * Creates a new OverlayModel for the specified space by using the given
     * <code>trackerURI</code>.
     * 
     * @param space
     *            the space associated to this model.
     * 
     * @param trackerURI
     *            the {@link URL} associated to the tracker to use in order to
     *            have an entry point in the structured p2p network.
     * @throws SemanticSpaceException
     *             if the Chord network hasn't been initialized or if the space
     *             doesn't exist.
     */
    public OverlayModel(final URI space, URL trackerURI) throws SemanticSpaceException {
        this.space = space;
        try {
            this.tracker = PAActiveObject.lookupActive(
                                SemanticChordTracker.class, trackerURI.toString());
        } catch (ActiveObjectCreationException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        
        this.setSpaceReference();
    }
    
    private void setSpaceReference() throws SemanticSpaceException {
        if (this.tracker.getNbOfPeersStored() == 0) {
            throw new SemanticSpaceException("Chord network not initialized.");
        }
        
        if (!((SpaceExistsResponseMessage) PAFuture.getFutureValue(
                this.tracker.getRandomPeer().send(
                        new SpaceExistsQueryMessage(space)))).getValue()) {
            
           throw new SpaceNotExistsException("Space '" + space + "' not exists.");
        }

        this.spaceReference = this.tracker.getSpaceReference(this.space);
    }
    
    /*
     * AbstractModel methods override
     */

    /**
     * {@inheritDoc}
     */
    public void addStatement(Resource arg0, URI arg1, Node arg2) throws ModelRuntimeException {
        try {
            this.spaceReference.addStatement(arg0, arg1, arg2.toSPARQL());
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void removeStatement(Resource arg0, URI arg1, Node arg2) throws ModelRuntimeException {
        try {
            this.spaceReference.addStatement(arg0, arg1, arg2.toSPARQL());
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void dump() {
        StringBuffer buf = new StringBuffer("OverlayModel(");
        buf.append("space=");
        buf.append(this.space);
        buf.append(")");
        System.out.println(buf.toString());
    }

    /**
     * {@inheritDoc}
     */
    public URI getContextURI() {
        return this.space;
    }

    /**
     * Returns always <code>false</code>.
     * 
     * @param arg0
     *            the to check with.
     * 
     * @return always <code>false</code>.
     */
    public boolean isIsomorphicWith(Model arg0) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public BlankNode createBlankNode() {
        return RDF2GoBuilder.createBlankNode();
    }

    /**
     * {@inheritDoc}
     */
    public BlankNode createBlankNode(String arg0) {
        return RDF2GoBuilder.createBlankNode(arg0);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isValidURI(String arg0) {
        boolean isValid = true;
        try {
            RDF2GoBuilder.createURI(arg0);
        } catch (IllegalArgumentException e) {
            isValid = false;
        }
        return isValid;
    }

    /**
     * {@inheritDoc}
     */
    public ClosableIterator<Statement> iterator() {
        ClosableIterator<Statement> res = null;
        try {
            res = this.spaceReference.sparqlConstruct(
                    "CONSTRUCT { ?s ?p ?o } WHERE { ?s ?p ?o }").iterator();
        } catch (SemanticSpaceException e) {
            e.printStackTrace();
        }
        return res;
    }

    /**
     * Returns always <code>false</code> because a distributed model cannot be
     * locked.
     * 
     * @return always <code>false</code> because a distributed model cannot be
     *         locked.
     */
    public boolean isLocked() {
        return false;
    }

    /**
     * Operation not support on this distributed model.
     */
    public void lock() throws LockException {
        throw new OperationNotSupportedException();
    }

    /**
     * Operation not support on this distributed model.
     */
    public void unlock() {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public ClosableIterator<Statement> findStatements(ResourceOrVariable arg0, UriOrVariable arg1,
            NodeOrVariable arg2) throws ModelRuntimeException {
        try {
            return this.spaceReference.findStatements(RDF2GoBuilder.createTriplePattern(arg0, arg1,
                    arg2));
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean sparqlAsk(String query) throws ModelRuntimeException {
        try {
            return this.spaceReference.sparqlAsk(query);
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public ClosableIterable<Statement> sparqlConstruct(String arg0) throws ModelRuntimeException{
        ClosableIterable<Statement> res = null;
        try {
            res = this.spaceReference.sparqlConstruct(arg0);
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
        return res;
    }

    /**
     * {@inheritDoc}
     */
    public ClosableIterable<Statement> sparqlDescribe(String arg0) throws ModelRuntimeException {
        ClosableIterable<Statement> res = null;
        try {
            res = this.spaceReference.sparqlDescribe(arg0);
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }
        return res;
    }

    /**
     * {@inheritDoc}
     */
    public QueryResultTable sparqlSelect(String arg0) throws ModelRuntimeException {
        QueryResultTable res = null;

        try {
            res = this.spaceReference.sparqlSelect(arg0);
        } catch (SemanticSpaceException e) {
            throw new SpaceModelRuntimeException(e);
        }

        return res;
    }

    /**
     * {@inheritDoc}
     */
    public void readFrom(Reader arg0) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public void readFrom(InputStream arg0) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public void readFrom(Reader arg0, Syntax arg1) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public void writeTo(Writer arg0) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public void writeTo(OutputStream arg0) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * {@inheritDoc}
     */
    public void writeTo(Writer arg0, Syntax arg1) throws IOException, ModelRuntimeException {
        throw new OperationNotSupportedException();
    }

    /**
     * Operation not supported.
     */
    public String getNamespace(String arg0) {
        throw new OperationNotSupportedException();
    }

    /**
     * Operation not supported.
     */
    public Map<String, String> getNamespaces() {
        throw new OperationNotSupportedException();
    }

    /**
     * Operation not supported.
     */
    public void removeNamespace(String arg0) {
        throw new OperationNotSupportedException();
    }

    /**
     * Operation not supported.
     */
    public void setNamespace(String arg0, String arg1) throws IllegalArgumentException {
        throw new OperationNotSupportedException();
    }

}
