package org.objectweb.proactive.extensions.structuredp2p.api;

import java.util.concurrent.Callable;

import org.objectweb.proactive.ActiveObjectCreationException;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.core.node.Node;
import org.objectweb.proactive.core.node.NodeException;
import org.objectweb.proactive.extensions.structuredp2p.overlay.OverlayType;
import org.objectweb.proactive.extensions.structuredp2p.tracker.Tracker;
import org.objectweb.proactive.extensions.structuredp2p.tracker.TrackerCAN;
import org.objectweb.proactive.extensions.structuredp2p.tracker.TrackerChord;

/**
 * {@link TrackerFactory} provides some static methods in order to create easily
 * a new active tracker.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: TrackerFactory.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class TrackerFactory {

    /**
     * Creates a new Tracker ActiveObject.
     * 
     * @param type
     *            the type of the tracker, which is one of {@link OverlayType}.
     * @return the new Tracker object created.
     * @throws ActiveObjectCreationException
     * @throws NodeException
     */
    public static Tracker newActiveTracker(OverlayType type) 
            throws ActiveObjectCreationException, NodeException {
        return TrackerFactory.newActiveTracker(type, "default");
    }

    /**
     * Creates a new Tracker ActiveObject.
     * 
     * @param type
     *            the type of the tracker, which is one of {@link OverlayType}.
     * @param node
     *            the node to use for deployment.
     * @return the new Tracker object created.
     * @throws ActiveObjectCreationException
     * @throws NodeException
     */
    public static Tracker newActiveTracker(OverlayType type, Node node)
            throws ActiveObjectCreationException, NodeException {
        return TrackerFactory.newActiveTracker(type, "default", node);
    }

    /**
     * Creates a new Tracker ActiveObject.
     * 
     * @param type
     *            the type of the tracker, which is one of {@link OverlayType}.
     * @param associatedNetworkName
     *            the name of the network to which the tracker is associated.
     * @return the new Tracker object created.
     * @throws ActiveObjectCreationException
     * @throws NodeException
     */
    public static Tracker newActiveTracker(OverlayType type, String associatedNetworkName)
            throws ActiveObjectCreationException, NodeException {
        return TrackerFactory.newActiveTracker(type, associatedNetworkName, null);
    }

    /**
     * Creates a new Tracker ActiveObject.
     * 
     * @param type
     *            the type of the tracker, which is one of {@link OverlayType}.
     * @param node
     *            the node to use for deployment.
     * @param associatedNetworkName
     *            the name of the network to which the tracker is associated.
     * @return the new Tracker object created.
     * @throws ActiveObjectCreationException
     * @throws NodeException
     */
    public static Tracker newActiveTracker(OverlayType type, String associatedNetworkName, 
            Node node) throws ActiveObjectCreationException, NodeException {
        Class<? extends Tracker> classType;

        if (type == OverlayType.CAN) {
            classType = TrackerCAN.class;
        } else if (type == OverlayType.CHORD) {
            classType = TrackerChord.class;
        } else {
            throw new IllegalArgumentException("Unknown '" + type + "' OverlayType");
        }

        return (Tracker) PAActiveObject.newActive(classType,
                new Object[] { associatedNetworkName }, node);
    }

    /**
     * Creates the specified number of trackers in parallel.
     * 
     * @param type
     *            the type of the tracker.
     * @param number
     *            the number of trackers to create.
     * @return the trackers created.
     */
    public static Tracker[] newActiveTrackerInParallel(final OverlayType type, int number) {
        return Executor.execute(Tracker.class, new Callable<Tracker>() {
            public Tracker call() throws Exception {
                return TrackerFactory.newActiveTracker(type);
            }
        }, number);
    }
    
}
