package org.objectweb.proactive.extensions.structuredp2p.messages;

import java.io.Serializable;
import java.util.UUID;

import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.router.Router;
import org.objectweb.proactive.extensions.structuredp2p.router.RouterStore;

/**
 * @author Laurent Pellegrino
 * 
 * @version $Id: SynchronousMessage.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public abstract class SynchronousMessage<K> implements Routable<K>, Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * Universally unique identifier used in order to identify the response.
     */
    private UUID uuid;

    /**
     * The key used in order to route the query on the network.
     */
    private K keyToReach;

    /**
     * The number of peers through which the message has
     * passed.
     */
    private int hopCount = 0;

    /**
     * Constructs a new SynchronousMessage with the specified identifier and
     * keyToReach.
     * 
     * @param uuid
     *            the unique identifier associated to the message.
     * @param keyToReach
     *            the key to reach (i.e. routes until reaching the {@link Peer}
     *            managing this key).
     */
    public SynchronousMessage(UUID uuid, K keyToReach) {
        this.uuid = uuid;
        this.keyToReach = keyToReach;
    }

    /**
     * Returns the universally unique identifier used in order to identify the
     * response.
     * 
     * @return the universally unique identifier used in order to identify the
     *         response.
     */
    public UUID getUUID() {
        return this.uuid;
    }

    /**
     * Returns the timestamp of the creation of the message.
     * 
     * @return the timestamp of the creation of the message.
     */
    public abstract long getDispatchTimestamp();

    /**
     * Increments the number of steps by the specified number.
     * 
     * @param increment
     *            the size of the increment.
     */
    public void incrementNbStepsBy(int increment) {
        this.hopCount += increment;
    }

    /**
     * Returns the key to reach (i.e. the peer containing this key is the
     * receiver of this message).
     * 
     * @return the key to reach (i.e. the peer containing this key is the
     *         receiver of this message).
     */
    public K getKeyToReach() {
        return this.keyToReach;
    }

    /**
     * Returns the number of peers through which the message has passed in order
     * to reach the <code>keyToReach</code>.
     * 
     * @return the number of peers through which the message has passed in order
     *         to reach the <code>keyToReach</code>.
     */
    public int getHopCount() {
        return this.hopCount;
    }

    @SuppressWarnings("unchecked")
    public void route(StructuredOverlay overlay) {
        Class<SynchronousMessage<K>> clazz =
                (Class<SynchronousMessage<K>>) this.getClass();
        RouterStore routerStore = RouterStore.getInstance();
        if (!routerStore.contains(clazz)) {
            routerStore.store(clazz, this.getRouter());
        }
        ((Router<SynchronousMessage<K>, K>) routerStore.get(clazz)).route(overlay, this);
    }
    
    /**
     * Sets the key to reach.
     * 
     * @param keyToReach
     *            the new key to reach.
     */
    public void setKeyToReach(K keyToReach) {
        this.keyToReach = keyToReach;
    }

    public void setNbSteps(int nbSteps) {
        this.hopCount = nbSteps;
    }

}
