package org.objectweb.proactive.extensions.structuredp2p.messages.query;

import java.util.UUID;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.response.AbstractResponseMessage;

/**
 * An <code>AbstractQueryMessage</code> is an abstraction of a query that can be
 * sent on a structured peer-to-peer network in order to find some data by a key
 * which is an object of type <code>K</code> . In response an object of type
 * {@link AbstractResponseMessage} is returned with the desired data.
 * <p>
 * A query is performed step by step by using one-way mechanism. So, we can't
 * say when the response will be returned. Suppose that the peer A is sending a
 * query in order to reach the peer B managing the key <code>keyToReach</code>.
 * The first step consists in setting the <code>keyToReach</code> to
 * <code>keyToFound</code>. After that the query is sent step by step until the
 * peer managing the <code>keyToReach</code> is found. When it is found, the
 * <code>keyToReach</code> change for <code>keyFromSender</code>. At this time
 * the response is routed to the sender in the opposite direction without
 * necessarily using the same route. This last point depends on the concrete
 * type of query which can implement the desired behavior.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: AbstractQueryMessage.java 5022 2010-08-14 16:14:45Z plaurent $
 * 
 * @see SynchronousMessage
 * @see AbstractResponseMessage
 */
public abstract class AbstractQueryMessage<K> extends SynchronousMessage<K> {

    private static final long serialVersionUID = 1L;

    /**
     * Timestamp of the creation of the message.
     */
    private long dispatchTimestamp = System.currentTimeMillis();

    /**
     * Constructs a new query message with the specified <code>keyToReach</code>.
     * 
     * @param keyToReach
     *            the key to reach.
     */
    public AbstractQueryMessage(K keyToReach) {
        super(UUID.randomUUID(), keyToReach);
    }

    /**
     * Constructs a new query message with the specified <code>uuid</code>,
     * <code>keyToReach</code> and <code>dispatchTimestamp</code>.
     * 
     * @param uuid
     *            the universally unique identifier associated to the query.
     * @param keyToReach
     *            the key to reach.
     * @param dispatchTimestamp
     *            the dispatch timestamp of the query.
     */
    public AbstractQueryMessage(UUID uuid, K keyToReach, long dispatchTimestamp) {
        super(uuid, keyToReach);
        this.dispatchTimestamp = dispatchTimestamp;
    }

    /**
     * Creates an {@link AbstractResponseMessage} in accordance to the type of
     * the current {@link AbstractQueryMessage}.
     * 
     * @return an {@link AbstractResponseMessage} in accordance to the type of
     * the current {@link AbstractQueryMessage}.
     */
    public abstract AbstractResponseMessage<?> createResponseMessage();

    /**
     * {@inheritDoc}
     */
    public long getDispatchTimestamp() {
        return this.dispatchTimestamp;
    }

}
