package org.objectweb.proactive.extensions.structuredp2p.messages.response;

import org.objectweb.proactive.extensions.structuredp2p.api.messages.Response;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.query.AbstractQueryMessage;

/**
 * An <code>AbstractQueryResponse</code> is an abstract response associated to
 * an abstract {@link AbstractQueryMessage}. This class contains some
 * informations like the latency (in ms), the number of steps for the send, the
 * number of steps for the reception, ...
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: SynchronousResponseMessage.java 3127 2010-01-06 15:59:42Z
 *          laurentp $
 */
public abstract class AbstractResponseMessage<K> 
                            extends SynchronousMessage<K> {

    private static final long serialVersionUID = 1L;

    private long dispatchTimestamp;
    
    private long deliveryTimestamp;

    private int outboundHopCount;
    
    private int latency = -1;

    /**
     * Constructs a new response message with the specified <code>query</code>
     * and <code>keyToReach</code>.
     * 
     * @param response
     *            the response to merge with.
     * @param keyToReach
     *            the key used in order to route the response to it recipient.
     */
    public AbstractResponseMessage(AbstractResponseMessage<K> response, K keyToReach) {
        super(response.getUUID(), keyToReach);
        this.dispatchTimestamp = response.getDispatchTimestamp();
        this.outboundHopCount = response.getOutboundHopCount();
        super.incrementNbStepsBy(response.getReturnHopCount());
    }

    /**
     * Constructs a new response message with the specified <code>query</code>
     * and <code>keyToReach</code>.
     * 
     * @param query
     *            the query which creates the response.
     * @param keyToReach
     *            the key used in order to route the response to it recipient.
     */
    public AbstractResponseMessage(AbstractQueryMessage<K> query, K keyToReach) {
        super(query.getUUID(), keyToReach);
        this.dispatchTimestamp = query.getDispatchTimestamp();
        this.outboundHopCount = query.getHopCount();
    }

    /**
     * Converts the current response (from private API) to a response from the
     * public API.
     * 
     * @return a response from the public API.
     */
    public abstract Response createResponse();

    /**
     * Returns the timestamp of the delivery.
     * 
     * @return the timestamp of the delivery.
     */
    public long getDeliveryTimestamp() {
        return this.deliveryTimestamp;
    }

    /**
     * Returns the timestamps of the dispatch of the initial query.
     * 
     * @return the timestamps of the dispatch of the initial query.
     */
    public long getDispatchTimestamp() {
        return this.dispatchTimestamp;
    }

    /**
     * Returns the latency (in milliseconds). It is the time between the moment
     * of the creation of the message and when the response has been received.
     * 
     * @return the latency between the moment of the creation of the message
     *         (dispatch) and when the response has been received (delivery).
     * 
     * @see #getDeliveryTimestamp()
     * @see #getDispatchTimestamp()
     */
    public int getLatency() {
        if (this.latency < 0) {
            throw new IllegalStateException(
                    "The response has not been receive from network after a query (latency=" + this.latency + ")");
        }
        return this.latency;
    }

    /**
     * Returns the number of peers through which the return message has passed.
     * 
     * @return the number of peers through which the return message has passed.
     */
    public int getReturnHopCount() {
        return super.getHopCount();
    }

    /**
     * Returns the number of peers through which the outbound message has
     * passed.
     * 
     * @return the number of peers through which the outbound message has
     *         passed.
     */
    public int getOutboundHopCount() {
        return this.outboundHopCount;
    }
    
    /**
     * Returns the total number of steps : the number of steps to send the query
     * plus the number of steps to receive the response.
     * 
     * @return the total number of steps : the number of steps to send the query
     *         plus the number of steps to receive the response.
     */
    public int getTotalHopCount() {
        return this.outboundHopCount + super.getHopCount();
    }

    /**
     * Sets the delivery time of the response (ie. when the response has been
     * received). At this step, the latency is automatically calculated.
     */
    public void setDeliveryTime() {
        if (this.latency != -1) {
            throw new IllegalStateException("Delivery time has already been set");
        }

        this.deliveryTimestamp = System.currentTimeMillis();
        this.latency = (int) (this.deliveryTimestamp - this.dispatchTimestamp);
        if (this.latency < 0) {
            this.latency = 0;
        }
    }

    public void setLatency(int duration) {
        this.latency = duration;
    }
    
    public void setNbStepsForSingleTrip(int nbStepsForSingleTrip) {
        this.outboundHopCount = nbStepsForSingleTrip;
    }

    /**
     * Returns a string containing a concise, human-readable description of this
     * object.
     * 
     * @return a string containing a concise, human-readable description of this
     *         object.
     */
    public String toString() {
        return "uuid=" + this.getUUID();
    }

}
