package org.objectweb.proactive.extensions.structuredp2p.overlay;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import org.objectweb.proactive.Body;
import org.objectweb.proactive.EndActive;
import org.objectweb.proactive.InitActive;
import org.objectweb.proactive.RunActive;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.core.body.exceptions.BodyTerminatedException;
import org.objectweb.proactive.extensions.structuredp2p.api.PeerFactory;
import org.objectweb.proactive.extensions.structuredp2p.api.messages.Query;
import org.objectweb.proactive.extensions.structuredp2p.api.messages.Response;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.NetworkAlreadyJoinedException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.NetworkNotJoinedException;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.operations.ResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.Operation;

/**
 * A Peer contains all operations which are common to peer-to-peer protocols.
 * These operations are the join and leave but also the send operation in order
 * to send messages. A Peer is composed of a {@link StructuredOverlay} which
 * allows to have several implementations of common operations for each protocol
 * to implement.
 * 
 * Warning, this class must not be instantiate directly. In order to create a
 * new active peer you must use the {@link PeerFactory}.
 * 
 * @author Alexandre Trovato
 * @author Fanny Kilanga
 * @author Laurent Pellegrino
 * 
 * @version $Id: Peer.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class Peer implements Comparable<Object>, InitActive, EndActive, RunActive, Serializable {

    private static final long serialVersionUID = 1L;

    private Set<UUID> peersWhichAreLeaving = new HashSet<UUID>();

    protected StructuredOverlay structuredOverlay;

    /**
     * Indicates if the join operation has already called and succeeded or not.
     */
    private boolean activated = false;

    private Peer stub;

    /**
     * The no-argument constructor as commanded by ProActive.
     */
    public Peer() {

    }

    /**
     * Constructs a new peer using the specified overlay. Warning, you must use
     * {@link PeerFactory} to instantiate a new active peer.
     * 
     * @param overlay
     *            the overlay to set to the new peer.
     */
    public Peer(StructuredOverlay overlay) {
        this.structuredOverlay = overlay;
        this.structuredOverlay.setLocalPeer(this);
    }

    /**
     * Returns debug information as String.
     * 
     * @return debug information as String.
     */
    public String dump() {
        return this.structuredOverlay.dump();
    }

    public boolean equals(Object obj) {
        return this.structuredOverlay.equals(obj);
    }

    /**
     * Returns the {@link Body} of the current active object.
     * 
     * @return the {@link Body} of the current active object.
     */
    public Body getBody() {
        return PAActiveObject.getBodyOnThis();
    }

    public UUID getIdentifier() {
        return this.structuredOverlay.getIdentifier();
    }

    /**
     * Returns the peers which are neighbors and which are leaving.
     * 
     * @return the peers which are neighbors and which are leaving.
     */
    public Set<UUID> getPeersWhichArePreparingToLeave() {
        return this.peersWhichAreLeaving;
    }

    /**
     * Returns the {@link StructuredOverlay} which is used by the peer.
     * 
     * @return the {@link StructuredOverlay} which is used by the peer.
     */
    public StructuredOverlay getStructuredOverlay() {
        return this.structuredOverlay;
    }

    /**
     * Returns the stub associated to the current peer.
     * 
     * @return the stub associated to the current peer.
     */
    public Peer getStub() {
        return this.stub;
    }

    /**
     * Returns the type of overlay for the current peer.
     * 
     * @return the type of overlay for the current peer.
     */
    public OverlayType getType() {
        return this.structuredOverlay.getType();
    }

    /**
     * Indicates whether the join operation has already be performed or not.
     * 
     * @return <code>true</code> if the peer is activated, <code>false</code>
     *         otherwise.
     */
    public boolean isActivated() {
        return this.activated;
    }

    public void setActivated(boolean activated) {
        this.activated = activated;
    }

    /**
     * Forces the current peer to join an existing network by using the
     * specified peer reference.
     * 
     * @param remotePeer
     *            the peer used as entry point.
     * @return Returns <code>true</code> if the operation has succeeded,
     *         <code>false</code> otherwise.
     * @throws NetworkAlreadyJoinedException
     *             if the current peer has already join a network.
     */
    public boolean join(Peer remotePeer) throws NetworkAlreadyJoinedException {
        if (this.activated) {
            throw new NetworkAlreadyJoinedException();
        }

        if (this.structuredOverlay.join(remotePeer)) {
            this.activated = true;
            return true;
        }

        return false;
    }

    /**
     * Forces the current peer to leave the network it has joined.
     * 
     * @return Returns <code>true</code> if the operation has succeeded,
     *         <code>false</code> otherwise.
     * @throws NetworkNotJoinedException
     *             if the current peer try leave without having joined a
     *             network.
     */
    public boolean leave() throws NetworkNotJoinedException {
        if (!this.activated) {
            throw new NetworkNotJoinedException();
        }

        if (this.structuredOverlay.leave()) {
            this.activated = false;
            return true;
        }

        return false;
    }

    public void update() {
        this.structuredOverlay.update();
    }

    /**
     * Notify the peer which execute this command that the specified peer has
     * ended to leave.
     * 
     * @param peerHavingTerminatedLeftOperation
     *            the peer which has terminated the left operation.
     * @return <code>true</code> if the notification has succeeded,
     *         <code>false</code> otherwise.
     */
    public boolean notifyNeighborEndLeave(UUID peerHavingTerminatedLeftOperation) {
        return this.peersWhichAreLeaving.remove(peerHavingTerminatedLeftOperation);
    }

    /**
     * Notify the current peer which execute this command that the specified
     * peer is leaving.
     * 
     * @param peerLeaving
     *            the identifier of the peer which is leaving.
     * @return <code>true</code> if the notification has succeeded,
     *         <code>false</code> otherwise.
     */
    public boolean notifyNeighborStartLeave(UUID peerLeaving) {
        this.peersWhichAreLeaving.add(peerLeaving);
        return true;
    }

    /**
     * {@inheritDoc}
     */
    public void initActivity(Body body) {
        this.stub = (Peer) PAActiveObject.getStubOnThis();
        this.structuredOverlay.initActivity(body);
    }

    public void endActivity(Body body) {
        // TODO uncomment when leave operation works for CAN

        // if (this.activated) {
        // try {
        // this.leave();
        // } catch (StructuredP2PException e) {
        // e.printStackTrace();
        // }
        // }
    }

    /**
     * {@inheritDoc}
     */
    public void runActivity(Body body) {
        this.structuredOverlay.runActivity(body);
    }

    /**
     * Indicates if the peer is accessible.
     * 
     * @return <code>true</code> if object is accessible, otherwise throws a
     *         {@link BodyTerminatedException} exception.
     */
    public boolean isAlive() {
        return true;
    }

    /**
     * Sends a {@link Query} on the network from the current peer.
     * 
     * @param query
     *            the query to send.
     * @return the response in agreement with the type of query sent.
     * 
     * @exception DispatchException
     *                if a problem occurs while the query is dispatched.
     */
    public Response send(Query query) throws DispatchException {
        return this.structuredOverlay.getQueryManager().dispatch(query);
    }

    /**
     * Sends a {@link SynchronousMessage}.
     * 
     * @param msg
     *            the message to send.
     * @return the response in agreement with the type of query sent.
     */
    public SynchronousMessage<?> send(SynchronousMessage<?> msg) {
        return this.structuredOverlay.getQueryManager().process(msg);
    }

    public void route(SynchronousMessage<?> msg) {
        this.structuredOverlay.getQueryManager().route(msg);
    }

    /**
     * Sends a {@link Operation} to a known {@link Peer}.
     * 
     * @param msg
     *            the message to send.
     * 
     * @return the response in agreement with the type of message sent.
     */
    public ResponseOperation send(Operation msg) {
        return msg.handle(this.structuredOverlay);
    }

    /**
     * Sets the overlay.
     * 
     * @param structuredOverlay
     *            the new overlay to set.
     */
    public void setStructuredOverlay(StructuredOverlay structuredOverlay) {
        this.structuredOverlay = structuredOverlay;
    }

    /**
     * {@inheritDoc}
     */
    public int compareTo(Object obj) {
        return this.structuredOverlay.compareTo(obj);
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        return this.structuredOverlay.toString();
    }

}
