package org.objectweb.proactive.extensions.structuredp2p.overlay;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import org.objectweb.proactive.extensions.structuredp2p.api.messages.Query;
import org.objectweb.proactive.extensions.structuredp2p.api.messages.Response;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.DispatchException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.PostProcessException;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.PreProcessException;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;
import org.objectweb.proactive.extensions.structuredp2p.messages.response.AbstractResponseMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * QueryManager manages queries to route.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: QueryManager.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public abstract class QueryManager implements Serializable {

    private static final long serialVersionUID = 1L;

    private static final transient Logger logger = 
        LoggerFactory.getLogger(QueryManager.class);

    protected StructuredOverlay overlay;

    private Set<UUID> synchronousMessagesIdentifiersMet = 
        Collections.synchronizedSet(new HashSet<UUID>());

    private Map<UUID, SynchronousMessageEntry> synchronousMessagesReceived = 
        new ConcurrentHashMap<UUID, SynchronousMessageEntry>();

    public QueryManager() {

    }

    public abstract SynchronousMessageEntry mergeAndSetSynchronousMessage(SynchronousMessage<?> msg);

    public abstract SynchronousMessageEntry setSynchronousMessage(SynchronousMessage<?> msg);

    /**
     * Method called just before the {@link #process(SynchronousMessage)} method
     * in order to convert the specified query which is from the public API to
     * the private API but also to perform some operation on the specified query
     * which will be processed.
     * 
     * @param query
     *            the query to manipulate.
     * @return a query from the private API.
     * @throws PreProcessException
     *             if an error occurs during the pre-processing operation.
     */
    protected abstract SynchronousMessage<?> preProcess(Query query) 
                                                throws PreProcessException;

    
    /**
     * Method called just after the {@link #process(SynchronousMessage)} method
     * in order to convert the specified response which is from the private API
     * to the public API but also to perform some operation on the specified
     * response.
     * 
     * @param initialQuery
     *            the initial query from the public API.
     * @param response
     *            the response generated by the process operation.
     * @return a response from the public API.
     * @throws PostProcessException
     *             if an error occurs during the post-processing operation.
     */
    protected abstract Response postProcess(
                                    Query initialQuery, 
                                    AbstractResponseMessage<?> response) 
                                                throws PostProcessException;

    /**
     * Dispatches the query on the overlay.It consists in three steps: to
     * pre-process the query, to process it, and to post-process the response
     * returned by the process operation.
     * 
     * @param query
     *            the {@link Query} to dispatch.
     * @return a {@link Response} associated to the query type.
     */
    public Response dispatch(Query query) throws DispatchException {
        return this.postProcess(query, this.process(this.preProcess(query)));
    }

    /**
     * The process algorithm consist to send the specified message and to put a
     * synchronization point in order to wait the response.
     * 
     * @param msg
     *            the message to send on the network.
     * @return a response according to the type of message sent.
     */
    public AbstractResponseMessage<?> process(SynchronousMessage<?> msg) {
        if (logger.isDebugEnabled()) {
            logger.debug("QueryManager.process({}) from {}", msg.getUUID(), this.overlay);
        }

        this.processInitialSent(msg);
        this.waitForSynchronization(msg);
        return this.processFinalReception(msg);
    }

    protected void processInitialSent(SynchronousMessage<?> msg) {
        this.addSynchronousMessageIdentifierMet(msg.getUUID());
        msg.route(this.overlay);
    }

    protected void waitForSynchronization(SynchronousMessage<?> msg) {
        if (logger.isDebugEnabled()) {
            StringBuffer buf = new StringBuffer();
            buf.append("Waiting for ");
            buf.append(this.synchronousMessagesReceived.get(
                        msg.getUUID()).getNumberOfResponseWaiting());
            buf.append(" response(s) with uuid=");
            buf.append(msg.getUUID());
            buf.append(" on ");
            buf.append(this.overlay);
            QueryManager.logger.debug(buf.toString());
        }
        
        synchronized (this.synchronousMessagesReceived) {
            while (this.synchronousMessagesReceived.get(msg.getUUID()).getStatus() 
                        != SynchronousMessageEntry.Status.ALL_RESPONSES_RECEIVED) {
                try {
                    this.synchronousMessagesReceived.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    protected AbstractResponseMessage<?> processFinalReception(SynchronousMessage<?> msg) {
        AbstractResponseMessage<?> response = 
        	(AbstractResponseMessage<?>) this.synchronousMessagesReceived
                .remove(msg.getUUID()).getSynchronousMessage();
        response.setDeliveryTime();

        if (logger.isDebugEnabled()) {
            QueryManager.logger.debug(
                    "Final response received for uuid={} on {}", 
                    response.getUUID(), 
                    this.overlay);
        }

        return response;
    }

    public void putFinalResponseAndNotifyInitialSender(SynchronousMessage<?> response) {
        SynchronousMessageEntry entry = this.overlay.getSynchronousMessagesReceived().get(
                response.getUUID());
        entry.incrementNbResponsesReceivedBy(1);
        entry.setSynchronousMessage(response);

        synchronized (this.overlay.getSynchronousMessagesReceived()) {
            this.overlay.getSynchronousMessagesReceived().notifyAll();
        }
    }

    public Map<UUID, SynchronousMessageEntry> getSynchronousMessagesReceived() {
        return this.synchronousMessagesReceived;
    }

    public Set<UUID> getSynchronousMessagesIdentifiersMet() {
        return this.synchronousMessagesIdentifiersMet;
    }

    public void addSynchronousMessageIdentifierMet(UUID uuid) {
        synchronized (this.synchronousMessagesIdentifiersMet) {
            this.synchronousMessagesIdentifiersMet.add(uuid);
        }
    }

    public void route(SynchronousMessage<?> msg) {
        msg.route(this.overlay);
    }

    public void setOverlay(StructuredOverlay overlay) {
        this.overlay = overlay;
    }

}
