package org.objectweb.proactive.extensions.structuredp2p.overlay;

import java.io.Serializable;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.objectweb.proactive.Body;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.StructuredP2PException;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;

/**
 * Defines an abstract class that all structured peer-to-peer protocols must
 * extend.
 * 
 * @author Alexandre Trovato
 * @author Fanny Kilanga
 * @author Laurent Pellegrino
 * 
 * @version $Id: StructuredOverlay.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public abstract class StructuredOverlay implements Serializable {

    private static final long serialVersionUID = 1L;

    private UUID identifier;

    private QueryManager queryManager;

    private Peer localPeer;

    public StructuredOverlay() {

    }

    /**
     * Constructor.
     * 
     * @param peer
     *            the peer which is associated to the overlay.
     * @param queryManager
     *            the query manager to associate to the overlay.
     */
    public StructuredOverlay(Peer peer, QueryManager queryManager) {
        this(queryManager);
        this.localPeer = peer;
    }

    public StructuredOverlay(QueryManager queryManager) {
        this.identifier = UUID.randomUUID();
        this.queryManager = queryManager;
        this.queryManager.setOverlay(this);
    }

    /**
     * Adds a peer to the network.
     * 
     * @param remotePeer
     *            the peer (entry point) which is used in order to join the
     *            network.
     * @return <code>true</code> if join operation has succeeded,
     *         <code>false</code> otherwise.
     * @throws StructuredP2PException
     */
    public abstract boolean join(Peer remotePeer);

    public abstract boolean leave();

    public abstract int compareTo(Object obj);

    public abstract boolean equals(Object obj);

    public abstract String toString();

    public abstract void initActivity(Body body);

    public abstract void runActivity(Body body);

    public abstract OverlayType getType();

    public abstract int getUpdateIntervalTime();

    public abstract void update();

    public UUID getIdentifier() {
        return this.identifier;
    }

    /**
     * Returns the current peer that use this overlay.
     * 
     * @return the current peer that use this overlay.
     */
    public Peer getLocalPeer() {
        return this.localPeer;
    }

    /**
     * Returns the stub associated to the local peer.
     * 
     * @return the stub associated to the local peer.
     */
    public Peer getRemotePeer() {
        return this.localPeer.getStub();
    }

    public QueryManager getQueryManager() {
        return this.queryManager;
    }

    public Map<UUID, SynchronousMessageEntry> getSynchronousMessagesReceived() {
        return this.queryManager.getSynchronousMessagesReceived();
    }

    public Set<UUID> getSynchronousMessagesIdentifiersMet() {
        return this.queryManager.getSynchronousMessagesIdentifiersMet();
    }

    public void addSynchronousMessageIdentifier(UUID uuid) {
        this.queryManager.addSynchronousMessageIdentifierMet(uuid);
    }

    public SynchronousMessageEntry mergeAndSetSynchronousMessage(SynchronousMessage<?> msg) {
        return this.queryManager.mergeAndSetSynchronousMessage(msg);
    }

    public void setLocalPeer(Peer localPeer) {
        this.localPeer = localPeer;
    }

    public void route(SynchronousMessage<?> msg) {
        this.queryManager.route(msg);
    }

    public abstract String dump();

}
