package org.objectweb.proactive.extensions.structuredp2p.overlay.can;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.objectweb.proactive.Body;
import org.objectweb.proactive.Service;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.body.request.Request;
import org.objectweb.proactive.core.util.ProActiveRandom;
import org.objectweb.proactive.core.util.converter.MakeDeepCopy;
import org.objectweb.proactive.extensions.structuredp2p.operations.ResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.BooleanResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.EmptyResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.Operation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.AddAndUpdateNeighborOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.JoinOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.JoinResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.LeaveOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.MergeOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.RemoveNeighborOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.can.UpdateZoneCacheOperation;
import org.objectweb.proactive.extensions.structuredp2p.overlay.OverlayType;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.QueryManager;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * AbstractCANOverlay is an implementation of Content-Addressable Network (CAN)
 * protocol. Each peer manages neighbors from a {@link NeighborTable} and is
 * composed of a {@link Zone} which indicates the space associated to resources
 * to manage.
 * 
 * @author Alexandre Trovato
 * @author Fanny Kilanga
 * @author Laurent Pellegrino
 * 
 * @version $Id: AbstractCANOverlay.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public abstract class AbstractCANOverlay extends StructuredOverlay {

    private static final long serialVersionUID = 1L;

    private static final Logger logger = LoggerFactory.getLogger(AbstractCANOverlay.class);

    public static final int NB_DIMENSIONS = 3;

    public static final int UPDATE_TIMEOUT = 1000;

    private NeighborTable neighborTable = new NeighborTable();

    private LinkedList<SplitEntry> splitHistory = new LinkedList<SplitEntry>();

    private Zone zone = new Zone();

    public AbstractCANOverlay() {
        super();
    }

    /**
     * Constructor.
     * 
     * @param localPeer
     *            the local peer reference.
     * 
     * @param queryManager
     *            the {@link QueryManager} to use.
     */
    public AbstractCANOverlay(Peer localPeer, QueryManager queryManager) {
        super(localPeer, queryManager);
    }

    public AbstractCANOverlay(QueryManager queryManager) {
        super(queryManager);
    }

    protected abstract void affectDataReceived(JoinResponseOperation msg);

    protected abstract void mergeDataReceived(MergeOperation msg);

    protected abstract Object retrieveAllData();

    /**
     * Add a split entry to the split history.
     * 
     * @param dimension
     *            the last dimension used to split.
     * @param direction
     *            the last direction used by the split.
     * @return <code>true</code> if the save has succeeded, <code>false</code>
     *         otherwise.
     */
    private boolean saveSplit(int dimension, int direction) {
        return this.splitHistory.add(new SplitEntry(dimension, direction));
    }

    /**
     * Check neighbors list for all dimensions in order to see if a peer is
     * always a neighbor, if not, it is removed from its neighbors.
     */
    private void removeOutdatedNeighbors() {
        this.removeOutdatedNeighbors(-1);
    }

    /**
     * Check neighbors list following a dimension in order to see if a peer is
     * always a neighbor, if it is, the neighbor is updated with the current
     * zone, else, it is removed from its neighbors.
     * 
     * @param dimensionToSkip
     *            the dimension to not check.
     */
    private void removeOutdatedNeighbors(int dimensionToSkip) {
        List<UUID> neighborsToRemove = new ArrayList<UUID>();

        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            if (dim != dimensionToSkip) {
                for (int direction = 0; direction < 2; direction++) {
                    for (NeighborEntry entry : this.neighborTable.get(dim, direction).values()) {
                        if (this.zone.getAbutsDimensionWith(entry.getZone()) == -1) {
                            neighborsToRemove.add(entry.getIdentifier());
                        }
                    }
                }
            }
        }

        for (UUID identifier : neighborsToRemove) {
            this.neighborTable.remove(identifier);
        }
    }

    /**
     * Indicates if the specified coordinate is in the managed zone.
     * 
     * @param coordinate
     *            the coordinate to check.
     * 
     * @return <code>true</code> if the coordinates are in the zone,
     *         <code>false</code> otherwise.
     */
    public boolean contains(CoordinateValue[] coordinate) {
        return this.zone.contains(coordinate);
    }

    public boolean contains(List<CoordinateValue> coordinate) {
        assert coordinate.size() == AbstractCANOverlay.NB_DIMENSIONS;
        return this.contains(coordinate.toArray(new CoordinateValue[] {}));
    }

    /**
     * Indicates if the dimension index of the current zone contains the
     * specified coordinate.
     * 
     * @param dimension
     *            the dimension index used for the check.
     * 
     * @param coordinate
     *            the coordinate to check.
     * 
     * @return <code>0</code> if the coordinate is contained by the zone on the
     *         specified axe, <code>-1</code> if the coordinate is smaller than
     *         the line which is managed by the specified dimension,
     *         <code>1</code> otherwise.
     */
    public int contains(int dimension, CoordinateValue coordinate) {
        return this.getZone().contains(dimension, coordinate);
    }

    /**
     * Returns the neighbor table for the current zone managed.
     * 
     * @return the neighbors of the managed zone.
     */
    public NeighborTable getNeighborTable() {
        return this.neighborTable;
    }

    /**
     * Gets a random dimension number. The minimum dimension number is
     * <code>0</code>. The maximum dimension number is defined by
     * {@link AbstractCANOverlay#NB_DIMENSIONS}.
     * 
     * @return the random dimension number.
     */
    private int getRandomDimension() {
        return ProActiveRandom.nextInt(AbstractCANOverlay.NB_DIMENSIONS);
    }

    /**
     * Gets a random direction number.
     * 
     * @return the random direction number.
     */
    private int getRandomDirection() {
        return ProActiveRandom.nextInt(2);
    }

    /**
     * Returns the history of the splits.
     * 
     * @return the history of the splits.
     */
    public List<SplitEntry> getSplitHistory() {
        return this.splitHistory;
    }

    /**
     * Returns the zone which is managed by the overlay.
     * 
     * @return the zone which is managed by the overlay.
     */
    public Zone getZone() {
        return this.zone;
    }

    protected abstract Object transfertData(int directionInv, Zone[] newZones);

    public String dump() {
        StringBuilder buf = new StringBuilder();
        buf.append("Peer managing ");
        buf.append(this);
        buf.append(" with ID=");
        buf.append(this.getIdentifier());
        buf.append(" has neighbor(s):\n");

        for (NeighborEntry neighbor : this.neighborTable) {
            buf.append("  - ");
            buf.append(neighbor.getZone());
            buf.append(", ID=");
            buf.append(neighbor.getIdentifier());
            buf.append("\n");
        }
        return buf.toString();
    }

    /**
     * Handles the specified {@link JoinOperation}.
     * 
     * @param msg
     *            the message to handle.
     * @return a response associated to the initial message after having handled
     *         it.
     */
    @SuppressWarnings("unchecked")
    public JoinResponseOperation handleJoinMessage(JoinOperation msg) {
        int dimension = this.getRandomDimension();
        int direction = this.getRandomDirection();
        int directionInv = AbstractCANOverlay.getOppositeDirection(direction);

        // Gets the next dimension to split onto
        if (!this.splitHistory.isEmpty()) {
            dimension = AbstractCANOverlay.getNextDimension(
                    this.splitHistory.removeLast().getDimension());
        }

        // Splits zones
        Zone[] newZones = null;
        try {
            newZones = this.getZone().split(dimension);
        } catch (ZoneException e) {
            e.printStackTrace();
        }

        // Updates the zone of the peer which is already on the network
        // (this represents the old peer).
        this.setZone(newZones[direction]);
        this.saveSplit(dimension, direction);

        Object dataResult = this.transfertData(directionInv, newZones);

        // Neighbors affected for the new peer which want to join the network
        NeighborTable neighborsOfThePeerWhichJoin = new NeighborTable();
        neighborsOfThePeerWhichJoin.addAll(this.neighborTable);
        neighborsOfThePeerWhichJoin.removeAll(dimension, direction);
        neighborsOfThePeerWhichJoin.add(new NeighborEntry(this.getIdentifier(), 
                    this.getRemotePeer(), this.getZone()), dimension, direction);

        // Update the neighbors of the peer which is already on the network
        for (NeighborEntry entryToNotify : 
            this.neighborTable.get(dimension, directionInv).values()) {
            PAFuture.waitFor(entryToNotify.getStub().send(
                    new RemoveNeighborOperation(
                            this.getIdentifier(), dimension, direction)));
        }
        this.neighborTable.removeAll(dimension, directionInv);
        this.removeOutdatedNeighbors();

        LinkedList<SplitEntry> newHistory = null;
        try {
            newHistory = (LinkedList<SplitEntry>) MakeDeepCopy.WithObjectStream
                    .makeDeepCopy(this.splitHistory);
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

        // Updates zone cache for peers which are back the current peer because
        // the new peer which join hasn't them as neighbors and cannot update
        // them.
        for (NeighborEntry entry : this.neighborTable.get(dimension, direction).values()) {
            PAFuture.waitFor(entry.getStub().send(
                    new UpdateZoneCacheOperation(this.getIdentifier(), newZones[direction])));
        }

        return new JoinResponseOperation(dimension, directionInv, this.getIdentifier(),
                newZones[direction], newZones[directionInv], neighborsOfThePeerWhichJoin,
                dataResult, newHistory);
    }

    /**
     * {@inheritDoc}
     */
    public boolean join(Peer remotePeerOnNetwork) {
        JoinResponseOperation response = (JoinResponseOperation) PAFuture
                .getFutureValue(remotePeerOnNetwork.send(new JoinOperation(this.getRemotePeer())));
        logger.debug("Join message sent and handled");

        /* Actions on local peer : the peer which join an existing peer */
        this.setZone(response.getZoneAffected());
        this.splitHistory = response.getSplitHistory();
        this.saveSplit(response.getDimension(), response.getDirection());
        this.neighborTable = response.getNeighborsAffected();

        this.affectDataReceived(response);

        for (int dimension = 0; dimension < AbstractCANOverlay.NB_DIMENSIONS; dimension++) {
            for (int direction = 0; direction < 2; direction++) {
                for (NeighborEntry entry : this.neighborTable.get(dimension, direction).values()) {
                    PAFuture
                            .waitFor(entry.getStub().send(
                                    new AddAndUpdateNeighborOperation(new NeighborEntry(this
                                            .getIdentifier(),
                                            this.getRemotePeer(), this.zone),
                                            new NeighborEntry(response
                                                    .getPeerToUpdate(), null, response
                                                    .getZoneForUpdate()),
                                            dimension, AbstractCANOverlay
                                                    .getOppositeDirection(direction))));
                }
            }
        }

        this.removeOutdatedNeighbors();
        logger.debug("Neighbors updated");

        return true;
    }

    /**
     * {@inheritDoc}
     */
    public boolean leave() {
        /*
         * The current peer associated to this overlay is the only peer on the
         * network.
         */
        if (this.neighborTable.size() == 0) {
            return true;
        }

        SplitEntry lastSplitEntry = this.splitHistory.pop();
        int lastDimension = lastSplitEntry.getDimension();
        int lastDirection = lastSplitEntry.getDirection();

        Map<UUID, NeighborEntry> neighborsToMergeWith = this.neighborTable.get(lastDimension,
                AbstractCANOverlay.getOppositeDirection(lastDirection));

        /*
         * Notify all the neighbors the current peer is preparing to leave. That
         * force the neighbors to wait the current peer has finished to leave
         * before to handle new requests.
         */
        for (NeighborEntry entry : this.getNeighborTable()) {
            entry.getStub().notifyNeighborStartLeave(this.getIdentifier());
        }

        /*
         * Terminate the current body in order to notify all the neighbors that
         * they can't send message to the current remote peer. If they try they
         * will receive a BlockingRequestReceiverException : this exception is
         * used by the tracker in order to detect a peer which is preparing to
         * leave when it have to add a peer on the network.
         */

        // TODO BlockingRequestReceiver
        // ((BlockingRequestReceiver) ((MigratableBody)
        // super.getLocalPeer().getBody())
        // .getRequestReceiver()).blockReception();

        switch (neighborsToMergeWith.size()) {
            case 0:
                /* We are alone on this pitiless world : nothing to do */
                break;
            case 1:
                neighborsToMergeWith.values().iterator().next().getStub().send(
                        new MergeOperation(lastDimension, lastDirection, this.getIdentifier(), this
                                .getZone(), new NeighborTable(), this.retrieveAllData()));
                break;
            default:
                Zone zoneToSplit = this.getZone();
                Zone[] newZones = null;

                /*
                 * For the last dimension and direction of the split, we split
                 * N-1 times, where N is the number of neighbors in the last
                 * dimension and last reverse direction from the current peer.
                 */
                for (NeighborEntry entry : neighborsToMergeWith.values()) {
                    try {
                        newZones = zoneToSplit.split(
                                        AbstractCANOverlay.getNextDimension(lastDimension), 
                                        this.neighborTable.getNeighborEntry(
                                                entry.getIdentifier()).getZone().getCoordinateMax(
                                                        AbstractCANOverlay.getNextDimension(lastDimension)));
                    } catch (ZoneException e) {
                        e.printStackTrace();
                    }

                    NeighborTable neighborsOfCurrentNeighbor = new NeighborTable();
                    neighborsOfCurrentNeighbor.addAll(this.neighborTable);

                    zoneToSplit = newZones[1];

                    /*
                     * Merge the new zones obtained with the suitable neighbors.
                     */
                    // FIXME The given Data are not good : we give all the
                    // resources to each
                    // neighbors to merge with
                    entry.getStub().send(
                                    new MergeOperation(lastDimension,
                                            lastDirection, this.getIdentifier(), newZones[0],
                                            neighborsOfCurrentNeighbor,
                                            this.retrieveAllData()));
                }
                break;
        }

        // this.neighborsDataStructure.removeAll(lastDimension,
        // CANOverlay.getOppositeDirection(lastDirection));

        /*
         * Send LeaveOperation in order to update the neighbors list.
         */
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                for (NeighborEntry entry : this.neighborTable.get(dim, direction).values()) {
                    if (!neighborsToMergeWith.containsKey(entry.getIdentifier())) {
                        entry.getStub().send(
                                new LeaveOperation(this.getIdentifier(), neighborsToMergeWith
                                        .values(), dim, AbstractCANOverlay
                                        .getOppositeDirection(direction)));
                    }
                }
            }
        }

        // TODO BlockingRequestReceiver
        // ((BlockingRequestReceiver) ((MigratableBody)
        // super.getLocalPeer().getBody())
        // .getRequestReceiver()).acceptReception();

        /*
         * Notify all the old neighbors of the peer which is leaving that it has
         * terminated the leave operation.
         */
        for (NeighborEntry entry : this.getNeighborTable()) {
            entry.getStub().notifyNeighborEndLeave(this.getIdentifier());
        }
        for (NeighborEntry entry : neighborsToMergeWith.values()) {
            entry.getStub().notifyNeighborEndLeave(this.getIdentifier());
        }

        /*
         * Set all neighbors reference to null.
         */
        this.neighborTable.removeAll();

        return true;
    }

    /**
     * Send a {@link Operation} to a list {@link Peer}.
     * 
     * @param remotePeers
     *            the list of peers to which to send the message.
     * @param msg
     *            the message to send.
     * 
     * @return the list of responses in agreement with the type of message sent.
     * @throws Exception
     *             this exception appears when a message cannot be send to a
     *             peer.
     */
    public List<ResponseOperation> sendTo(List<Peer> remotePeers, Operation msg)
            throws Exception {
        List<ResponseOperation> responses = new ArrayList<ResponseOperation>(
                remotePeers.size());

        for (Peer remotePeer : remotePeers) {
            responses.add(remotePeer.send(msg));
        }

        return responses;
    }

    /**
     * Send a {@link Operation} to a peers that are in the specified
     * {@link NeighborTable}.
     * 
     * @param table
     *            the peers managed by the data structure to which to send the
     *            message.
     * @param msg
     *            the message to send.
     * 
     * @return the list of responses in agreement with the type of message sent.
     * @throws Exception
     *             this exception appears when a message cannot be send to a
     *             peer.
     */
    public List<ResponseOperation> sendTo(NeighborTable table,
            Operation msg) throws Exception {
        List<ResponseOperation> responses = new ArrayList<ResponseOperation>();

        for (int dimension = 0; dimension < AbstractCANOverlay.NB_DIMENSIONS; dimension++) {
            for (int direction = 0; direction < 2; direction++) {
                for (NeighborEntry entry : table.get(dimension, direction).values()) {
                    responses.add(entry.getStub().send(msg));
                }
            }
        }

        return responses;
    }

    /**
     * Sets the new split history.
     * 
     * @param history
     *            the new split history to set.
     */
    public void setHistory(LinkedList<SplitEntry> history) {
        this.splitHistory = history;
    }

    /**
     * Sets the new zone covered.
     * 
     * @param zone
     *            the new zone covered.
     */
    public void setZone(Zone zone) {
        this.zone = zone;
    }

    public OverlayType getType() {
        return OverlayType.CAN;
    }

    public void dumpNeighbors() {
        logger.debug("Peer managing " + this.zone);
        NeighborTable neighborTable = this.getNeighborTable();

        for (int dimension = 0; dimension < AbstractCANOverlay.NB_DIMENSIONS; dimension++) {
            for (int direction = 0; direction < 2; direction++) {
                for (NeighborEntry entry : neighborTable.get(dimension, direction).values()) {
                    logger.debug("  * " + entry.getZone() + ", dimension=" + dimension
                            + ", direction="
                            + direction);
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        return this.zone.toString();
    }

    /**
     * {@inheritDoc}
     */
    public void update() {
        // TODO
    }

    /**
     * Returns the timeout in milliseconds to wait before to check the overlay.
     * 
     * @return the timeout in milliseconds to wait before to check the overlay.
     */
    public int getUpdateIntervalTime() {
        return AbstractCANOverlay.UPDATE_TIMEOUT;
    }

    /**
     * Returns the next dimension following the specified dimension.
     * 
     * @param dimension
     *            the specified dimension.
     * @return the next dimension following the specified dimension.
     */
    public static int getNextDimension(int dimension) {
        return (dimension + 1) % AbstractCANOverlay.NB_DIMENSIONS;
    }

    /**
     * Returns the opposite direction of the specified direction.
     * 
     * @param direction
     *            the specified direction.
     * 
     * @return the opposite direction.
     */
    public static int getOppositeDirection(int direction) {
        return (direction + 1) % 2;
    }

    /**
     * Returns the previous dimension following the specified dimension.
     * 
     * @param dimension
     *            the specified dimension.
     * @return the previous dimension following the specified dimension.
     */
    public static int getPreviousDimension(int dimension) {
        int dim = dimension - 1;
        if (dim < 0) {
            dim = AbstractCANOverlay.NB_DIMENSIONS;
        }
        return dim;
    }

    /*
     * NeighborTable shortcuts
     */

    /**
     * Indicates if the data structure contains the specified {@link Peer}
     * identifier as neighbor.
     * 
     * @param peerID
     *            the peer identifier to look for.
     * @return <code>true</code> if the {@link NeighborTable} contains the peer
     *         identifier as neighbor, <code>false</code> otherwise.
     */
    public boolean hasNeighbor(UUID peerID) {
        return this.neighborTable.contains(peerID);
    }

    /*
     * Messages handling
     */

    public EmptyResponseOperation handleAddAndUpdateNeighborMessage(AddAndUpdateNeighborOperation msg) {
        int dimension = msg.getDimension();
        int direction = msg.getDirection();
        this.neighborTable.add(msg.getNeighborEntryToAdd(), dimension, direction);

        NeighborEntry entry = this.neighborTable.getNeighborEntry(
                msg.getNeighborEntryToUpdate().getIdentifier());

        if (entry != null) {
            entry.setZone(msg.getNeighborEntryToUpdate().getZone());
            // this.neighborTable.add(entry, dimension, direction);
        }
        if (logger.isDebugEnabled()) {
            logger.debug("Zone from peer ID '" + msg.getNeighborEntryToUpdate().getIdentifier()
                    + "'" + " has" + (entry == null ? "n't" : "") + " been updated to "
                    + msg.getNeighborEntryToUpdate().getZone()
                    + " for peer " + this + " which have a neighbor with the previous identifier.");
        }

        return new EmptyResponseOperation();
    }

    public EmptyResponseOperation handleLeaveMessage(LeaveOperation msg) {
        this.neighborTable.remove(msg.getPeerHavingLeft());

        for (NeighborEntry entry : msg.getNeighborsToMergeWith()) {
            if (!this.getNeighborTable().contains(entry.getIdentifier())) {
                this.neighborTable.add(entry, msg.getDimension(), msg.getDirection());
            }
        }
        this.removeOutdatedNeighbors();

        return new EmptyResponseOperation();
    }

    public EmptyResponseOperation handleMergeMessage(MergeOperation msg) {
        try {
            this.setZone(this.zone.merge(msg.getZoneToReallocate()));
            this.mergeDataReceived(msg);
        } catch (ZoneException e) {
            e.printStackTrace();
        }

        int dimension = msg.getDimension();
        int direction = msg.getDirection();

        int index = -1, t = 0;
        for (SplitEntry entry : this.splitHistory) {
            if (entry.getDimension() == dimension && entry.getDirection() == direction) {
                index = t;
            }
            t++;
        }

        if (index >= 0) {
            this.splitHistory.remove(index);
        }

        this.neighborTable.remove(msg.getPeerToMergeWith(), dimension, direction);
        this.neighborTable.addAll(msg.getNeighborsToReallocate());
        this.removeOutdatedNeighbors();

        return new EmptyResponseOperation();
    }

    public BooleanResponseOperation handleRemoveNeighborMessage(RemoveNeighborOperation msg) {
        if (msg.getDirection() == NeighborTable.ANY_DIRECTION) {
            return new BooleanResponseOperation(this.neighborTable.remove(msg.getPeerIdentifier()));
        } else {
            return new BooleanResponseOperation(this.neighborTable.remove(msg.getPeerIdentifier(),
                    msg.getDimension(), msg.getDirection()));
        }
    }

    /*
     * Specific implementation of ProActive concepts.
     */
    public void initActivity(Body body) {
        PAActiveObject.setImmediateService("send");
        PAActiveObject.setImmediateService("route");
    }

    public void runActivity(Body body) {
        Service service = new Service(body);

        while (body.isActive()) {
            Request request = null;
            if (super.getLocalPeer().getPeersWhichArePreparingToLeave().size() > 0) {
                /*
                 * boolean receiveFromLeaver = false; request =
                 * service.blockingRemoveOldest();
                 * 
                 * for (Peer peer : this.peersWhichAreLeaving) { if
                 * (peer.getBody().getID().equals(request.getSender().getID()))
                 * { System.out.println("OK FIND !!!"); receiveFromLeaver =
                 * true; service.serve(request); break; } }
                 * 
                 * if (!receiveFromLeaver) { System.out.println("NT FIND");
                 */
                service.serveOldest("notifyNeighborEndLeave");
                // }
            } else {
                request = service.blockingRemoveOldest(this.getUpdateIntervalTime());

                if (request == null) {
                    this.update();
                } else {
                    service.serve(request);
                }
            }
        }
    }

    public int compareTo(Object obj) {
        return this.equals(obj) ? 0 : -1;
    }

    public boolean equals(Object obj) {
        if (super.getLocalPeer().getType() != ((Peer) obj).getType()) {
            return false;
        }

        AbstractCANOverlay objOverlay = (AbstractCANOverlay) (((Peer) obj).getStructuredOverlay());

        return (this.getZone().equals(objOverlay.getZone()) && this.getNeighborTable().equals(
                objOverlay.getNeighborTable()));
    }

}
