package org.objectweb.proactive.extensions.structuredp2p.overlay.can;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import org.objectweb.proactive.core.util.ProActiveRandom;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;

/**
 * Data structure used in order to store the neighbors of a {@link Peer} of type
 * CAN where neighbors are in a specified dimension and direction.
 * 
 * @author Alexandre Trovato
 * @author Fanny Kilanga
 * @author Laurent Pellegrino
 * 
 * @version $Id: NeighborTable.java 5022 2010-08-14 16:14:45Z plaurent $
 */
@SuppressWarnings("unchecked")
public class NeighborTable implements Iterable<NeighborEntry>, Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * Any direction : inferior or superior.
     */
    public static final int ANY_DIRECTION = -1;

    /**
     * Inferior direction in comparison to a specified peer on a given
     * dimension.
     */
    public static final int INFERIOR_DIRECTION = 0;

    /**
     * Superior direction in comparison to a specified peer on a given
     * dimension.
     */
    public static final int SUPERIOR_DIRECTION = 1;

    /**
     * Contains neighbors categorized by dimension, direction. The neighbors are
     * in a two-dimensional array of {@link ConcurrentHashMap}. Each line corresponds
     * to a dimension and the number of columns is always equal to two.
     */
    private Map<UUID, NeighborEntry>[][] entries = new ConcurrentHashMap[AbstractCANOverlay.NB_DIMENSIONS][2];

    /**
     * Constructs a new NeighborTable.
     */
    public NeighborTable() {
        for (int i = 0; i < AbstractCANOverlay.NB_DIMENSIONS; i++) {
            this.entries[i][NeighborTable.INFERIOR_DIRECTION] = new ConcurrentHashMap<UUID, NeighborEntry>();
            // this.entries[i][NeighborTable.INFERIOR_DIRECTION] = Maps
            // .synchronizedBiMap(this.entries[i][NeighborTable.INFERIOR_DIRECTION]);
            this.entries[i][NeighborTable.SUPERIOR_DIRECTION] = new ConcurrentHashMap<UUID, NeighborEntry>();
            // this.entries[i][NeighborTable.SUPERIOR_DIRECTION] = Maps
            // .synchronizedBiMap(this.entries[i][NeighborTable.SUPERIOR_DIRECTION]);
        }
    }

    /**
     * Add a new neighbor at the specified <code>dimension</code>,
     * <code>direction</code>.
     * 
     * @param entry
     *            the {@link NeighborEntry} to add.
     * @param dimension
     *            the dimension index (must be in <code>0</code> and
     *            {@link AbstractCANOverlay#NB_DIMENSIONS - 1} include).
     * @param direction
     *            the direction ({@link #INFERIOR_DIRECTION} or
     *            {@link #SUPERIOR_DIRECTION}).
     */
    public void add(NeighborEntry entry, int dimension, int direction) {
        this.entries[dimension][direction].put(entry.getIdentifier(), entry);
    }

    /**
     * Add all the neighbors from the specified {@link NeighborTable}.
     * 
     * @param table
     *            the neighbors to add.
     */
    public void addAll(NeighborTable table) {
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                this.entries[dim][direction].putAll(table.get(dim, direction));
            }
        }
    }

    public Peer getRandomNeighbor(int dimension, int direction) {
        NeighborEntry[] neighbors = this.get(dimension, direction).values().toArray(
                new NeighborEntry[] {});
        return neighbors[ProActiveRandom.nextInt(neighbors.length)].getStub();
    }

    /**
     * Returns a neighbor (which is a {@link Peer}) randomly selected among
     * neighbors validating constraints for all dimensions less than the
     * specified dimension.
     * 
     * @param ownerZone
     *            the zone of the peer on which this method is called.
     * @param coordinateToReach
     *            the coordinate to reach.
     * @param dimension
     *            the dimension to look for.
     * @param direction
     *            the direction to look for
     * @return a neighbor randomly selected among neighbors validating
     *         constraints for all dimensions less than the specified dimension.
     */
    public Peer getRandomNeighborValidatingConstraints(Zone ownerZone,
            CoordinateValue[] coordinateToReach, int dimension, int direction) {

        List<NeighborEntry> neighborsValidatingConstraints = new ArrayList<NeighborEntry>();
        boolean validatesPreviousDimensions;

        for (NeighborEntry entry : this.entries[dimension][direction].values()) {
            validatesPreviousDimensions = true;

            for (int j = 0; j < dimension; j++) {
                if (coordinateToReach[j] != null
                        && entry.getZone().contains(j, coordinateToReach[j]) != 0) {
                    validatesPreviousDimensions = false;
                    break;
                }
            }

            if (validatesPreviousDimensions) {
                neighborsValidatingConstraints.add(entry);
            }
        }

        if (neighborsValidatingConstraints.size() == 0) {
            return null;
        } else {
            return neighborsValidatingConstraints.get(
                    ProActiveRandom.nextInt(neighborsValidatingConstraints.size())).getStub();
        }
    }

    /**
     * Returns all the neighbors of the data structure categorized by dimension
     * and direction.
     * 
     * @return all the neighbors of the data structure categorized by dimension
     *         and direction.
     */
    public Map<UUID, NeighborEntry>[][] getAll() {
        return this.entries;
    }

    /**
     * Returns all the neighbors on the specified dimension.
     * 
     * @param dimension
     *            the dimension to use (dimension start to <code>0</code> and
     *            max is defined by {@link AbstractCANOverlay#NB_DIMENSIONS} -
     *            1).
     * @return the neighbors on the specified dimension.
     */
    public Map<UUID, NeighborEntry>[] get(int dimension) {
        return this.entries[dimension];
    }

    /**
     * Returns all the neighbors on the specified <code>dimension</code> and
     * <code>direction</code> .
     * 
     * @param dimension
     *            the dimension to use (dimension start to <code>0</code> and
     *            max is defined by {@link AbstractCANOverlay#NB_DIMENSIONS} -
     *            1).
     * @param direction
     *            the direction ({@link #INFERIOR_DIRECTION} or
     *            {@link #SUPERIOR_DIRECTION}).
     * @return all the neighbors on the specified dimension and direction.
     */
    public Map<UUID, NeighborEntry> get(int dimension, int direction) {
        return this.entries[dimension][direction];
    }

    /**
     * Returns an entry from the specified {@link Peer} identifier if it is
     * found.
     * 
     * @param peerIdentifier
     *            the identifier used for the lookup.
     * @return the zone found or <code>null</code>.
     */
    public NeighborEntry getNeighborEntry(UUID peerIdentifier) {
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                if (this.entries[dim][direction].containsKey(peerIdentifier)) {
                    return (NeighborEntry) this.entries[dim][direction].get(peerIdentifier);
                }
            }
        }

        return null;
    }

    /**
     * Indicates if a {@link NeighborEntry} identified by the specified
     * {@link Peer} identifier is contained by this data structure or not.
     * 
     * @param peerIdentifier
     *            the identifier to use for checking.
     * @return <code>true</code> if the data structure contains the peer
     *         identifier, <code>false</code> otherwise.
     */
    public boolean contains(UUID peerIdentifier) {
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                if (this.entries[dim][direction].containsKey(peerIdentifier)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Indicates if the data structure contains the specified {@link Peer} as
     * neighbor at the specified <code>dimension</code> and
     * <code>direction</code>.
     * 
     * @param peerIdentifier
     *            the identifier to use for checking.
     * @param dimension
     *            the dimension.
     * @param direction
     *            the direction.
     * 
     * @return <code>true</code> if the data structure contains the peer as
     *         neighbor, <code>false</code> otherwise.
     */
    public boolean contains(UUID peerIdentifier, int dimension, int direction) {
        return this.entries[dimension][direction].containsKey(peerIdentifier);
    }

    /**
     * {@inheritDoc}
     */
    public Iterator<NeighborEntry> iterator() {
        Map<UUID, NeighborEntry> neighborEntries = new ConcurrentHashMap<UUID, NeighborEntry>();

        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                neighborEntries.putAll(this.entries[dim][direction]);
            }
        }

        return neighborEntries.values().iterator();
    }

    /**
     * Removes the {@link NeighborEntry} identified by the specified peer
     * identifier.
     * 
     * @param peerIdentifier
     *            the identifier to use for checking.
     * @return <code>true</code> if the neighbor has been removed,
     *         <code>false</code> otherwise.
     */
    public boolean remove(UUID peerIdentifier) {
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                if (this.entries[dim][direction].containsKey(peerIdentifier)) {
                    this.entries[dim][direction].remove(peerIdentifier);
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Removes the {@link NeighborEntry} identified by the specified peer
     * identifier on the given dimension and direction.
     * 
     * @param peerIdentifier
     *            the identifier to use for checking.
     * @param dimension
     *            the dimension index (must be in <code>0</code> and
     *            {@link AbstractCANOverlay#NB_DIMENSIONS - 1} include).
     * @param direction
     *            the direction ({@link #INFERIOR_DIRECTION} or
     *            {@link #SUPERIOR_DIRECTION}).
     * @return <code>true</code> if the neighbor has been removed,
     *         <code>false</code> otherwise.
     */
    public boolean remove(UUID peerIdentifier, int dimension, int direction) {
        if (this.entries[dimension][direction].containsKey(peerIdentifier)) {
            this.entries[dimension][direction].remove(peerIdentifier);
            return true;
        }

        return false;
    }

    /**
     * Clears the data structure.
     */
    public void removeAll() {
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                this.removeAll(dim, direction);
            }
        }
    }

    /**
     * Removes all neighbors on a specified dimension and direction.
     * 
     * @param dimension
     *            the dimension index (must be in <code>0</code> and
     *            {@link AbstractCANOverlay#NB_DIMENSIONS - 1} include).
     * @param direction
     *            the direction ({@link #INFERIOR_DIRECTION} or
     *            {@link #SUPERIOR_DIRECTION}).
     */
    public void removeAll(int dimension, int direction) {
        this.entries[dimension][direction].clear();
    }

    /**
     * Returns the number of neighbors the data structure manages.
     * 
     * @return the number of neighbors the data structure manages.
     */
    public int size() {
        int nbEntries = 0;
        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                nbEntries += this.entries[dim][direction].size();
            }
        }
        return nbEntries;
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();

        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            for (int direction = 0; direction < 2; direction++) {
                buf.append("[");
                int i = 0;
                for (NeighborEntry entry : this.entries[dim][direction].values()) {
                    buf.append(entry.getZone());
                    if (i < this.entries[dim][direction].values().size() - 1) {
                        buf.append(",");
                    }
                    i++;
                }
                buf.append("]");

                if (direction == 0) {
                    buf.append("<--(dim ");
                    buf.append(dim);
                    buf.append(")-->");
                } else {
                    buf.append("\n");
                }
            }
        }
        return buf.toString();
    }

}
