package org.objectweb.proactive.extensions.structuredp2p.overlay.can;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.proactive.core.util.converter.MakeDeepCopy;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.LexicographicCoordinateValue;

/**
 * An zone defines a space (rectangle) which is completely logical and managed
 * by an {@link AbstractCANOverlay}. A zone is composed of two coordinates: a
 * minimum coordinate which corresponds to the lower left corner and a maximum
 * coordinates which corresponds to the upper right corner.
 * 
 * @author Alexandre Trovato
 * @author Fanny Kilanga
 * @author Laurent Pellegrino
 * 
 * @version $Id: Zone.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class Zone implements Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * The minimal value to manage (lower left corner).
     */
    public static final char LOWER_BOUND = '0'; // codePoint 0=48

    /**
     * The maximal value to manage (upper right corner).
     */
    public static final char UPPER_BOUND = '{'; // codePoint {=123

    /**
     * The type of coordinate to use.
     */
    public static final Class<? extends CoordinateValue> COORDINATE_TYPE = LexicographicCoordinateValue.class;

    /**
     * The maximum coordinate.
     */
    private CoordinateValue[] coordinateMax;

    /**
     * The minimum coordinate.
     */
    private CoordinateValue[] coordinateMin;

    /**
     * Constructs a new zone by using the {@link #LOWER_BOUND} and {@link #UPPER_BOUND} as coordinates.
     */
    public Zone() {
        CoordinateValue[] minCoordinate = new CoordinateValue[AbstractCANOverlay.NB_DIMENSIONS];
        CoordinateValue[] maxCoordinate = new CoordinateValue[AbstractCANOverlay.NB_DIMENSIONS];

        for (int i = 0; i < AbstractCANOverlay.NB_DIMENSIONS; i++) {
            CoordinateValue coordinateValue = null;
            try {
                coordinateValue = Zone.COORDINATE_TYPE.newInstance();
            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            }
            coordinateValue.setValue(new String(new char[] { Zone.LOWER_BOUND }));
            minCoordinate[i] = coordinateValue;

            try {
                coordinateValue = Zone.COORDINATE_TYPE.newInstance();
            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            }
            coordinateValue.setValue(new String(new char[] { Zone.UPPER_BOUND }));
            maxCoordinate[i] = coordinateValue;
        }

        this.coordinateMin = minCoordinate;
        this.coordinateMax = maxCoordinate;
    }

    /**
     * Constructs a new zone with the specified lower and upper bounds.
     * 
     * @param min
     *            the minimum coordinates.
     * @param max
     *            the maximum coordinates.
     * @throws ZoneException
     *             if a zone has the coordinate <code>min</code> and
     *             <code>max</code> equals in the same dimension.
     */
    public Zone(CoordinateValue[] min, CoordinateValue[] max) throws ZoneException {
        this.coordinateMin = min;
        this.coordinateMax = max;
    }

    /**
     * Check if the specified coordinate is in the zone managed.
     * 
     * @param coordinate
     *            the coordinate to check.
     * 
     * @return <code>true</code> if the coordinate is in the zone managed,
     *         <code>false</code> otherwise.
     */
    public boolean contains(CoordinateValue[] coordinate) {
        int i;
        boolean res = true;

        for (i = 0; i < AbstractCANOverlay.NB_DIMENSIONS; i++) {
            if (!(res &= (this.contains(i, coordinate[i]) == 0))) {
                return false;
            }
        }

        return res;
    }

    /**
     * Indicates if the zone contains the specified coordinate component
     * following the specified dimension.
     * 
     * @param dimension
     *            the dimension.
     * @param coordinate
     *            the coordinate to check.
     * @return <code>0</code> if the current coordinate is in the zone,
     *         <code>-1</code> if the coordinate is taller than the minimal
     *         coordinate of the zone and <code>1</code> if the coordinate is
     *         greater than the maximal coordinate of the zone and
     *         .((CANOverlay) this.firstPeer.getStructuredOverlay()).getZone()
     *         .getCoordinatesMin()
     */
    public int contains(int dimension, CoordinateValue coordinate) {
        if (coordinate.compareTo(this.getCoordinateMax(dimension)) >= 0) {
            return 1;
        } else if (coordinate.compareTo(this.getCoordinateMin(dimension)) < 0) {
            return -1;
        }

        return 0;
    }

    /**
     * {@inheritDoc}
     */
    public boolean equals(Object o) {
        if (!(o instanceof Zone)) {
            throw new IllegalArgumentException();
        }

        Zone zone = (Zone) o;

        int i;
        int nbDim = this.coordinateMax.length;

        for (i = 0; i < nbDim; i++) {
            if (!this.getCoordinateMax(i).equals(zone.getCoordinateMax(i))
                    || !this.getCoordinateMin(i).equals(zone.getCoordinateMin(i))) {
                return false;
            }
        }

        return true;
    }

    /**
     * Returns the dimension on which the current zone and the specified
     * <code>zone</code> are abuts.
     * 
     * @param zone
     *            the zone to use in order to find the dimension abuts.
     * @return the dimension on which the current zone and the specified
     *         <code>zone</code> are abuts or <code>-1</code> if they aren't
     *         abuts.
     */
    public int getAbutsDimensionWith(Zone zone) {
        int i;
        int nbDim = this.coordinateMax.length;

        for (i = 0; i < nbDim; i++) {
            if (this.isAbutsWith(zone, i)) {
                return i;
            }
        }

        return -1;
    }

    /**
     * Returns the maximum coordinate at a specified dimension that indicates
     * the zone which is managed by a peer.
     * 
     * @param dimension
     *            the dimension on which we recover the coordinates.
     * @return the maximum coordinate at a specified dimension that indicates
     *         the zone which is managed by a peer.
     */
    public CoordinateValue getCoordinateMax(int dimension) {
        return this.coordinateMax[dimension];
    }

    /**
     * Returns the minimum coordinate at a specified dimension that indicates
     * the zone which is managed by a peer.
     * 
     * @param dimension
     *            the dimension on which we recover the coordinates.
     * @return the minimum coordinate at the specified dimension that indicates
     *         the zone which is managed by a peer.
     */
    public CoordinateValue getCoordinateMin(int dimension) {
        return this.coordinateMin[dimension];
    }

    /**
     * Returns the coordinates of the zone.
     * 
     * @return the coordinates of the zone.
     */
    public CoordinateValue[][] getCoordinates() {
        return new CoordinateValue[][] { this.getCoordinateMin(), this.getCoordinateMax() };
    }

    /**
     * Returns the maximum coordinate that indicates the zone which is managed
     * by a peer.
     * 
     * @return the maximum coordinate that indicates the zone which is managed
     *         by a peer.
     */
    public CoordinateValue[] getCoordinateMax() {
        return this.coordinateMax;
    }

    /**
     * Returns the maximum coordinate that indicates the zone which is managed
     * by a peer.
     * 
     * @return the maximum coordinate that indicates the zone which is managed
     *         by a peer.
     */
    public List<CoordinateValue> getCoordinateMaxAsList() {
        List<CoordinateValue> list = new ArrayList<CoordinateValue>();
        for (CoordinateValue coord : this.coordinateMax) {
            list.add(coord);
        }
        return list;
    }

    /**
     * Returns the minimum coordinate that indicates the zone which is managed
     * by a peer.
     * 
     * @return the minimum coordinate that indicates the zone which is managed
     *         by a peer.
     */
    public CoordinateValue[] getCoordinateMin() {
        return this.coordinateMin;
    }

    /**
     * Returns the minimum coordinate that indicates the zone which is managed
     * by a peer.
     * 
     * @return the minimum coordinate that indicates the zone which is managed
     *         by a peer.
     */
    public List<CoordinateValue> getCoordinateMinAsList() {
        List<CoordinateValue> list = new ArrayList<CoordinateValue>();
        for (CoordinateValue coord : this.coordinateMin) {
            list.add(coord);
        }
        return list;
    }

    /**
     * Check the specified <code>zone</code> on the given <code>dimension</code>
     * is abuts with the current zone.
     * 
     * WARNING : if an zone is abuts with an another zone on the dimension 1 of
     * two dimension, the parameter dimension to specify is 0 because the two
     * zones share the same coordinate on the dimension 0 and not the dimension
     * 1.
     * 
     * @param zone
     *            the zone to check.
     * @param dimension
     *            the dimension to check.
     * @return <code>true</code> if zones are abuts, <code>false</code>
     *         otherwise.
     */
    public Boolean isAbutsWith(Zone zone, int dimension) {
        boolean dimResult = false;
        boolean borderResult = false;

        for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
            if (dim == dimension) {
                dimResult = this.getCoordinateMin(dim).equals(zone.getCoordinateMax(dim)) 
                         || this.getCoordinateMax(dim).equals(zone.getCoordinateMin(dim));
            } else {
                borderResult |= (CoordinateValue.isBetween(
                                    this.getCoordinateMin(dim), 
                                    zone.getCoordinateMin(dim), 
                                    zone.getCoordinateMax(dim)) || 
                                 CoordinateValue.isBetween(
                                    this.getCoordinateMax(dim), 
                                    zone.getCoordinateMin(dim), 
                                    zone.getCoordinateMax(dim))) || 
                                 CoordinateValue.isBetween(
                                    zone.getCoordinateMin(dim), 
                                    this.getCoordinateMin(dim), 
                                    this.getCoordinateMax(dim)) || 
                                 CoordinateValue.isBetween(
                                    zone.getCoordinateMax(dim), 
                                    this.getCoordinateMin(dim),
                                    this.getCoordinateMax(dim));
            }
        }

        return dimResult && borderResult;
    }

    /**
     * Merges the current zone with the specified <code>zone</code>.
     * 
     * @param zone
     *            the zone to merge with.
     * @return a new zone which is the merge between the current zone and the specified zone.
     * @throws ZoneException if zones to merge are not abuts.
     */
    public Zone merge(Zone zone) throws ZoneException {
        int abutsDimension = this.getAbutsDimensionWith(zone);

        if (abutsDimension == -1) {
            throw new ZoneException("Zones to merge are not abuts.");
        }

        CoordinateValue[] minCoord = null;
        CoordinateValue[] maxCoord = null;
        try {
            minCoord = (CoordinateValue[]) MakeDeepCopy.WithObjectStream
                        .makeDeepCopy(this.coordinateMin);
            maxCoord = (CoordinateValue[]) MakeDeepCopy.WithObjectStream
                        .makeDeepCopy(this.coordinateMax);
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
        minCoord[abutsDimension] = CoordinateValue.min(this.getCoordinateMin(abutsDimension), zone
                    .getCoordinateMin(abutsDimension));
        maxCoord[abutsDimension] = CoordinateValue.max(this.getCoordinateMax(abutsDimension), zone
                    .getCoordinateMax(abutsDimension));

        return new Zone(minCoord, maxCoord);
    }

    /**
     * Returns two zones representing the original one split into two following
     * a dimension.
     * 
     * @param dimension
     *            the dimension.
     * @return two zones representing the original one split into two following
     *         a dimension.
     * @throws ZoneException
     * @see Zone#split(int, CoordinateValue)
     */
    public Zone[] split(int dimension) throws ZoneException {
        return this.split(dimension, this.getCoordinateMin(dimension).getMiddleWith(
                this.getCoordinateMax(dimension)));
    }

    /**
     * Returns two zones representing the original one split into two following
     * the specified <code>dimension</code> at the specified
     * <code>coordinateComponent</code>.
     * 
     * @param dimension
     *            the dimension.
     * @param coordinateComponent
     *            the coordinate component.
     * @return two zones representing the original one split into two following
     *         the specified dimension at the specified coordinate.
     * @throws ZoneException
     */
    public Zone[] split(int dimension, CoordinateValue coordinateComponent) throws ZoneException {
        CoordinateValue[] coordinatesMaxCopy = null;
        CoordinateValue[] coordinatesMinCopy = null;

        try {
            coordinatesMaxCopy = (CoordinateValue[]) MakeDeepCopy.WithObjectStream
                    .makeDeepCopy(this
                            .getCoordinateMax());
            coordinatesMinCopy = (CoordinateValue[]) MakeDeepCopy.WithObjectStream
                    .makeDeepCopy(this
                            .getCoordinateMin());
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

        coordinatesMaxCopy[dimension] = coordinateComponent;
        coordinatesMinCopy[dimension] = coordinateComponent;

        return new Zone[] { 
                new Zone(this.getCoordinateMin(), coordinatesMaxCopy),
                new Zone(coordinatesMinCopy, this.getCoordinateMax()) };
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("(");

        for (int i = 0; i < this.coordinateMin.length; i++) {
            buf.append(this.coordinateMin[i].getValue());
            if (i != this.coordinateMin.length - 1) {
                buf.append(",");
            }
        }

        buf.append(") to (");

        for (int i = 0; i < this.coordinateMax.length; i++) {
            buf.append(this.coordinateMax[i].getValue());
            if (i != this.coordinateMax.length - 1) {
                buf.append(",");
            }
        }

        buf.append(")");

        return buf.toString();
    }

}
