package org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates;

import java.io.Serializable;

import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.AbstractCANOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.Zone;

/**
 * CoordinateValue represents one component of a coordinate. These coordinates
 * are used in {@link Zone} by {@link AbstractCANOverlay} in order to specify
 * the space which is managed by a {@link Peer} on a network.
 * 
 * @author Kilanga Fanny
 * @author Laurent Pellegrino
 * @author Trovato Alexandre
 * 
 * @version $Id: CoordinateValue.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public abstract class CoordinateValue implements Comparable<CoordinateValue>, Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * The String value associated to the coordinate.
     */
    private String value;

    /**
     * Constructor.
     */
    public CoordinateValue() {
    }

    /**
     * Constructor.
     */
    public CoordinateValue(String value) {
    	this.value = value;
    }

    /**
     * Compares two coordinates.
     * 
     * @param coordinate
     *            the value to compare with.
     * @return The value <code>0</code> if the argument coordinate is equal to
     *         this coordinate; a value less than <code>0</code> if this
     *         coordinate is less than the coordinate argument; and a value
     *         greater than <code>0</code> if this coordinate is greater than
     *         the coordinate argument.
     */
    public abstract int compareTo(CoordinateValue coordinate);

    /**
     * Returns the middle coordinate between the current coordinate and the
     * specified coordinate.
     * 
     * @param coordinate
     *            the specified coordinate.
     * 
     * @return the middle coordinate between the current coordinate and the
     *         specified coordinate.
     */
    public abstract CoordinateValue getMiddleWith(CoordinateValue coordinate);

    /**
     * Indicates if the specified coordinate <code>current</code> is between the
     * coordinate <code>a</code> and coordinate <code>b</code>.
     * 
     * @param a
     *            a coordinate.
     * 
     * @param b
     *            a coordinate.
     * 
     * @return <code>true</code> if the coordinate <code>current</code> is
     *         between <code>a</code> and <code>b</code>. <code>false</code>
     *         otherwise.
     */
    public static boolean isBetween(CoordinateValue current, CoordinateValue a, CoordinateValue b) {
        if (a.compareTo(b) < 0) {
            return (current.compareTo(a) >= 0) && (current.compareTo(b) < 0);
        } else if (a.compareTo(b) > 0) {
            return (current.compareTo(b) >= 0) && (current.compareTo(a) < 0);
        }

        return false;
    }

    /**
     * Returns the smallest coordinate among two specified coordinate.
     * 
     * @return the smallest coordinate among two specified coordinate.
     */
    public static CoordinateValue min(CoordinateValue coord1, CoordinateValue coord2) {
        if (coord1.compareTo(coord2) < 0) {
            return (coord1);
        } else {
            return (coord2);
        }
    }

    /**
     * Returns the greatest coordinate among two specified coordinate.
     * 
     * @return the greatest coordinate among two specified coordinate.
     */
    public static CoordinateValue max(CoordinateValue coord1, CoordinateValue coord2) {
        if (coord1.compareTo(coord2) > 0) {
            return (coord1);
        } else {
            return coord2;
        }
    }

    /**
     * Returns the value of the coordinate.
     * 
     * @return the value of the coordinate.
     */
    public String getValue() {
        return this.value;
    }

    /**
     * Sets the value.
     * 
     * @param value
     *            the new value to set.
     */
    public void setValue(String value) {
        this.value = value;
    }

    /**
     * {@inheritDoc}
     */
    public boolean equals(Object obj) {
        return this.value.equals(((CoordinateValue) obj).getValue());
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        return this.getValue();
    }
}
