package org.objectweb.proactive.extensions.structuredp2p.overlay.chord;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.UUID;

import org.objectweb.proactive.Body;
import org.objectweb.proactive.Service;
import org.objectweb.proactive.api.PAActiveObject;
import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.body.exceptions.BodyTerminatedException;
import org.objectweb.proactive.core.body.request.Request;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.ChordException;
import org.objectweb.proactive.extensions.structuredp2p.operations.BooleanResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.IsAliveOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.AddEntriesOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.AddEntryOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.EntryResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.FindSuccessorOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.FindSuccessorResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetEntryOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetKeyOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetKeyResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetPredecessorOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetPredecessorResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetSuccessorsFTOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetSuccessorsFTResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.NotifyPredecessorOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.RemoveEntryOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.SetPredecessorOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.SetSuccessorOperation;
import org.objectweb.proactive.extensions.structuredp2p.overlay.OverlayType;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * AbstractChordOverlay is an implementation of Chord protocol. Each chord peer
 * is composed of a predecessor and successor link but also of a
 * {@link FingerTable} in order to accelerate the lookup for resource. Each node
 * and resources are identified by a unique hashed identifier which is
 * represented by a {@link Key}.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: AbstractChordOverlay.java 3127 2010-01-06 15:59:42Z laurentp$
 */
public abstract class AbstractChordOverlay<T> extends StructuredOverlay {

    private static final long serialVersionUID = 1L;

    private static final transient Logger logger = LoggerFactory.getLogger(AbstractChordOverlay.class);

    public static final int STABILIZE_INTERVAL_TIME = 400;

    private static final int TOLERANT_FAILURES = 
        (int) Math.round(Math.log(Key.NB_BITS) / Math.log(2));

    private int fingerIndexToFix;

    private Key key;

    private Peer predecessor;

    private Peer successor;

    private FingerTable fingerTable;

    private final Peer[] successorsForTolerantFailures = new Peer[TOLERANT_FAILURES];

    private Entries<T> entries = new Entries<T>();

    /**
     * Constructor. An identifier will be automatically assigned.
     */
    public AbstractChordOverlay() {
        super(new DefaultChordQueryManager());
    }

    /**
     * Constructor.
     * 
     * @param id
     *            the identifier to assign to the node.
     */
    public AbstractChordOverlay(BigInteger id) {
        this();
        this.key = new Key(id);
    }

    /**
     * Constructor.
     * 
     * @param id
     *            the identifier to assign to the node.
     */
    public AbstractChordOverlay(String id) {
        this();
        this.key = new Key(id);
    }

    /**
     * {@inheritDoc}
     */
    public String dump() {
        return this.toString();
    }

    /**
     * Creates a new Chord network. It is called only by the first peer which
     * want to create the network.
     * 
     * @return <code>true</code> if the creation has succeeded,
     *         <code>false</code> otherwise.
     * @throws ChordException
     *             if the peer which calls the operation has already joined a
     *             network.
     */
    public boolean create() throws ChordException {
        if (super.getLocalPeer().isActivated()) {
            throw new ChordException("Peer has already joined or created a network");
        }

        if (logger.isInfoEnabled()) {
            logger.info("Node {} has created a new network!", this.key);
        }
        
        this.predecessor = null;
        this.successor = super.getRemotePeer();
        super.getLocalPeer().setActivated(true);

        this.update();

        return true;
    }

    /**
     * Joins a Chord ring by using an existing node in the Chord ring.
     * 
     * @param node
     *            a bootstrapping node.
     */
    public boolean join(Peer node) {
        this.predecessor = null;
        this.successor = ((FindSuccessorResponseOperation) PAFuture.getFutureValue(
                node.send(new FindSuccessorOperation(this.key)))).getSuccessor();

        // FIXME really ugly but we need to stabilize
        while (this.predecessor == null) {
            Thread.yield();
            this.update();
        }

        if (logger.isInfoEnabled()) {
            logger.info("Node {} has joined the network from node {}.",
                        this.key,
                        ((GetKeyResponseOperation) PAFuture.getFutureValue(
                                node.send(new GetKeyOperation()))).getKey());
        }
        return true;
    }

    /**
     * Leaves the overlay by informing neighbors that it want to leave the
     * network.
     * 
     * @return <code>true</code> if leave success, <code>false</code> otherwise.
     */
    public boolean leave() {
        try {
            if (!this.getKey().equals(
                    ((GetKeyResponseOperation) PAFuture.getFutureValue(this.successor
                            .send(new GetKeyOperation()))).getKey())) {
                // Gives all entries to successor
                PAFuture.waitFor(this.successor.send(new AddEntriesOperation<T>(this.entries
                        .getEntries())));
                PAFuture.waitFor(this.predecessor.send(new SetSuccessorOperation(this.successor)));
                PAFuture.waitFor(this.successor.send(new SetPredecessorOperation(this.predecessor)));
            }
        } catch (BodyTerminatedException e) {
            // we try to update a peer which has left or crashed
        } catch (NullPointerException e) {
            // some scenarios implies that the predecessor we try to update has
            // been detected as dead just before to send the message by
            // checkPredecessor
        }

        if (logger.isInfoEnabled()) {
            logger.info("Node {} has left the network", this.key);
        }

        return true;
    }

    /**
     * Lookups a successor of given key.
     * 
     * @param id
     *            an identifier to lookup.
     * @return the successor of given identifier.
     */
    public Peer findSuccessor(Key id) {
        if (logger.isTraceEnabled()) {
            logger.trace("{}.findSuccessor({})", this.key, id);
        }

        Key successorKey = null;

        try {
            successorKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(
                    this.successor.send(new GetKeyOperation()))).getKey();
            if (this.key.equals(successorKey)) {
                return super.getRemotePeer();
            }

            if (id.isBetween(this.key, successorKey) || id.compareTo(successorKey) == 0) {
                return this.successor;
            } else {
                Peer node = this.closestPrecedingNode(id);

                if (this.key.compareTo(((GetKeyResponseOperation) PAFuture.getFutureValue(node
                        .send(new GetKeyOperation()))).getKey()) == 0) {
                    return super.getRemotePeer();
                }

                return ((FindSuccessorResponseOperation) PAFuture.getFutureValue(node
                        .send(new FindSuccessorOperation(id)))).getSuccessor();
            }
        } catch (Exception e) {
            if (this.fixSuccessor().equals(super.getRemotePeer())) {
                return super.getRemotePeer();
            } else {
                return ((FindSuccessorResponseOperation) PAFuture.getFutureValue(
                        this.successor.send(new FindSuccessorOperation(id)))).getSuccessor();
            }
        }
    }

    /**
     * Lookups in the finger table the closest node which manages the key given
     * in parameter.
     * 
     * @param id
     *            the identifier used in order to lookup.
     * @return the closest node which manages the specified key.
     */
    public Peer closestPrecedingNode(Key id) {
        if (logger.isTraceEnabled()) {
            logger.trace("{}.closestPrecedingNode({})", this.key.getValue(), id.getValue());
        }

        for (int i = Key.NB_BITS - 1; i >= 0; i--) {
            Finger finger = fingerTable.getFinger(i);

            try {
                Key fingerKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(finger.getNode()
                        .send(new GetKeyOperation()))).getKey();
                if (fingerKey.isBetween(this.key, id)) {
                    return finger.getNode();
                }
            } catch (BodyTerminatedException e) {
                // wait for next stabilization...
            }
        }

        for (int i = TOLERANT_FAILURES - 1; i > 0; i--) {
            Peer candidate = this.successorsForTolerantFailures[i];

            try {
                Key candidateKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(candidate
                        .send(new GetKeyOperation()))).getKey();
                if (candidateKey.isBetween(this.key, id)) {
                    return candidate;
                }
            } catch (BodyTerminatedException e) {
                // ...
            }
        }

        return super.getRemotePeer();
    }

    /**
     * Launch a new stabilization step in order to ensure that lookups execute
     * correctly as the set of participating nodes changes. Chord must ensure
     * that each node's successor pointer is up to date.
     */
    public void update() {
        if (logger.isTraceEnabled()) {
            logger.trace("Node {} is performing stabilize algorithm", this.key);
        }

        this.fixSuccessorsForFaultTolerance();
        this.stabilize();
        this.fixFingers();
        this.checkPredecessor();
    }

    /**
     * Verifies the successor, and tells the successor about this node. It is
     * called periodically (each {@code #STABILIZE_INTERVAL_TIME} ms).
     * 
     * @see #runActivity(Body)
     */
    public void stabilize() {
        if (logger.isTraceEnabled()) {
            logger.trace("Node {} is stabilizing...", this.key);
        }

        Peer node = null;
        try {
            node = ((GetPredecessorResponseOperation) PAFuture.getFutureValue(
                    this.successor.send(new GetPredecessorOperation()))).getPredecessor();

            if (node != null) {
                Key nodeKey = null;
                Key successorKey = null;

                nodeKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(node
                        .send(new GetKeyOperation()))).getKey();

                successorKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(
                        this.successor.send(new GetKeyOperation()))).getKey();

                if (nodeKey.isBetween(this.key, successorKey)) {
                    this.successor = node;
                }
            }
            PAFuture.waitFor(this.successor
                    .send(new NotifyPredecessorOperation(super.getRemotePeer())));
        } catch (Exception e) {
            this.fixSuccessor();
        }
    }

    /**
     * Returns the closest alive successor available by looking up in the
     * successorsForTolerantFailures list first and then the finger table.
     * 
     * @return the closest alive successor.
     */
    public Peer firstLiveSuccessor() {
        return this.firstLiveSuccessor(1);
    }

    public Peer firstLiveSuccessor(int index) {
        if (logger.isTraceEnabled()) {
            logger.trace("{}.closestAliveSuccessor({})", this.key, index);
        }
        
        Peer current = null;
        if (index < this.successorsForTolerantFailures.length) {
            try {
                current = this.successorsForTolerantFailures[index];
                current.send(new IsAliveOperation());
                return current;
            } catch (Exception e) {
                return this.firstLiveSuccessor(index + 1);
            }
        } else if (index < Key.NB_BITS + TOLERANT_FAILURES) {
            try {
                current = this.fingerTable.getFinger(index - TOLERANT_FAILURES).getNode();
                current.send(new IsAliveOperation());
                return current;
            } catch (Exception e) {
                return this.firstLiveSuccessor(index + 1);
            }
        } else {
            return super.getRemotePeer();
        }
    }

    /**
     * Called when the peer specified in parameter think it might be the
     * predecessor of the current peer.
     * 
     * @param node
     *            the new predecessor supposed.
     * @return <code>true</code> if the call has succeeded, <code>false</code>
     *         otherwise.
     */
    public boolean notifyPredecessor(Peer node) {
        if (logger.isTraceEnabled()) {
            logger.trace("Node {} is notifying predecessor...", this.key);
        }
        
        try {
            Key nodeKey = null;
            nodeKey = ((GetKeyResponseOperation) PAFuture.getFutureValue(node
                    .send(new GetKeyOperation()))).getKey();

            if (this.predecessor == null
                    || nodeKey.isBetween(
                            ((GetKeyResponseOperation) PAFuture.getFutureValue(
                                    this.predecessor.send(
                                            new GetKeyOperation()))).getKey(), this.key)) {
                this.predecessor = node;
                this.transfertEntries(node, nodeKey);
            }
        } catch (BodyTerminatedException e) {
            this.fixSuccessor();
        }
        return true;
    }

    /**
     * @param node
     * @param nodeKey
     */
    private void transfertEntries(Peer node, Key nodeKey) {
        Map<Key, Entry<T>> entriesCheckingConstraints = new HashMap<Key, Entry<T>>();
        Iterator<Key> it = this.entries.getEntries().keySet().iterator();
        Key currentKey;

        while (it.hasNext()) {
            currentKey = it.next();
            if (((this.key.compareTo(nodeKey) <= 0 && currentKey.compareTo(nodeKey) <= 0 && currentKey
                    .compareTo(this.key) > 0) || (this.key.compareTo(nodeKey) > 0 && currentKey
                    .compareTo(nodeKey) <= 0))) {
                entriesCheckingConstraints.put(currentKey, this.entries.getEntries()
                        .get(currentKey));
                it.remove();
            }
        }

        if (entriesCheckingConstraints.size() > 0) {
            PAFuture.waitFor(node.send(new AddEntriesOperation<T>(entriesCheckingConstraints)));
        }
    }

    /**
     * Refreshes an entry of the finger table each time it is called. The size
     * of the finger table is defined by the length of the {@link Key}.
     */
    public void fixFingers() {
        if (logger.isTraceEnabled()) {
            logger.debug("Node {} is fixing fingers...", this.key);
        }
        this.fingerIndexToFix++;

        if (this.fingerIndexToFix >= Key.NB_BITS) {
            this.fingerIndexToFix = 0;
        }

        Finger finger = this.fingerTable.getFinger(this.fingerIndexToFix);
        Peer successorFound = (Peer) PAFuture.getFutureValue(this.findSuccessor(new Key(finger
                .getStartIndex())));
        finger.setNode(successorFound);
    }

    /**
     * Checks if predecessor is always running or not. If not, predecessor is
     * set to <code>null</code> in order to be update in a next stabilization
     * step..
     */
    public void checkPredecessor() {
        if (logger.isTraceEnabled()) {
            logger.trace("Node {} is checking predecessor...", this.key);
        }
        
        if (this.predecessor != null) {
            try {
                PAFuture.waitFor(this.predecessor.send(new IsAliveOperation()));
            } catch (Exception e) {
                this.predecessor = null;
            }
        }
    }

    /**
     * Updates successors list for fault-tolerance.
     */
    public void fixSuccessorsForFaultTolerance() {
        Peer[] successors = null;

        try {
            successors = ((GetSuccessorsFTResponseOperation) PAFuture.getFutureValue(this.successor
                    .send(new GetSuccessorsFTOperation()))).getSuccessors();
            for (int i = 1; i < TOLERANT_FAILURES; i++) {
                this.successorsForTolerantFailures[i] = successors[i - 1];
            }
            this.successorsForTolerantFailures[0] = this.successor;
        } catch (Exception e) {
            this.fixSuccessor();
        }
    }

    public Peer fixSuccessor() {
        this.successor = this.firstLiveSuccessor();
        return this.successor;
    }

    /**
     * Puts a new {@link Entry} on the overlay by looking for the correct peer
     * to use.
     * 
     * @param entry
     *            the entry to add.
     * 
     * @return Returns <code>true</code> if the operation has succeeded,
     *         <code>false</code> otherwise.
     */
    public boolean putEntry(Entry<T> entry) {
        Peer node = this.findSuccessor(entry.getKey());
        if (logger.isInfoEnabled()) {
            logger.info("("
                    + entry.getKey()
                    + ","
                    + entry.getValue()
                    + ") has been added on node "
                    + ((GetKeyResponseOperation) PAFuture.getFutureValue(node
                            .send(new GetKeyOperation()))).getKey());
        }

        return ((BooleanResponseOperation) PAFuture.getFutureValue(node.send(new AddEntryOperation<T>(
                entry)))).getValue();
    }

    /**
     * Lookups an {@link Entry} by using a given key.
     * 
     * @param key
     *            the key to use for the lookup.
     * @return the data found or <code>null</code> if the data is not on the
     *         overlay or if the Chord network is not stabilized.
     */
    @SuppressWarnings("unchecked")
    public Entry<T> findEntry(Key key) {
        if (logger.isDebugEnabled()) {
            logger.debug("{}.findEntry({})", this.key, key);
        }
        Peer node = this.findSuccessor(key);

        Entry<T> entryFound = ((EntryResponseOperation<T>) PAFuture.getFutureValue(node
                .send(new GetEntryOperation(key)))).getEntryFound();

        if (logger.isInfoEnabled()) {
            if (entryFound == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Successor found is {} but it doesn't contains key", node);
                }
                logger.info("Nobody on the network contains the key");
            } else {
                logger.info("Key "
                        + key
                        + " has been found on node "
                        + ((GetKeyResponseOperation) PAFuture.getFutureValue(node
                                .send(new GetKeyOperation()))).getKey()
                        + " and associated value is '" + entryFound + "'");
            }
        }

        return entryFound;
    }

    /**
     * Removes a data by using the specified key.
     * 
     * @param key
     *            the key to find.
     * @return <code>true</code> if the remove has succeeded, <code>false</code>
     *         otherwise.
     */
    public boolean remove(Key key) {
        return ((BooleanResponseOperation) PAFuture.getFutureValue(this.findSuccessor(key).send(
                new RemoveEntryOperation(key)))).getValue();
    }

    /**
     * Adds an entry to the local store.
     * 
     * @param entry
     *            the entry to add.
     * @return <code>true<code> if operation has succeeded. <code>false</code>
     *         otherwise.
     */
    public boolean addEntry(Entry<T> entry) {
        this.entries.getEntries().put(entry.getKey(), entry);
        return true;
    }

    /**
     * Adds all the given entries.
     * 
     * @param entries
     *            the entries to add.
     * @return <code>true<code> if operation has succeeded. <code>false</code>
     *         otherwise.
     */
    public boolean addEntries(Map<Key, Entry<T>> entries) {
        this.entries.getEntries().putAll(entries);
        return true;
    }

    /**
     * Returns an entry by key from the local store.
     * 
     * @param key
     *            the key to use in order to find the entry.
     * @return the entry found or <code>null</code>.
     */
    public Entry<T> getEntry(Key key) {
        return this.entries.getEntry(key);
    }

    /**
     * Returns all the entries managed by the peer.
     * 
     * @return all the entries managed by the peer.
     */
    public Entries<T> getEntries() {
        return this.entries;
    }

    /**
     * Returns the number of entries that the peer is managing.
     * 
     * @return the number of entries that the peer is managing.
     */
    public int getNbEntries() {
        return this.entries.getEntries().size();
    }

    /**
     * Retrieves between to bounds.
     * 
     * @param from
     *            lower bound.
     * @param to
     *            upper bound.
     * @return the entry found for the given range.
     */
    public Map<Key, Entry<T>> transfertEntries(Key from, Key to) {
        return this.entries.getEntriesBetween(from, to);
    }

    /**
     * Removes an entry identified by a specified {@link Key}.
     * 
     * @param key
     *            the key to use.
     * @return <code>true<code> if operation has succeeded. <code>false</code>
     *         otherwise.
     */
    public boolean removeEntry(Key key) {
        return this.entries.getEntries().remove(key) != null;
    }

    /**
     * Returns the key associated to this peer.
     * 
     * @return the key associated to this peer.
     */
    public Key getKey() {
        return this.key;
    }

    /**
     * Returns the predecessor of this peer.
     * 
     * @return the predecessor of this peer.
     */
    public Peer getPredecessor() {
        return this.predecessor;
    }

    /**
     * Returns the successor of this peer.
     * 
     * @return the successor of this peer.
     */
    public Peer getSuccessor() {
        return this.successor;
    }

    /**
     * Returns the finger table associated to this peer.
     * 
     * @return the finger table associated to this peer.
     */
    public FingerTable getFingerTable() {
        return this.fingerTable;
    }

    /**
     * Sets the predecessor.
     * 
     * @param predecessor
     *            the new predecessor to set.
     */
    public void setPredecessor(Peer predecessor) {
        this.predecessor = predecessor;
    }

    /**
     * Sets the successor.
     * 
     * @param successor
     *            the new successor to set.
     */
    public void setSuccessor(Peer successor) {
        this.successor = successor;
    }

    public Peer[] getSuccessorsForTolerantFailures() {
        return successorsForTolerantFailures;
    }

    /**
     * Sets the finger table.
     * 
     * @param fingerTable
     *            the new finger table to use.
     */
    public void setFingerTable(FingerTable fingerTable) {
        this.fingerTable = fingerTable;
    }

    public int compareTo(Object obj) {
        if (super.getLocalPeer().getType() != ((Peer) obj).getType()) {
            throw new IllegalArgumentException("Objects to compare have not the same type");
        }

        return this.key.compareTo(
                ((GetKeyResponseOperation) PAFuture.getFutureValue(
                        ((Peer) obj).send(new GetKeyOperation()))).getKey());
    }

    /**
     * {@inheritDoc}
     */
    public boolean equals(Object obj) {
        return this.compareTo((Peer) obj) == 0;
    }

    /**
     * {@inheritDoc}
     */
    public int hashCode() {
        return this.key.getValue().hashCode();
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("Node ");
        buf.append(this.getKey().getValue());
        buf.append(" has P=");
        if (this.predecessor == null) {
            buf.append("null");
        } else {
            try {
                buf.append(((GetKeyResponseOperation) PAFuture.getFutureValue(this.predecessor
                        .send(new GetKeyOperation()))).getKey());
            } catch (BodyTerminatedException e) {
                buf.append("?");
            }
        }
        buf.append(", S=");
        try {
            buf.append(((GetKeyResponseOperation) PAFuture.getFutureValue(this.successor
                    .send(new GetKeyOperation()))).getKey());
        } catch (BodyTerminatedException e) {
            buf.append("?");
        }
        buf.append(", FT=[");
        for (int i = 0; i < Key.NB_BITS; i++) {
            Finger finger = fingerTable.getFinger(i);
            try {
                buf.append(((GetKeyResponseOperation) PAFuture.getFutureValue(finger.getNode().send(
                        new GetKeyOperation()))).getKey());
            } catch (BodyTerminatedException e) {
                // just ignore it because peer has left and finger
                // table will be update in next stabilization
                buf.append("?");
            }
            if (i != Key.NB_BITS - 1) {
                buf.append(" ");
            }
        }
        buf.append("]");
        buf.append(", S=[");

        for (int j = 0; j < this.successorsForTolerantFailures.length; j++) {
            try {
                buf.append(((GetKeyResponseOperation) PAFuture
                        .getFutureValue(this.successorsForTolerantFailures[j]
                                .send(new GetKeyOperation()))).getKey());
            } catch (Exception e) {
                buf.append("?");
            }

            if (j < this.successorsForTolerantFailures.length - 1) {
                buf.append(",");
            }
        }
        buf.append("]");
        buf.append(", E=[");

        int i = 0;
        for (Entry<T> entry : this.entries) {
            buf.append("(");
            buf.append(entry.getKey());
            buf.append(",");
            buf.append(entry.getValue());
            buf.append(")");
            if (i < this.entries.getEntries().size() - 1) {
                buf.append(", ");
            }
            i++;
        }
        buf.append("]");

        return buf.toString();
    }

    /**
     * {@inheritDoc}
     */
    public void initActivity(Body body) {
        PAActiveObject.setImmediateService("send");

        if (this.key == null) {
            this.key = new Key(UUID.randomUUID().toString());
        }

        for (int i = 0; i < this.successorsForTolerantFailures.length; i++) {
            this.successorsForTolerantFailures[i] = super.getRemotePeer();
        }

        this.fingerTable = new FingerTable(this, super.getRemotePeer());
        if (logger.isDebugEnabled()) {
            logger.info("Node {} is created", this.key);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void runActivity(Body body) {
        Service service = new Service(body);
        while (body.isActive()) {
            Request request = service
                    .blockingRemoveOldest(AbstractChordOverlay.STABILIZE_INTERVAL_TIME);

            // timeout has expired and no request to serve
            if (request == null && super.getLocalPeer().isActivated()) {
                this.update();
            } else {
                service.serve(request);
            }
        }
    }

    public OverlayType getType() {
        return OverlayType.CHORD;
    }

    public int getUpdateIntervalTime() {
        return AbstractChordOverlay.STABILIZE_INTERVAL_TIME;
    }

}
