package org.objectweb.proactive.extensions.structuredp2p.overlay.chord;

import java.io.Serializable;
import java.math.BigInteger;

import org.apache.log4j.Logger;
import org.objectweb.proactive.extensions.structuredp2p.util.HashUtil;

/**
 * A {@link Key} represents an identifier used by {@link AbstractChordOverlay} and
 * {@link Entries} that are stored.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: Key.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public final class Key implements Comparable<Key>, Serializable {

    private static final long serialVersionUID = 1L;

    private static final Logger logger = Logger.getLogger(Key.class);

    public static final int NB_BITS = 8;

    public static final BigInteger MAX_VALUE = BigInteger.valueOf(2 << (NB_BITS - 1));

    private final BigInteger value;

    /**
     * Constructor.
     * 
     * @param id
     *            the identifier to use modulo the size of the key which is
     *            {@link Key#NB_BITS}.
     */
    public Key(int id) {
        if (id < 0) {
            throw new AssertionError("id must be positive");
        } else {
            this.value = BigInteger.valueOf(id % (int) Math.pow(2, Key.NB_BITS));
        }
    }

    /**
     * Constructor.
     * 
     * @param id
     *            the identifier to use modulo the size of the key which is
     *            {@link Key#NB_BITS}.
     */
    public Key(BigInteger id) {
        if (id.compareTo(BigInteger.ZERO) < 0) {
            throw new AssertionError("Number must be positive");
        } else {
            this.value = id.mod(BigInteger.valueOf(2 << (Key.NB_BITS - 1)));
        }
    }

    /**
     * Constructor.
     * 
     * @param id
     *            the identifier to use that will be hashed modulo the size of
     *            the key which is {@link Key#NB_BITS}.
     */
    public Key(String id) {
        this.value = HashUtil.getSHA1KeyRestrictedToIntervalByModulo(
                                id, 2 << (Key.NB_BITS - 1));
    }

    /**
     * {@inheritDoc}
     */
    public int compareTo(Key key) {
        return this.value.compareTo(key.getValue());
    }

    /**
     * Indicated if the current key is between two others.
     * 
     * @param fromKey
     *            lower bound.
     * @param toKey
     *            upper bound.
     * @return
     *         <code>true<code> if the current key is between the two others, <code>false</code>
     *         otherwise.
     */
    public boolean isBetween(Key fromKey, Key toKey) {
        boolean result;

        if (fromKey.compareTo(toKey) == 0) {
            result = this.compareTo(fromKey) != 0;
        } else if (fromKey.compareTo(toKey) < 0) {
            result = this.compareTo(fromKey) > 0 && this.compareTo(toKey) < 0;
        } else {
            result = this.compareTo(fromKey) > 0 || this.compareTo(toKey) < 0;
        }

        if (logger.isTraceEnabled()) {
            logger.trace(fromKey.getValue() + " < " + this.value + " < " + toKey.getValue() + " ? "
                    + result);
        }

        return result;
    }

    /**
     * {@inheritDoc}
     */
    public boolean equals(Object obj) {
        return this.compareTo((Key) obj) == 0;
    }

    /**
     * {@inheritDoc}
     */
    public int hashCode() {
        return this.value.hashCode();
    }

    /**
     * Returns the value associated to this key.
     * 
     * @return the value associated to this key.
     */
    public BigInteger getValue() {
        return this.value;
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        return this.value.toString();
    }
    
}
