package org.objectweb.proactive.extensions.structuredp2p.router;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A router defines how to route a {@link SynchronousMessage}. It provides some
 * methods to add some functionalities at different steps of the routing.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: Router.java 5028 2010-08-15 10:55:06Z plaurent $
 * 
 * @param <T>
 *            the type of message to route.
 * @param <K>
 *            the type of the key used to check if the constraints are
 *            validated.
 */
public abstract class Router<T extends SynchronousMessage<K>, K> {

    private static final transient Logger logger = LoggerFactory.getLogger(Router.class);

    protected ConstraintsValidator<K> constraintsValidator;

    /**
     * Initialize a router with the specified <code>validator</code>.
     * 
     * @param validator
     *            the {@link ConstraintsValidator} to use in order to check the
     *            constraints.
     */
    public Router(ConstraintsValidator<K> validator) {
        this.constraintsValidator = validator;
    }

    protected void handle(StructuredOverlay overlay, T msg) {
        this.preHandle(overlay, msg);
        this.performHandle(overlay, msg);
        this.postHandle(overlay, msg);
    }

    public void route(StructuredOverlay overlay, T msg) {
        this.preRoute(overlay, msg);
        this.performRoute(overlay, msg);
        this.postRoute(overlay, msg);
    }

    /**
     * This method is called just before
     * {@link Router#performHandle(StructuredOverlay, SynchronousMessage)}. You
     * can override it for custom action.
     */
    protected void preHandle(StructuredOverlay overlay, T msg) {
        logger.trace("preHandle()");
    }

    /**
     * This method is called just before
     * {@link Router#performRoute(StructuredOverlay, SynchronousMessage)}. You
     * can override it for custom action.
     */
    protected void preRoute(StructuredOverlay overlay, T msg) {
        logger.trace("preRoute()");
    }

    /**
     * This method is called just after
     * {@link Router#performHandle(StructuredOverlay, SynchronousMessage)}. You
     * can override for custom action.
     */
    protected void postHandle(StructuredOverlay overlay, T msg) {
        logger.trace("postHandle()");
    }

    /**
     * This method is called just after
     * {@link Router#postRoute(StructuredOverlay, SynchronousMessage)}. You can
     * override it for custom action.
     */
    protected void postRoute(StructuredOverlay overlay, T msg) {
        logger.trace("postRoute()");
    }

    protected void onDestinationReached(StructuredOverlay overlay, T msg) {
        logger.trace("onDestinationReached()");
    }

    /**
     * Handles the specified <code>msg</msg> on the given <code>overlay</code>.
     * 
     * @param overlay
     *            the {@link StructuredOverlay} used to handle the message.
     * @param msg
     *            the {@link SynchronousMessage} to handle.
     */
    protected abstract void performHandle(StructuredOverlay overlay, T msg);

    /**
     * Make decision to route the message to an another peer or to handle it by
     * calling {@link #handle(StructuredOverlay, SynchronousMessage)}. The
     * message must be handled if the constraints are validated by using the
     * {@link ConstraintsValidator}.
     * 
     * @param overlay
     *            the {@link StructuredOverlay} to use in order to make the
     *            decision.
     * @param msg
     *            the {@link SynchronousMessage} which have to be routed.
     */
    protected abstract void performRoute(StructuredOverlay overlay, T msg);

    /**
     * Validates the contraints by using the specified {@link StructuredOverlay}
     * and key K.
     * 
     * @param overlay
     *            the {@link StructuredOverlay} on which the constraints are
     *            checked.
     * @param key
     *            the key to use to check the constraints.
     * @return <code>true</code> if the constraints are validated,
     *         <code>false</code> otherwise.
     * 
     * @see ConstraintsValidator#validatesKeyConstraints(StructuredOverlay,
     *      Object)
     */
    protected boolean validatesKeyConstraints(StructuredOverlay overlay, K key) {
        return this.constraintsValidator.validatesKeyConstraints(overlay, key);
    }

}
