package org.objectweb.proactive.extensions.structuredp2p.router.can;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessage;
import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;
import org.objectweb.proactive.extensions.structuredp2p.messages.query.AbstractQueryMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.AbstractCANOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.NeighborTable;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.requests.BlockingRequestReceiverException;
import org.objectweb.proactive.extensions.structuredp2p.router.Router;
import org.objectweb.proactive.extensions.structuredp2p.util.LoggerUtil;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Router used to route {@link SynchronousMessage}s from a {@link Peer} to an
 * another {@link Peer}.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: UnicastQueryRouter.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class UnicastQueryRouter<T extends AbstractQueryMessage<CoordinateValue[]>> 
                                  extends Router<T, CoordinateValue[]> {

    private static final transient Logger logger;

    private static final transient boolean isDebugEnabled;
    
    static {
        logger = LoggerFactory.getLogger(UnicastQueryRouter.class);
        isDebugEnabled = logger.isDebugEnabled();
    }
    
    public UnicastQueryRouter(ConstraintsValidator<CoordinateValue[]> validator) {
        super(validator);
    }

    protected void performHandle(StructuredOverlay overlay, T msg) {
        if (isDebugEnabled) {
            logger.debug("performHandle({}, msg={})", overlay, msg.getUUID());
        }
        this.onDestinationReached(overlay, msg);
        msg.createResponseMessage().route(overlay);
    }

    protected void performRoute(StructuredOverlay overlay, T msg) {
        AbstractCANOverlay overlayCAN = ((AbstractCANOverlay) overlay);
        CoordinateValue[] coordinatesToReach = msg.getKeyToReach();

        if (isDebugEnabled) {
            StringBuilder buf = new StringBuilder();
            buf.append("performRoute(overlay, msg=");
            buf.append(msg.getUUID());
            buf.append(", keyToReach=");
            buf.append(LoggerUtil.coordinateAsString(coordinatesToReach));
            buf.append(", currentZone=");
            buf.append(((AbstractCANOverlay) overlay).getZone());
            logger.debug(buf.toString());
        }

        if (msg.getHopCount() == 0) {
            overlay.getSynchronousMessagesReceived().put(
                    msg.getUUID(), new SynchronousMessageEntry(1, 0));
        }

        if (overlayCAN.contains(coordinatesToReach)) {
            if (isDebugEnabled) {
                logger.debug("Peer which contains the key to reach has been found");
            }
            this.handle(overlay, msg);
        } else {
            int direction;
            int pos;
            
            if (isDebugEnabled) {
                logger.debug("Route to other peer because current peer {} does not manage the key to reach {}", 
                        overlay,  LoggerUtil.coordinateAsString(coordinatesToReach));
            }

            for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
                direction = NeighborTable.SUPERIOR_DIRECTION;
                pos = overlayCAN.contains(dim, coordinatesToReach[dim]);

                if (pos == -1) {
                    direction = NeighborTable.INFERIOR_DIRECTION;
                }

                if (pos != 0) {
                    if (overlayCAN.getNeighborTable().get(dim, direction).size() > 0) {
                        Peer neighborChosen = 
                            overlayCAN.getNeighborTable().getRandomNeighborValidatingConstraints(
                                    overlayCAN.getZone(), coordinatesToReach, dim, direction);

                        try {
                            msg.incrementNbStepsBy(1);
                            neighborChosen.route(msg);
                            if (logger.isDebugEnabled()) {
                                logger.debug("Message sent to {}", neighborChosen);
                            }
                        } catch (BlockingRequestReceiverException e) {
                            // overlay.bufferizeSynchronousMessage(msg);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        break;
                    }
                }
            }
        }
    }

}
