package org.objectweb.proactive.extensions.structuredp2p.router.can;

import org.objectweb.proactive.extensions.structuredp2p.messages.SynchronousMessageEntry;
import org.objectweb.proactive.extensions.structuredp2p.messages.response.AbstractResponseMessage;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.objectweb.proactive.extensions.structuredp2p.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.AbstractCANOverlay;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.NeighborTable;
import org.objectweb.proactive.extensions.structuredp2p.overlay.can.coordinates.CoordinateValue;
import org.objectweb.proactive.extensions.structuredp2p.requests.BlockingRequestReceiverException;
import org.objectweb.proactive.extensions.structuredp2p.router.Router;
import org.objectweb.proactive.extensions.structuredp2p.validator.ConstraintsValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Laurent Pellegrino
 * 
 * @version $Id: UnicastResponseRouter.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class UnicastResponseRouter<T extends AbstractResponseMessage<CoordinateValue[]>>
        extends Router<T, CoordinateValue[]> {

    private static final transient Logger logger;

    private static final transient boolean isDebugEnabled;
    
    static {
        logger = LoggerFactory.getLogger(UnicastResponseRouter.class);
        isDebugEnabled = logger.isDebugEnabled();
    }
    
    public UnicastResponseRouter(ConstraintsValidator<CoordinateValue[]> validator) {
        super(validator);
    }

    protected void performHandle(StructuredOverlay overlay, T msg) {
        if (isDebugEnabled) {
            logger.debug("performHandle(overlay, msg={})", msg.getUUID());
        }
        overlay.getQueryManager().putFinalResponseAndNotifyInitialSender(msg);
    }

    protected void performRoute(StructuredOverlay overlay, T msg) {
        AbstractCANOverlay canOverlay = ((AbstractCANOverlay) overlay);
        CoordinateValue[] coordinatesToReach = msg.getKeyToReach();

        if (isDebugEnabled) {
            StringBuffer buf = new StringBuffer();
            buf.append("performRoute(overlay, msg=");
            buf.append(msg.getUUID());
            buf.append("), keyToReach=(");

            for (int i = 0; i < coordinatesToReach.length; i++) {
                buf.append(coordinatesToReach[i]);
                if (i < coordinatesToReach.length - 1) {
                    buf.append(",");
                }
            }
            buf.append("), currentZone=");
            buf.append(((AbstractCANOverlay) overlay).getZone());
            logger.debug(buf.toString());
        }

        if (msg.getHopCount() == 0) {
            overlay.getSynchronousMessagesReceived().put(
                    msg.getUUID(), new SynchronousMessageEntry(1, 0));
        }

        if (canOverlay.contains(coordinatesToReach)) {
            if (isDebugEnabled) {
                logger.debug("Peer which contains the key to reach has been found");
            }
            this.handle(overlay, msg);
        } else {
            int direction;
            int pos;
            if (isDebugEnabled) {
                logger.debug("Route to other peer because current peer does not manage the key to reach");
            }

            for (int dim = 0; dim < AbstractCANOverlay.NB_DIMENSIONS; dim++) {
                direction = NeighborTable.SUPERIOR_DIRECTION;
                pos = canOverlay.contains(dim, coordinatesToReach[dim]);

                if (pos == -1) {
                    direction = NeighborTable.INFERIOR_DIRECTION;
                }

                if (pos != 0) {
                    if (canOverlay.getNeighborTable().get(dim, direction).size() > 0) {
                        Peer nearestPeer = 
                            canOverlay.getNeighborTable().getRandomNeighborValidatingConstraints(
                                    canOverlay.getZone(), coordinatesToReach, dim, direction);
                        try {
                            msg.incrementNbStepsBy(1);
                            nearestPeer.route(msg);
                            if (isDebugEnabled) {
                                logger.debug("Message sent to peer {}", nearestPeer);
                            }
                        } catch (BlockingRequestReceiverException e) {
                            // overlay.bufferizeSynchronousMessage(msg);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        break;
                    } else{
                        logger.error("No neighbor to send to");
                    }
                }
            }
        }
    }

}
