package org.objectweb.proactive.extensions.structuredp2p.tracker;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.proactive.Body;
import org.objectweb.proactive.api.PAFuture;
import org.objectweb.proactive.core.body.exceptions.BodyTerminatedException;
import org.objectweb.proactive.core.util.ProActiveRandom;
import org.objectweb.proactive.extensions.structuredp2p.exceptions.StructuredP2PException;
import org.objectweb.proactive.extensions.structuredp2p.operations.IsAliveOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.CreateNetworkOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetKeyOperation;
import org.objectweb.proactive.extensions.structuredp2p.operations.chord.GetKeyResponseOperation;
import org.objectweb.proactive.extensions.structuredp2p.overlay.OverlayType;
import org.objectweb.proactive.extensions.structuredp2p.overlay.Peer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Specific implementation of {@link Tracker} for a Chord overlay network.
 * 
 * @author Laurent Pellegrino
 * 
 * @version $Id: TrackerChord.java 5022 2010-08-14 16:14:45Z plaurent $
 */
public class TrackerChord extends Tracker {

    private static final long serialVersionUID = 1L;

    private static final transient Logger logger = LoggerFactory.getLogger(TrackerChord.class);

    public static final int CHECK_INTERVAL_TIME = 3000;

    private boolean isActivated = false;

    public TrackerChord() {
        super(OverlayType.CHORD);
    }

    public TrackerChord(String associatedNetworkName) {
        super(OverlayType.CHORD, associatedNetworkName);
    }

    public TrackerChord(String associatedNetworkName, String trackerName) {
        super(OverlayType.CHORD, associatedNetworkName, trackerName);
    }

    public TrackerChord(double probabilityToStorePeer) {
        super(OverlayType.CHORD, probabilityToStorePeer);
    }

    /**
     * {@inheritDoc}
     */
    public boolean addOnNetwork(Peer peer) throws IllegalArgumentException {
        if (super.getNbOfPeersStored() == 0) {
            super.getStoredPeers().add(peer);
            PAFuture.waitFor(peer.send(new CreateNetworkOperation()));

            if (logger.isInfoEnabled()) {
                logger.info("Chord node "
                        + ((GetKeyResponseOperation) PAFuture.getFutureValue(peer
                                .send(new GetKeyOperation()))).getKey()
                        + " has created a new network");
            }
        } else {
            Peer peerToJoin = super.getRandomPeer();

            try {
                if ((Boolean) PAFuture.getFutureValue(peer.join(peerToJoin))) {
                    if (ProActiveRandom.nextDouble() < this.getProbabilityToStorePeer()) {
                        super.getStoredPeers().add(peer);
                    }

                    if (logger.isInfoEnabled()) {
                        logger.info("Chord node "
                                + ((GetKeyResponseOperation) PAFuture.getFutureValue(peer
                                        .send(new GetKeyOperation()))).getKey() + " has joined");
                    }
                }
            } catch (BodyTerminatedException e) {
                if (logger.isInfoEnabled()) {
                    logger.info("A dead peer maintained as entry point has been detected");
                }
                super.getStoredPeers().remove(peerToJoin);
                this.addOnNetwork(peer);
            } catch (StructuredP2PException e) {
                e.printStackTrace();
            }
        }

        return true;
    }

    /**
     * {@inheritDoc}
     */
    public void endActivity(Body body) {
        this.isActivated = false;
        super.endActivity(body);
    }

    /**
     * Checks if peers maintains as entry point are alive. If not, their
     * reference is thrown.
     */
    private void updateStoredPeers() {
        logger.debug("Checking peers maintain as entry point...");

        List<Peer> failedPeers = new ArrayList<Peer>(super.getNbOfPeersStored() / 2);
        Peer currentPeer = null;

        for (int i = 0; i < super.getNbOfPeersStored(); i++) {
            try {
                currentPeer = super.getStoredPeerAt(i);
                currentPeer.send(new IsAliveOperation());
            } catch (BodyTerminatedException e) {
                failedPeers.add(currentPeer);
            }
        }

        for (Peer peer : failedPeers) {
            super.removePeer(peer);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void initActivity(Body body) {
        super.initActivity(body);
        
        // Cannot use runActivity for update : in case of a stored peer failure,
        // the addOnNetwork function will be called recursively and the
        // updateStoredPeers method will never be called. Indeed the blocking
        // serve in runActivity can perform a method call (for update) only if
        // it has no entry in it queue. This is not the case because of recursive 
        // calls.
        new Thread(new Runnable() {
            public void run() {
                while (TrackerChord.this.isActivated) {
                    TrackerChord.this.updateStoredPeers();
                    try {
                        Thread.sleep(TrackerChord.CHECK_INTERVAL_TIME);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
        }).start();
    }

}
