/*
 * Copyright (C) 2010
 * 
 * STI Innsbruck ICT - Technologie Park Innsbruck, 2nd Floor Technikerstrasse
 * 21a A-6020 Innsbruck, Austria
 * 
 * Contact: reto.krummenacher@sti2.at gerald.schrempf@sti2.at (development)
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

package at.sti2.semanticspaces.impl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.ModelFactory;
import org.ontoware.rdf2go.RDF2Go;
import org.ontoware.rdf2go.exception.MalformedQueryException;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.ModelSet;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.impl.StatementImpl;
import org.ontoware.rdf2go.model.impl.URIGenerator;
import org.ontoware.rdf2go.model.node.URI;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;

import at.sti2.semanticspaces.api.ISemanticSpace;
import at.sti2.semanticspaces.api.ISpaceModel;
import at.sti2.semanticspaces.api.enums.SpaceModelType;
import at.sti2.semanticspaces.api.enums.SpaceRelation;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.api.exceptions.SpaceAlreadyExistsException;
import at.sti2.semanticspaces.api.exceptions.SpaceNotExistsException;
import at.sti2.semanticspaces.api.exceptions.SpaceTypeNotSupportedException;
import at.sti2.semanticspaces.api.spaceOntology.SpaceOntology;
import at.sti2.semanticspaces.impl.model.FederationModel;
import at.sti2.semanticspaces.impl.model.LODModel;
import at.sti2.semanticspaces.impl.model.SpaceModel;
import at.sti2.semanticspaces.impl.queries.AskQuery;

/**
 * 
 * repository-properties can be passed via a repository-file or
 * properties-object.
 * 
 * possible repository-properties: RDF2GO_REGISTER_REPOSITORY (needed),
 * RDF2GO_REGISTER_OVERLAY (optional).
 * 
 * see also the needed properties for the ModelFactory: ModelFactory.STORAGE,
 * ModelFactory.REASONING (both needed)
 * 
 * @author Gerald Schrempf
 * 
 */

public class SemanticSpace implements ISemanticSpace {
	/**
	 * register string for a Repository ModelFactory in the properties-file.
	 */
	static final String RDF2GO_REGISTER_REPOSITORY = "repository";

	/**
	 * register string for an Overlay ModelFactory in the properties-file.
	 */
	static final String RDF2GO_REGISTER_OVERLAY = "overlay";

	/**
	 * default property-file for configuration. (src/main/resources)
	 */
	static final String DEFAULT_CONFIG_FILE = "DefaultConfig.properties";

	/**
	 * modelSet for the underlying repository. Cannot be null.
	 */
	ModelSet repository;

	/**
	 * modelSet for distributed spaces. Can be null.
	 */
	ModelSet overlay;

	/**
	 * contains all different kinds of space-models. local HashMap, not stored
	 * persistent.
	 */
	private Map<URI, SpaceModel> spaceModels = new HashMap<URI, SpaceModel>();

	/**
	 * holds temporary all federations of this kernel. Map<federationURI, Set of
	 * SpaceModels>
	 */
	private Map<URI, FederationModel> federations = new HashMap<URI, FederationModel>();

	/**
	 * uses the default-properties. (sesame-repository, no overlay)
	 * 
	 * @throws SemanticSpaceException
	 */
	public SemanticSpace() throws SemanticSpaceException {
		this(DEFAULT_CONFIG_FILE);
	}

	/**
	 * reads a property-file.
	 * 
	 * @param propertyFile
	 * @throws SemanticSpaceException
	 */
	public SemanticSpace(String propertyFile) throws SemanticSpaceException {
		this(PropertyReader.getProperties(propertyFile));
	}

	/**
	 * initiates a new local repository/overlay connection. creates the
	 * metadataSpace (ignored if already exists). the metadataSpace contains all
	 * information about space relations.
	 * 
	 * @param propertiesFile
	 *            The name of the property file.
	 * @throws SemanticSpaceException
	 */
	public SemanticSpace(Properties props) throws SemanticSpaceException {
		rdf2goRegister(props, RDF2GO_REGISTER_REPOSITORY);
		
		repository = getModelFactory(props, RDF2GO_REGISTER_REPOSITORY);
		overlay = getModelFactory(props, RDF2GO_REGISTER_OVERLAY);

		try {
			createSpace(SpaceOntology.metadataSpace);
		} catch (SpaceAlreadyExistsException e) {
			// do nothing.
		}
	}

	/**
	 * registers a rdf2go framework.
	 * @param props
	 * @param rdf2goRegisterRepository
	 */
	private void rdf2goRegister(Properties props,
			String registerKey) {
		if (props.containsKey(registerKey)) {
			RDF2Go.register(props.getProperty(registerKey));
		}
	}

	/**
	 * registers an rdf2go ModelFactory, creates a new ModelSet and opens it.
	 * returns null if not possible.
	 * 
	 * @param props
	 * @param registerKey
	 * @return
	 * @throws SemanticSpaceException 
	 */
	private ModelSet getModelFactory(Properties props, String registerKey) throws SemanticSpaceException {
		if (props.containsKey(registerKey)) {
			ModelFactory mf;
			try {
				ClassLoader classLoader = getClass().getClassLoader();
				String property = props.getProperty(registerKey);
				
				Class<?> modelFactory = classLoader.loadClass(property);
				mf = (ModelFactory) modelFactory.newInstance();
			} catch (ClassNotFoundException e) {
				throw new SemanticSpaceException(e);
			} catch (InstantiationException e) {
				throw new SemanticSpaceException(e);
			} catch (IllegalAccessException e) {
				throw new SemanticSpaceException(e);
			}

			ModelSet ms = mf.createModelSet(props);
			ms.open();
			return ms;
		}
		return null;
	}

	// ***** SPACE CREATION METHODS *****//
	/**
	 * creates a repository space.
	 * 
	 * @return
	 */
	private SpaceModel createSpace(URI space) throws SemanticSpaceException {
		return createSpace(space, SpaceModelType.LocalRepository);
	}

	/**
	 * creates a new space with the specified type.
	 * 
	 * @param space
	 * @param type
	 * @return
	 */
	public SpaceModel createSpace(URI space, SpaceModelType type)
			throws SemanticSpaceException {
		checkSpaceNotExists(space);
		assertSpaceType(type);
		createMetadata(SpaceOntology.rootSpace,
				SpaceRelation.tsonto_hasSubspace.getURI(), space);
		createMetadata(space, SpaceOntology.tsonto_isTypeOf, type.getURI());
		return getSpace(space);
	}

	/**
	 * checks if space does not exist. throws an exception otherwise.
	 * 
	 * @param space
	 * @throws SpaceAlreadyExistsException
	 */
	private void checkSpaceNotExists(URI space)
			throws SpaceAlreadyExistsException {
		if (spaceExists(space)) {
			throw new SpaceAlreadyExistsException("space " + space
					+ " already exists");
		}
	}

	/**
	 * checks if this space-type is supported. throws an exception otherwise.
	 * 
	 * @param type
	 * @throws SpaceTypeNotSupportedException
	 */
	private void assertSpaceType(SpaceModelType type)
			throws SpaceTypeNotSupportedException {
		if (type == SpaceModelType.DistributedSpace && overlay == null) {
			throw new SpaceTypeNotSupportedException(
					"OverlayModelFactory not registered/accessible!");
		}
	}

	/**
	 * creates a subspace of a parent with the specified type.
	 * 
	 * @param space
	 * @param type
	 * @param parent
	 * @return
	 */
	public SpaceModel createSpace(URI space, SpaceModelType type, URI parent)
			throws SemanticSpaceException {
		checkSpaceExists(parent);
		createSpace(space, type);
		createMetadata(parent, SpaceRelation.tsonto_hasSubspace.getURI(), space);
		return getSpace(space);
	}

	/**
	 * returns the space if it already exists. if its not contained in the local
	 * HashMap, it will get loaded and recovered in a SpaceModel. throws an
	 * exception if the space does not exist.
	 * 
	 * this method really creates a new repository-space if its the first call
	 * for this URI. spaceModel.open() is called internally.
	 * 
	 * @throws SemanticSpaceException
	 * @throws ModelRuntimeException
	 * @throws MalformedQueryException
	 */
	public SpaceModel getSpace(URI space) throws SemanticSpaceException {
		checkSpaceExists(space);

		SpaceModel spaceModel;
		if (spaceModels.containsKey(space)) {
			spaceModel = spaceModels.get(space);
		} else {
			spaceModel = new SpaceModel(
					recoverModel(space, getSpaceType(space)), this);
			spaceModels.put(space, spaceModel);
		}
		spaceModel.open();
		return spaceModel;
	}

	/**
	 * checks if the space exists. throws an exception otherwise.
	 * 
	 * @param space
	 * @throws SpaceNotExistsException
	 */
	private void checkSpaceExists(URI space) throws SpaceNotExistsException {
		if (!spaceExists(space)) {
			throw new SpaceNotExistsException("space " + space
					+ " does not exist");
		}
	}

	/**
	 * returns the spaceModels.
	 * 
	 * @param spaces
	 * @return
	 */
	public Set<ISpaceModel> getSpaces(Set<URI> spaces)
			throws SemanticSpaceException {
		Set<ISpaceModel> spaceModels = new HashSet<ISpaceModel>();
		for (URI space : spaces) {
			spaceModels.add(getSpace(space));
		}
		return spaceModels;
	}

	/**
	 * gets the model for a space, with the specified type.
	 * 
	 * @param space
	 * @param type
	 * @return
	 * @throws SemanticSpaceException
	 * @throws SemanticSpaceException
	 */
	private Model recoverModel(URI space, SpaceModelType type)
			throws SemanticSpaceException {
		assertSpaceType(type);

		switch (type) {
		case LocalRepository: // gets an existing space or creates it
			return repository.getModel(space);
		case DistributedSpace:
			return overlay.getModel(space);
		case LOD:
			return new LODModel(space);
		default:
			throw new SemanticSpaceException(
					"modeltype not supported for recovering: "
							+ type.toString());
		}
	}

	/**
	 * get all subspaces of a space.
	 * 
	 * @param space
	 * @return
	 * @throws SemanticSpaceException
	 */
	public Set<ISpaceModel> getSubspaces(URI space)
			throws SemanticSpaceException {
		return getSpaces(getSubspacesInternal(space, space, new HashSet<URI>()));
	}

	/**
	 * list all spaces.
	 * 
	 * @return
	 */
	public Set<URI> listSpaces() {
		Set<URI> existingUris = null;
		try {
			existingUris = queryRelation(SpaceOntology.rootSpace,
					SpaceRelation.tsonto_hasSubspace.getURI());
			existingUris.remove(SpaceOntology.metadataSpace);
		} catch (MalformedQueryException e) {
			e.printStackTrace();
		}
		return existingUris;
	}

	/**
	 * list all federations.
	 */
	public Set<URI> listFederations() {
		return federations.keySet();
	}

	// ***** SPACE RELATIONS *****//
	/**
	 * create federation. supports all kinds of spaces.
	 * 
	 * @throws SemanticSpaceException
	 */
	public URI createFederation(Set<ISpaceModel> spaces)
			throws SemanticSpaceException {
		checkSpacesExists(spaces);
		URI federationURI = URIGenerator.createNewRandomUniqueURI();

		FederationModel federationModel = new FederationModel(federationURI,
				spaces);
		federations.put(federationURI, federationModel);
		return federationURI;
	}

	/**
	 * delete federation.
	 * 
	 * @throws SemanticSpaceException
	 */
	public void deleteFederation(URI federationURI)
			throws SemanticSpaceException {
		if (!federations.containsKey(federationURI)) {
			throw new SpaceNotExistsException("federation " + federationURI
					+ " does not exist");
		}
		federations.remove(federationURI);
	}

	/**
	 * returns a federation.
	 * 
	 * @param fedURI
	 * @return
	 */
	public FederationModel getFederation(URI fedURI) {
		return federations.get(fedURI);
	}

	/**
	 * public modification method for SpaceRelations.
	 */
	public void createSpaceRelation(ISpaceModel sSpace, SpaceRelation relation,
			ISpaceModel oSpace) throws SemanticSpaceException {
		createMetadata(sSpace.getContextURI(), relation.getURI(), oSpace
				.getContextURI());
	}

	/**
	 * public modification method for SpaceRelations.
	 */
	public void removeSpaceRelation(ISpaceModel sSpace, SpaceRelation relation,
			ISpaceModel oSpace) throws SemanticSpaceException {
		removeMetadata(sSpace.getContextURI(), relation.getURI(), oSpace
				.getContextURI());
	}

	/**
	 * internal method to create statements in metadataSpace. does not check if
	 * subject/object exists. commits the changes directly afterwards.
	 * 
	 * @param subject
	 * @param relation
	 * @param object
	 * @throws SemanticSpaceException
	 */
	private void createMetadata(URI subject, URI relation, URI object) {
		repository.getModel(SpaceOntology.metadataSpace).addStatement(
				new StatementImpl(null, subject, relation, object));
		commitMetadata();
	}

	private void removeMetadata(URI subject, URI relation, URI object)
			throws SemanticSpaceException {
		repository.getModel(SpaceOntology.metadataSpace).removeStatement(
				new StatementImpl(null, subject, relation, object));
		commitMetadata();
	}

	/**
	 * commits the changes in the metadataSpace.
	 */
	private void commitMetadata() {
		// repository.getModel(SpaceOntology.metadataSpace).close();
		// repository.getModel(SpaceOntology.metadataSpace).open();
		repository.getModel(SpaceOntology.metadataSpace).commit();
	}

	/**
	 * checks if the space exists. metadata-space contains all spaceModels and
	 * their relations.
	 * 
	 * @param space
	 * @return
	 */
	private boolean spaceExists(URI space) {
		return repository.sparqlAsk(new AskQuery(SpaceOntology.metadataSpace,
				"ASK { <" + SpaceOntology.rootSpace + "> <"
						+ SpaceRelation.tsonto_hasSubspace.getURI() + "> <"
						+ space + "> }").toString());
	}

	/**
	 * checks if the space exists.
	 * 
	 * @param space
	 * @throws SpaceNotExistsException
	 */
	private void checkSpaceExists(ISpaceModel space)
			throws SpaceNotExistsException {
		if (!spaceExists(space.getContextURI())) {
			throw new SpaceNotExistsException("space " + space.getContextURI()
					+ " does not exist");
		}
	}

	/**
	 * checks if the spaces exists.
	 * 
	 * @param spaces
	 * @throws SpaceNotExistsException
	 */
	private void checkSpacesExists(Set<ISpaceModel> spaces)
			throws SpaceNotExistsException {
		for (ISpaceModel space : spaces) {
			checkSpaceExists(space);
		}
	}

	/**
	 * recursive internal method to get all subspaces of a space.
	 * 
	 * @param originatingURI
	 * @param spaceURI
	 * @param alreadyVisited
	 * @return all subspaces in this level
	 * @throws MalformedQueryException
	 * @throws SemanticSpaceException
	 */
	private Set<URI> getSubspacesInternal(URI originatingURI, URI spaceURI,
			HashSet<URI> alreadyVisited) {
		Set<URI> subspaces = new HashSet<URI>();
		if (alreadyVisited.contains(spaceURI)) {
			return subspaces;
		}
		alreadyVisited.add(spaceURI);

		Set<URI> uris = queryRelation(spaceURI,
				SpaceRelation.tsonto_hasSubspace.getURI());
		Iterator<URI> iterator = uris.iterator();
		while (iterator.hasNext()) {
			URI uri = iterator.next();
			subspaces.add(uri);
			subspaces.addAll(getSubspacesInternal(originatingURI, uri,
					alreadyVisited));
		}
		return subspaces;
	}

	/**
	 * queries a space relation from 1 space to * spaces. 1 level, not
	 * recursive.
	 * 
	 * @param space
	 * @param relation
	 * @return spaces
	 * @throws MalformedQueryException
	 * @throws SemanticSpaceException
	 */
	private Set<URI> queryRelation(URI space, URI relation) {
		HashSet<URI> uris = new HashSet<URI>();
		String construct = getSparqlConstruct(space, relation);

		try {
			ClosableIterable<Statement> statements = repository.getModel(
					SpaceOntology.metadataSpace).sparqlConstruct(construct);
			ClosableIterator<Statement> iterator = statements.iterator();
			while (iterator.hasNext()) {
				uris.add(iterator.next().getObject().asURI());
			}
		} catch (MalformedQueryException e) {
			e.printStackTrace();
		}
		return uris;
	}

	/**
	 * creates a sparql-construct query to find objects of the specified
	 * relation.
	 * 
	 * @param relation
	 * @return
	 */
	private String getSparqlConstruct(URI space, URI relation) {
		String construct = "Construct {<" + space + "> <" + relation
				+ "> ?o} WHERE {<" + space + "> <" + relation + "> ?o}";
		return construct;
	}

	/**
	 * gets the type of the space from metadataSpace. there should be exactly
	 * one type entry for a space.
	 * 
	 * @param space
	 * @throws SemanticSpaceException
	 * @return
	 */
	private SpaceModelType getSpaceType(URI space)
			throws SemanticSpaceException {
		Set<URI> uris = queryRelation(space, SpaceOntology.tsonto_isTypeOf);
		Iterator<URI> iterator = uris.iterator();

		if (!iterator.hasNext() || uris.size() > 1) {
			throw new SemanticSpaceException("error: there are " + uris.size()
					+ " type-entries for the space " + space);
		}

		return SpaceModelType.getSpaceType(iterator.next());
	}

	/**
	 * shutdown needed for persistent storage of BigOWLIM repository.
	 */
	@Override
	public void shutdown() throws SemanticSpaceException {
		try {
			((Repository) repository.getUnderlyingModelSetImplementation())
					.shutDown();
		} catch (RepositoryException e) {
			throw new SemanticSpaceException(e);
		}
	}
}
