/*
 * Copyright (C) 2010

 * STI Innsbruck
 * ICT - Technologie Park Innsbruck, 2nd Floor
 * Technikerstrasse 21a
 * A-6020 Innsbruck, Austria 
 * 
 * Contact:
 * reto.krummenacher@sti2.at
 * gerald.schrempf@sti2.at (development)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 */
package at.sti2.semanticspaces.impl.model;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Map;

import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.exception.LockException;
import org.ontoware.rdf2go.exception.MalformedQueryException;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.Syntax;
import org.ontoware.rdf2go.model.impl.AbstractModel;
import org.ontoware.rdf2go.model.node.BlankNode;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;
import org.ontoware.rdf2go.model.node.Variable;
import org.openrdf.query.BooleanQuery;
import org.openrdf.query.GraphQuery;
import org.openrdf.query.GraphQueryResult;
import org.openrdf.query.QueryLanguage;
import org.openrdf.rdf2go.GraphIterable;
import org.openrdf.rdf2go.RepositoryQueryResultTable;
import org.openrdf.rdf2go.StatementIterable;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.http.HTTPRepository;

import at.sti2.semanticspaces.api.exceptions.NoSparqlEndpointException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.api.exceptions.SpaceModelRuntimeException;
import at.sti2.semanticspaces.impl.queries.AskQuery;

/**
 * wrapper-class for linked open data (LOD). implements the interface:
 * RDF2Go.Model. Connection to LOD space is opened and closed in every query.
 * stateless querying.
 * 
 * @author gersch
 * 
 */
public class LODModel extends AbstractModel {
	private static final long serialVersionUID = 1L;

	/**
	 * stores the URI of this LOD-space.
	 */
	final URI endpointURI;

	/**
	 * the SPARQL endpoint.
	 */
	HTTPRepository sparqlEndpoint;

	public LODModel(URI space) throws SemanticSpaceException {
		checkEndpointURI(space);

		endpointURI = space;
		sparqlEndpoint = new HTTPRepository(endpointURI.toString(), "");

		try {
			sparqlEndpoint.initialize();
		} catch (RepositoryException e) {
			throw new SpaceModelRuntimeException(e);
		}

		checkEndpointLODFunctionality();
	}

	/**
	 * checks if endpoint URI is a valid and accessible URL. needs to be checked
	 * before, otherwise the HTTPRepository constructor will not return.
	 * 
	 * @param space
	 * @throws SemanticSpaceException
	 */
	private void checkEndpointURI(URI space) throws NoSparqlEndpointException {
		URL url = null;
		HttpURLConnection urlConnection = null;
		int val = 0;

		try {
			url = new URL(space.toString());
			urlConnection = (HttpURLConnection) url.openConnection();
			val = urlConnection.getResponseCode();
		} catch (Exception e) {
			throw new NoSparqlEndpointException(e);
		}

		if (val != 200) {
			throw new NoSparqlEndpointException(
					"Endpoint URI not accessible over HTTP: " + space);
		}
	}

	/**
	 * checks if endpoint accepts a SPARQL ask query. the field "sparqlEndpoint"
	 * has to be initialized. throws an exception if endpoint does not accept
	 * this SPARQL query.
	 * 
	 * @throws SemanticSpaceException
	 * 
	 * @throws SemanticSpaceException
	 */
	private void checkEndpointLODFunctionality() throws SemanticSpaceException {
		try {
			sparqlAsk(new AskQuery(endpointURI, "ASK { ?s ?p ?o }").toString());
		} catch (MalformedQueryException e) {
			throw new NoSparqlEndpointException(e);
		} catch (ModelRuntimeException e) {
			throw new NoSparqlEndpointException(e);
		}
	}

	@Override
	public void addStatement(Resource arg0, URI arg1, Node arg2)
			throws ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	@Override
	public void removeStatement(Resource arg0, URI arg1, Node arg2)
			throws ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void dump() {
		ModelHelper.throwOperationNotSupportedException();
	}

	public URI getContextURI() {
		return endpointURI;
	}

	public boolean isIsomorphicWith(Model arg0) {
		return false;
	}

	public BlankNode createBlankNode() {
		return null;
	}

	public BlankNode createBlankNode(String arg0) {
		return null;
	}

	public boolean isValidURI(String arg0) {
		return false;
	}

	public ClosableIterator<Statement> iterator() {
		return null;
	}

	public boolean isLocked() {
		return false;
	}

	public void lock() throws LockException {
	}

	public void unlock() {
	}

	/**
	 * get statements from SPARQL endpoint. mapping to sparqlConstruct query.
	 */
	public ClosableIterator<Statement> findStatements(
			ResourceOrVariable subject, UriOrVariable predicate,
			NodeOrVariable object) throws ModelRuntimeException {

		String SPO = convert(subject, "?s") + " " + convert(predicate, "?p")
				+ " " + convert(object, "?o");
		String query = "CONSTRUCT { " + SPO + " } WHERE { " + SPO + " }";

		return sparqlConstruct(query).iterator();
	}

	/**
	 * converts URIs to Strings with "<>" brackets around for a SPARQL construct
	 * query.
	 * 
	 * @param element
	 * @param variableName
	 * @return
	 */
	private String convert(NodeOrVariable element, String variableName) {
		if (Variable.ANY.equals(element)) {
			return variableName;
		} else {
			return "<" + element.toString() + ">";
		}
	}

	/**
	 * send graph query to LOD SPARQL endpoint.
	 */
	public ClosableIterable<Statement> sparqlConstruct(String constructQuery)
			throws ModelRuntimeException, MalformedQueryException {
		return new GraphIterable(graphQuery(constructQuery), null);
	}

	/**
	 * send boolean query to LOD SPARQL endpoint.
	 */
	@Override
	public boolean sparqlAsk(String askQuery) throws ModelRuntimeException,
			MalformedQueryException {
		RepositoryConnection conn = null;
		try {
			try {
				conn = sparqlEndpoint.getConnection();
				BooleanQuery booleanQuery = conn.prepareBooleanQuery(
						QueryLanguage.SPARQL, askQuery);
				return booleanQuery.evaluate();
			} finally {
				conn.close();
			}
		} catch (Exception e) {
			throw new SpaceModelRuntimeException(e);
		}
	}

	/**
	 * send graph query to LOD SPARQL endpoint.
	 */
	public ClosableIterable<Statement> sparqlDescribe(String describeQuery)
			throws ModelRuntimeException {
		return new StatementIterable(graphQuery(describeQuery), null);
	}

	/**
	 * send query to LOD SPARQL endpoint.
	 * 
	 * @param query
	 * @return
	 */
	private GraphQueryResult graphQuery(String query) {
		RepositoryConnection conn = null;
		try {
			try {
				conn = sparqlEndpoint.getConnection();
				GraphQuery graphQuery = conn.prepareGraphQuery(
						QueryLanguage.SPARQL, query);
				return graphQuery.evaluate();
			} finally {
				conn.close();
			}
		} catch (Exception e) {
			throw new SpaceModelRuntimeException(e);
		}
	}

	/**
	 * send tuple query to LOD SPARQL endpoint.
	 */
	public QueryResultTable sparqlSelect(String selectQuery)
			throws MalformedQueryException, ModelRuntimeException {
		RepositoryConnection conn = null;
		try {
			try {
				conn = sparqlEndpoint.getConnection();
				return new RepositoryQueryResultTable(selectQuery, conn);
			} finally {
				conn.close();
			}
		} catch (Exception e) {
			throw new SpaceModelRuntimeException(e);
		}
	}

	public void readFrom(Reader arg0) throws IOException, ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void readFrom(InputStream arg0) throws IOException,
			ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void readFrom(Reader arg0, Syntax arg1) throws IOException,
			ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void writeTo(Writer arg0) throws IOException, ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void writeTo(OutputStream arg0) throws IOException,
			ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public void writeTo(Writer arg0, Syntax arg1) throws IOException,
			ModelRuntimeException {
		ModelHelper.throwOperationNotSupportedException();
	}

	public String getNamespace(String arg0) {
		return null;
	}

	public Map<String, String> getNamespaces() {
		return null;
	}

	public void removeNamespace(String arg0) {
	}

	public void setNamespace(String arg0, String arg1)
			throws IllegalArgumentException {
	}

	/**
	 * closes all open resources.
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			sparqlEndpoint.shutDown();
		} catch (RepositoryException e) {
			throw new SpaceModelRuntimeException(e);
		}
		super.finalize();
	}
}