/*
 * Copyright (C) 2010
 * 
 * STI Innsbruck ICT - Technologie Park Innsbruck, 2nd Floor Technikerstrasse
 * 21a A-6020 Innsbruck, Austria
 * 
 * Contact: reto.krummenacher@sti2.at gerald.schrempf@sti2.at (development)
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

package at.sti2.semanticspaces.impl.model;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.QueryRow;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;

import at.sti2.semanticspaces.api.ISemanticSpaceListener;
import at.sti2.semanticspaces.api.ISpaceModel;
import at.sti2.semanticspaces.api.exceptions.OperationNotSupportedException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.api.exceptions.SpaceModelRuntimeException;

/**
 * provides helper-methods for different Model-Implementations like SpaceModel
 * and FederationModel. (needs access to getModel() method)
 * 
 * @author Gerald Schrempf
 * 
 */
public class ModelHelper {
	/**
	 * implements sparqlDescribe for a set of SpaceModels.
	 * 
	 * @param spaces
	 * @param query
	 * @param earlierResults
	 *            will be added to the total results.
	 * @return
	 * @throws ModelRuntimeException
	 */
	public static ClosableIterable<Statement> sparqlDescribe(
			Set<ISpaceModel> spaces, String query,
			ClosableIterator<Statement> earlierResults)
			throws ModelRuntimeException {

		List<ClosableIterator<Statement>> iterators = new LinkedList<ClosableIterator<Statement>>();
		if (earlierResults != null) {
			iterators.add(earlierResults);
		}

		for (ISpaceModel currSpace : spaces) {
			iterators.add(currSpace.sparqlDescribe(query).iterator());
		}
		return new FlattenClosableIterable<Statement>(iterators.iterator());
	}

	/**
	 * implements sparqlConstruct for a set of SpaceModels.
	 * 
	 * @param spaces
	 * @param query
	 * @param earlierResults
	 *            will be added to the total results.
	 * @return
	 * @throws ModelRuntimeException
	 */
	public static ClosableIterable<Statement> sparqlConstruct(
			Set<ISpaceModel> spaces, String query,
			ClosableIterator<Statement> earlierResults)
			throws ModelRuntimeException {

		List<ClosableIterator<Statement>> iterators = new LinkedList<ClosableIterator<Statement>>();
		if (earlierResults != null) {
			iterators.add(earlierResults);
		}

		for (ISpaceModel currSpace : spaces) {
			iterators.add(currSpace.sparqlConstruct(query).iterator());
		}
		return new FlattenClosableIterable<Statement>(iterators.iterator());
	}

	/**
	 * implements sparqlSelect for a set of SpaceModels.
	 * 
	 * @param spaces
	 * @param query
	 * @param earlierResults
	 *            will be added to the total results.
	 * @return
	 */
	public static QueryResultTable sparqlSelect(Set<ISpaceModel> spaces,
			String query, ClosableIterator<QueryRow> earlierResults)
			throws ModelRuntimeException {

		List<ClosableIterator<QueryRow>> queryResultTables = new LinkedList<ClosableIterator<QueryRow>>();
		List<String> variables = null;
		QueryResultTable results = null;

		if (earlierResults != null) {
			queryResultTables.add(earlierResults);
		}

		for (ISpaceModel currSpace : spaces) {
			results = currSpace.sparqlSelect(query);
			queryResultTables.add(results.iterator());
		}

		if (results != null) { // getVariables is needed only once
			variables = results.getVariables();
		}

		return new FlattenQueryResultTable(variables, queryResultTables
				.iterator());
	}

	/**
	 * implements sparqlAsk for a set of SpaceModels.
	 * 
	 * @param spaces
	 * @param query
	 * @return
	 */
	public static boolean sparqlAsk(Set<ISpaceModel> spaces, String query) {
		for (ISpaceModel currSpace : spaces) {
			if (currSpace.sparqlAsk(query)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * implements findStatements for a set of SpaceModels.
	 * 
	 * @param spaces
	 * @param arg0
	 * @param arg1
	 * @param arg2
	 * @param earlierResults
	 *            will be added to the total results.
	 * @return
	 * @throws ModelRuntimeException
	 */
	@SuppressWarnings("unchecked")
	public static ClosableIterator<Statement> findStatements(
			Set<ISpaceModel> spaces, ResourceOrVariable arg0,
			UriOrVariable arg1, NodeOrVariable arg2,
			ClosableIterator<Statement> earlierResults)
			throws ModelRuntimeException {

		List<ClosableIterator<Statement>> iterators = new LinkedList<ClosableIterator<Statement>>();
		if (earlierResults != null) {
			iterators.add(earlierResults);
		}

		for (ISpaceModel currSpace : spaces) {
			iterators.add(currSpace.findStatements(arg0, arg1, arg2));
		}

		return new FlattenClosableIterator(iterators.iterator());
	}

	/**
	 * subscribes to all spaces and returns a Map with the SpaceModels and its
	 * related subscriptionURI. If subscription is not possible on a space,
	 * continue without error.
	 * 
	 * @param spaces
	 * @param query
	 * @param listener
	 * @return
	 * @throws SemanticSpaceException
	 *             if subscribe is not possible.
	 */
	public static Map<ISpaceModel, URI> subscribeToSpaces(
			Set<ISpaceModel> spaces, String query,
			final ISemanticSpaceListener listener)
			throws SemanticSpaceException {
		Map<ISpaceModel, URI> subscriptions = new HashMap<ISpaceModel, URI>();

		for (ISpaceModel currSpace : spaces) {
			try {
				subscriptions.put(currSpace, currSpace.subscribe(query,
						listener));
			} catch (OperationNotSupportedException e) {
				// don't subscribe to this type of space.
			}
		}

		return subscriptions;
	}

	/**
	 * unsubscribes from all spaces-entries in the map.
	 * 
	 * @param map
	 * @throws SemanticSpaceException
	 *             if unsubscribe is not possible.
	 */
	public static void unsubscribeFromSpaces(Map<ISpaceModel, URI> map)
			throws SemanticSpaceException {
		for (Map.Entry<ISpaceModel, URI> entry : map.entrySet()) {
			entry.getKey().unsubscribe(entry.getValue());
		}
	}

	/**
	 * throws an exception with the name of the caller-method.
	 * 
	 * @throws ModelRuntimeException
	 */
	public static void throwOperationNotSupportedException()
			throws ModelRuntimeException {
		StackTraceElement[] stack = Thread.currentThread().getStackTrace();

		String exception;
		if (stack.length > 0) {
			exception = "'" + stack[stack.length - 1].getMethodName() + "'";
		} else {
			exception = "(stackTrace is empty, method unknown)";
		}

		throw new SpaceModelRuntimeException("the method-call " + exception
				+ " is not possible for a federation!");
	}
}
