/*
 * Copyright (C) 2010
 * 
 * STI Innsbruck ICT - Technologie Park Innsbruck, 2nd Floor Technikerstrasse
 * 21a A-6020 Innsbruck, Austria
 * 
 * Contact: reto.krummenacher@sti2.at gerald.schrempf@sti2.at (development)
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

package at.sti2.semanticspaces.impl.model;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.ontoware.aifbcommons.collection.ClosableIterable;
import org.ontoware.aifbcommons.collection.ClosableIterator;
import org.ontoware.rdf2go.exception.LockException;
import org.ontoware.rdf2go.exception.MalformedQueryException;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.Model;
import org.ontoware.rdf2go.model.QueryResultTable;
import org.ontoware.rdf2go.model.Statement;
import org.ontoware.rdf2go.model.Syntax;
import org.ontoware.rdf2go.model.impl.AbstractModel;
import org.ontoware.rdf2go.model.node.BlankNode;
import org.ontoware.rdf2go.model.node.Node;
import org.ontoware.rdf2go.model.node.NodeOrVariable;
import org.ontoware.rdf2go.model.node.Resource;
import org.ontoware.rdf2go.model.node.ResourceOrVariable;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.UriOrVariable;

import at.sti2.semanticspaces.api.ISemanticSpaceListener;
import at.sti2.semanticspaces.api.ISpaceModel;
import at.sti2.semanticspaces.api.exceptions.OperationNotSupportedException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.impl.SemanticSpace;
import at.sti2.semanticspaces.impl.queries.AskQuery;
import at.sti2.semanticspaces.impl.queries.ConstructQuery;
import at.sti2.semanticspaces.impl.queries.DescribeQuery;
import at.sti2.semanticspaces.impl.queries.SelectQuery;

import com.ontotext.trree.rdf2go.api.SubscribableModel;
import com.ontotext.trree.rdf2go.api.Subscriber;

/**
 * wrapper-class for different kinds of models. (OwlimModel, LODModel,
 * OverlayModel ...)
 * 
 * user has to call open() before accessing the other methods. modifications on
 * model are stored persistantly when close() is called.
 * 
 * @author Gerald Schrempf
 * 
 */
public class SpaceModel extends AbstractModel implements ISpaceModel {
	private static final long serialVersionUID = 1L;

	private SemanticSpace kernel = null; // needed to get subspaces

	/**
	 * separate model variable. same like super.model, but no type-casts are
	 * needed for access.
	 */
	private final Model model;

	/**
	 * Map for <subscriptionURI, subSubscriptions<subSpaceModel,
	 * subSubscriptionURI>>. it stores for every subscription on this space a
	 * Map for the subscriptions to the subspaces. the inner map stores the
	 * subspace-SpaceModel and the subscriptionURI.
	 */
	Map<URI, Map<ISpaceModel, URI>> subSubscriptions = new HashMap<URI, Map<ISpaceModel, URI>>();

	/**
	 * stores reference to kernel. set before creation of SpaceModel-objects.
	 */
	public void setExecutionContext(SemanticSpace executionContext) {
		kernel = executionContext;
	}

	/**
	 * returns the executionContext (kernel) which holds this SpaceModel. this
	 * context should allow the access to the rootModel and the space-relations.
	 */
	public SemanticSpace getExecutionContext() {
		return kernel;
	}

	/**
	 * store reference to underlying model and executionContext (SemSpaceImpl).
	 * 
	 * @param model
	 *            reference to underlying model.
	 * @param executionContext
	 *            reference to kernel which manages the spaces and their
	 *            relations.
	 */
	public SpaceModel(Model model, SemanticSpace executionContext) {
		super.model = model;
		this.model = model;
		setExecutionContext(executionContext);
	}

	/**
	 * adding not allowed on every model.
	 */
	@Override
	public void addStatement(Resource arg0, URI arg1, Node arg2)
			throws ModelRuntimeException {
		assertModel();
		model.addStatement(arg0, arg1, arg2);
		model.commit();
	}

	/**
	 * removing not allowed on every model.
	 */
	@Override
	public void removeStatement(Resource arg0, URI arg1, Node arg2)
			throws ModelRuntimeException {
		assertModel();
		model.removeStatement(arg0, arg1, arg2);
		model.commit();
	}

	@Override
	public void close() {
		assertModel();
		model.close();
		super.close();
	}

	@Override
	public Model open() {
		super.open();
		assertModel();
		return model.open();
	}

	public void dump() {
		assertModel();
		model.dump();
	}

	public URI getContextURI() {
		assertModel();
		return model.getContextURI();
	}

	public boolean isIsomorphicWith(Model arg0) {
		assertModel();
		return model.isIsomorphicWith(arg0);
	}

	public BlankNode createBlankNode() {
		assertModel();
		return model.createBlankNode();
	}

	public BlankNode createBlankNode(String arg0) {
		assertModel();
		return model.createBlankNode(arg0);
	}

	public boolean isValidURI(String arg0) {
		assertModel();
		return model.isValidURI(arg0);
	}

	public ClosableIterator<Statement> iterator() {
		assertModel();
		return model.iterator();
	}

	public boolean isLocked() {
		assertModel();
		return model.isLocked();
	}

	public void lock() throws LockException {
		assertModel();
		model.lock();
	}

	public void unlock() {
		assertModel();
		model.unlock();
	}

	/**
	 * findStatements for this space and its subspaces.
	 */
	public ClosableIterator<Statement> findStatements(ResourceOrVariable arg0,
			UriOrVariable arg1, NodeOrVariable arg2)
			throws ModelRuntimeException {
		assertModel();
		ClosableIterator<Statement> results;
		results = model.findStatements(arg0, arg1, arg2);

		return ModelHelper.findStatements(getSubspaces(), arg0, arg1, arg2,
				results);
	}

	/**
	 * sparqlConstruct. recursive over a space and all its subspaces.
	 */
	public ClosableIterable<Statement> sparqlConstruct(String query)
			throws ModelRuntimeException, MalformedQueryException {
		assertModel();
		ClosableIterable<Statement> result;

		String constructQuery = new ConstructQuery(getContextURI(), query)
				.toString();
		result = model.sparqlConstruct(constructQuery);

		return ModelHelper.sparqlConstruct(getSubspaces(), query, result
				.iterator());
	}

	/**
	 * sparqlDescribe. recursive over a space and all its subspaces.
	 */
	public ClosableIterable<Statement> sparqlDescribe(String query)
			throws ModelRuntimeException {
		assertModel();
		String describeQuery;
		ClosableIterable<Statement> results;

		describeQuery = new DescribeQuery(getContextURI(), query).toString();
		results = model.sparqlDescribe(describeQuery);

		return ModelHelper.sparqlDescribe(getSubspaces(), query, results
				.iterator());
	}

	/**
	 * sparqlSelect. recursive over a space and all its subspaces.
	 */
	public QueryResultTable sparqlSelect(String query)
			throws MalformedQueryException, ModelRuntimeException {
		assertModel();
		QueryResultTable results;
		String selectQuery;

		selectQuery = new SelectQuery(getContextURI(), query).toString();
		results = model.sparqlSelect(selectQuery);

		return ModelHelper.sparqlSelect(getSubspaces(), query, results
				.iterator());
	}

	/**
	 * sparqlAsk on space and subspaces.
	 */
	@Override
	public boolean sparqlAsk(String query) throws ModelRuntimeException,
			MalformedQueryException {
		assertModel();
		if (model.sparqlAsk(new AskQuery(getContextURI(), query).toString())) {
			return true;
		} else {
			return ModelHelper.sparqlAsk(getSubspaces(), query);
		}
	}

	public void readFrom(Reader arg0) throws IOException, ModelRuntimeException {
		assertModel();
		model.readFrom(arg0);
	}

	public void readFrom(InputStream arg0) throws IOException,
			ModelRuntimeException {
		assertModel();
		model.readFrom(arg0);
	}

	public void readFrom(Reader arg0, Syntax arg1) throws IOException,
			ModelRuntimeException {
		assertModel();
		model.readFrom(arg0, arg1);
	}

	public void writeTo(Writer arg0) throws IOException, ModelRuntimeException {
		assertModel();
		model.writeTo(arg0);
	}

	public void writeTo(OutputStream arg0) throws IOException,
			ModelRuntimeException {
		assertModel();
		model.writeTo(arg0);
	}

	public void writeTo(Writer arg0, Syntax arg1) throws IOException,
			ModelRuntimeException {
		assertModel();
		model.writeTo(arg0, arg1);
	}

	public String getNamespace(String arg0) {
		assertModel();
		return model.getNamespace(arg0);
	}

	public Map<String, String> getNamespaces() {
		assertModel();
		return model.getNamespaces();
	}

	public void removeNamespace(String arg0) {
		assertModel();
		model.removeNamespace(arg0);
	}

	public void setNamespace(String arg0, String arg1)
			throws IllegalArgumentException {
		assertModel();
		model.setNamespace(arg0, arg1);
	}

	/**
	 * Subscription to a space and its subspaces. maybe not possible for every
	 * model!
	 */
	public URI subscribe(String query, final ISemanticSpaceListener listener)
			throws SemanticSpaceException {
		assertModel();
		if (model instanceof SubscribableModel) {
			Subscriber subscriber = new Subscriber() {
				public void matchingStatement(URI subscription,
						Statement statement) {
					listener.notify(subscription, statement);
				}
			};
			URI subscription = ((SubscribableModel) model).subscribe(query,
					subscriber);

			/**
			 * subscribe to all subspaces.
			 */
			Map<ISpaceModel, URI> subspaceSubscriptions = ModelHelper
					.subscribeToSpaces(getSubspaces(), query, listener);
			subSubscriptions.put(subscription, subspaceSubscriptions);

			return subscription;
		} else {
			throw new OperationNotSupportedException(
					"'Subscribe' is not possible for this model: "
							+ model.getClass());
		}
	}

	/**
	 * Unsubscription from a space its subspaces. maybe not possible for every
	 * model!
	 */
	public void unsubscribe(URI subscription) throws SemanticSpaceException {
		assertModel();
		if (model instanceof SubscribableModel) {
			((SubscribableModel) model).unsubscribe(subscription);
			ModelHelper.unsubscribeFromSpaces(subSubscriptions
					.get(subscription));
		} else {
			throw new OperationNotSupportedException(
					"'Unsubscribe' is not possible for this model: "
							+ model.getClass());
		}
	}

	/**
	 * convenience method for getting the subspaces of this space.
	 * 
	 * @return
	 */
	private Set<ISpaceModel> getSubspaces() {
		assertModel();
		try {
			Set<ISpaceModel> spaceModels = getExecutionContext().getSubspaces(
					getContextURI());
			return spaceModels;
		} catch (SemanticSpaceException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	protected void finalize() throws Throwable {
		if (model.isOpen()) {
			model.close();
		}
		super.finalize();
	}
}
