package at.sti2.semanticspaces.test.lod;

import static org.junit.Assert.assertTrue;

import org.junit.BeforeClass;
import org.junit.Test;
import org.ontoware.rdf2go.exception.MalformedQueryException;
import org.ontoware.rdf2go.exception.ModelRuntimeException;
import org.ontoware.rdf2go.model.TriplePattern;
import org.ontoware.rdf2go.model.impl.TriplePatternImpl;
import org.ontoware.rdf2go.model.node.URI;
import org.ontoware.rdf2go.model.node.impl.URIImpl;

import at.sti2.semanticspaces.api.enums.SpaceModelType;
import at.sti2.semanticspaces.api.exceptions.NoSparqlEndpointException;
import at.sti2.semanticspaces.api.exceptions.SemanticSpaceException;
import at.sti2.semanticspaces.impl.SemanticSpace;
import at.sti2.semanticspaces.test.TestSemanticSpace;

public class TestSemanticSpacesLOD extends TestSemanticSpace {

	static final String CONFIG = "OwlimConfig.properties";
	private static final SpaceModelType localSpaceModelType = SpaceModelType.LocalRepository;
	private static final SpaceModelType modelType_LOD = SpaceModelType.LOD;

	/**
	 * possible sparql endpoints (LOD): http://dbpedia.org/sparql
	 */
	protected static URI testSpaceLOD = new URIImpl("http://dbpedia.org/sparql");

	protected static URI testInvalidEndpoint1 = new URIImpl(
			"http://biocyc.bio2rdf.org/spa_rql");
	protected static URI testInvalidEndpoint2 = new URIImpl(
			"http://www.google.at");

	protected static URI foafURI = new URIImpl("http://xmlns.com/foaf/0.1/");
	protected static URI typeURI = new URIImpl(
			"http://www.w3.org/1999/02/22-rdf-syntax-ns#type");
	protected static URI dataSourceURI = new URIImpl(
			"http://www.openlinksw.com/schema/attribution#DataSource");

	/**
	 * @throws InterruptedException
	 * @throws SemanticSpaceException
	 */
	@BeforeClass
	public static void setup() throws InterruptedException,
			SemanticSpaceException {
		preInitialize();
		TestSemanticSpace.spaceModelType = localSpaceModelType;
		semanticSpace = new SemanticSpace(CONFIG);
		postInitialize();
	}

	/**
	 * creates the needed testSpaces for the following test cases. checks if the
	 * spaces are really created.
	 * 
	 * @throws SemanticSpaceException
	 */
	@Test
	public void test_createSpacesLOD() throws SemanticSpaceException {
		semanticSpace.createSpace(testSpaceLOD, modelType_LOD);
	}

	/**
	 * test invalid HTTP URI.
	 * 
	 * @throws SemanticSpaceException
	 */
	@Test
	public void test_invalidEndpoint1() throws SemanticSpaceException {
		try {
			semanticSpace.createSpace(testInvalidEndpoint1, SpaceModelType.LOD);
			assertTrue("no valid HTTP URI!", false);
		} catch (NoSparqlEndpointException e) {
			assertTrue(true);
		}
	}

	/**
	 * test endpoint which is not a SPARQL endpoint.
	 * 
	 * @throws SemanticSpaceException
	 */
	@Test
	public void test_invalidEndpoint2() throws SemanticSpaceException {
		try {
			semanticSpace.createSpace(testInvalidEndpoint2, SpaceModelType.LOD);
			assertTrue("no SPARQL endpoint!", false);
		} catch (NoSparqlEndpointException e) {
			assertTrue(true);
		}
	}

	@Test
	public void test_selectLOD() throws MalformedQueryException,
			ModelRuntimeException, SemanticSpaceException {
		String query = "SELECT * WHERE {?S ?P ?O} LIMIT 10";
		assertCountSpaceModel(testSpaceLOD, query, 10);
	}

	@Test
	public void test_constructLOD() throws MalformedQueryException,
			ModelRuntimeException, SemanticSpaceException {
		String query = "CONSTRUCT { ?S ?P ?O } WHERE {?S ?P ?O } LIMIT 10";
		int i = super.countResults(semanticSpace.getSpace(testSpaceLOD)
				.sparqlConstruct(query).iterator());

		// printResult(query,
		// semanticSpace.getSpace(testSpaceLOD).sparqlConstruct(query).iterator());
		assertTrue("count " + i + " should be " + 10, i == 10);
	}

	@Test
	public void test_askLOD() throws MalformedQueryException,
			ModelRuntimeException, SemanticSpaceException {
		String query = "ASK { ?s ?p ?o }";
		assertTrue(semanticSpace.getSpace(testSpaceLOD).sparqlAsk(query));
	}

	@Test
	public void test_describeLOD() throws MalformedQueryException,
			ModelRuntimeException, SemanticSpaceException {
		String query = "DESCRIBE ?s where { ?s ?p ?o } LIMIT 10";
		int i = super.countResults(semanticSpace.getSpace(testSpaceLOD)
				.sparqlDescribe(query).iterator());

		// printResult(query,
		// semanticSpace.getSpace(testSpaceLOD).sparqlDescribe(query).iterator());
		assertTrue("count " + i + " should be greater than " + 10, i > 10);
	}

	@Test
	public void test_findStatementsLOD() throws MalformedQueryException,
			ModelRuntimeException, SemanticSpaceException {
		assertTrue(semanticSpace.getSpace(testSpaceLOD).findStatements(foafURI,
				typeURI, dataSourceURI).hasNext());

		TriplePattern triplePattern = new TriplePatternImpl(foafURI, typeURI,
				dataSourceURI);
		assertTrue(semanticSpace.getSpace(testSpaceLOD).findStatements(
				triplePattern).hasNext());
	}
}
